/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.metadata ;

import  java.util.Hashtable ;
import  java.util.Vector ;
import  java.util.Enumeration ;

import  JOP.persist.Hashcode ;

/**
    If instantiated with an instance of a specific class,
    this class will read the 'slot' meta-definition
    information and allow it to be retrieved.
    <p>
    WARNING: This class uses native methods
    <p>
    WARNING: Use of this class bypasses elements of Java's
    security model, specifically instances of any object can be
    created even if the calling application would not normally have
    access to those objects. Use with care.
*/

public class MetaClass
{
    private Object          obj ;
    private Hashtable       slotIndex ;
    private Vector          slots ;
    private static boolean  loaded = false ;

    /**
        Allows a MetaClass to be created for any valid Java classname.
        This method is similar to Class.newInstance() except that
        instances are created of any type even if the calling application
        does not have access.
    */
    public static MetaClass forName(String className) throws Exception
    {
        if (!loaded)
        {
            System.loadLibrary("MetaClass") ;

            loaded = true ;
        }

        Class c = Class.forName(className) ;

        if (c.equals(Class.forName("java.lang.Class")))
        {
            throw new Exception("Cannot access 'java.lang.Class'") ;
        }

        Object  o = null ;

        o = newInstance(c) ;   // smash it in

        // System.out.println("Object = " + className + " " + o.hashCode() ) ;

        return new MetaClass(o) ;
    }

    /**
        Create a new MetaClass for an existing object
    */
    public MetaClass(Object o) throws Exception
    {
        if (!loaded)
        {
            System.loadLibrary("MetaClass") ;

            loaded = true ;
        }

        if (o instanceof Class)
        {
            throw new Exception("Cannot access 'java.lang.Class'") ;
        }

        setObject( o ) ;
    }

    public String toString()
    {
        if (slots == null)
        {
            return null ;
        }
        else
        {
            return slots.toString() ;
        }
    }

    /**
        Dumps an object to the console
    */
    public void printObject()
    {
        for (Enumeration e = slots.elements() ; e.hasMoreElements() ; )
        {
            MetaSlot    s = (MetaSlot)e.nextElement() ;

            System.out.println(s) ;
        }
    }

    public Object getObject()
    {
        return obj ;
    }

    /**
        Calculate a hash for the object and return the
        value of the hash.
    */
    public int getHashcode() throws Exception
    {
        Hashcode h = new Hashcode() ;

        for (Enumeration e = slots.elements() ; e.hasMoreElements() ; )
        {
            MetaSlot s = (MetaSlot)e.nextElement() ;

            if (s.isAccessible())
            {
                s.calcHashcode(h) ;
            }
        }

        return h.getHashcode() ;
    }

    private synchronized void setObject(Object o)
    {
        // System.out.println("META SET: " + o.getClass().getName()) ;

        obj = o ;

        slotIndex = new Hashtable() ;
        slots = new Vector() ;

        Class   c = o.getClass() ;

        while (c != null)
        {
            String classname = c.getName() ;

            int slotCount = getSlotCount( classname ) ;

            // System.out.println("Fields = " + slotCount ) ;

            for (int i = 0 ; i < slotCount ; i++)
            {
                String  signature = getSlotSignature( classname, i  ) ;
                String  name = getSlotName( classname, i ) ;
                int     access = getSlotAccess( classname, i ) ;

                MetaSlot s = new MetaSlot( obj, classname, name, i, signature, access ) ;

                slotIndex.put( classname + "/" + name, s ) ;
                slots.addElement(s) ;
            }

            c = c.getSuperclass() ;
        }
    }

    /**
        Returns an enumerator of the slots of this
        object. The slots are returned in the order they
        are declared. The slots of the subclass are returned
        first, then the slots of the superclass and
        so on.
    */

    public Enumeration slots()
    {
        return slots.elements() ;
    }

    /**
        Returns a slot matching the slot field name and the slot
        classname. Returns 'null' if no matching slot is found.
    */

    public MetaSlot getSlot( String classname, String name ) throws BadSlotName
    {
        MetaSlot s = (MetaSlot)slotIndex.get(classname + "/" + name);

        if (s == null)
        {
            throw new BadSlotName("No slot '" + classname + " " + name + "'" ) ;
        }

        return s ;
    }

    private native int getSlotCount( String classname ) ;
    private native String getSlotSignature( String classname, int n );
    private native String getSlotName( String classname, int n );
    private native int getSlotAccess( String classname, int n );

    private static native Object newInstance(Class c) ;
}

