/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.metadata ;

import  JOP.persist.Hashcode ;

/**
    JOP allows objects to be marshalled directly using native methods.
    There is one instance of this class created for each 'slot' of
    the object.
    <p>
    WARNING: This class uses native methods
    <p>
    WARNING: Use of this class bypasses elements of Java's
    security model, specifically any slot of an object can be
    accessed even if the calling application would not normally have
    access to those slots. Use with care.
*/
public final class MetaSlot
{
    private String  classname ;     // name of the declared class
    private String  signature ;     // signature
    private int     access ;        // access mask
    private Object  object ;        // used by the DLL
    private String  name ;          // name of this slot
    private int     index ;         // slot number

    /**
        This class can only be created by JOP.metadata.MetaClass.
    */
    MetaSlot( Object o, String c, String n, int i, String s, int a )
    {
        object = o ;
        classname = c ;
        name = n ;
        index = i ;
        signature = s ;
        access = a ;
    }

    /**
        Return the class name for a slot
    */
    public String getClassName()
    {
        return classname ;
    }

    /**
        Return the signature of a slot. The trailing ';' is removed.
    */
    public String getSignature()
    {
        if (signature.endsWith(";"))
        {
            return signature.substring(0, signature.length() - 1 ) ;
        }
        else
        {
            return signature ;
        }
    }

    /**
        Return the name for a slot
    */
    public String getSlotName()
    {
        return name ;
    }

    /**
        Return true of the slot is accessible. Static and transient slots
        will return false.
    */
    public boolean isAccessible()
    {
        if (((access & 0x08) == 0x08) ||
             (access & 0x80) == 0x80)
        {
            return false ;
        }
        else
        {
            return true ;
        }
    }

    /**
        Return true of the slot is an array
    */
    public boolean isArray()
    {
        if (signature.charAt(0) == '[')
        {
            return true ;
        }
        else
        {
            return false ;
        }
    }

    /**
        Print out the slot
    */
    public String toString()
    {
        StringBuffer output = new StringBuffer(classname + " " + name + " " + signature + " " + access + " = ") ;

        try
        {
            if (signature.charAt(0) == '[')
            {
                Object  o[] = convertToObjectArray() ;
                if (o == null)
                {
                    output.append("null") ;
                }
                else
                {
                    output.append('{') ;
                    for (int i = 0 ; i < o.length ; i++)
                    {
                        if (i > 0)
                        {
                            output.append(',') ;
                        }

                        if (o[i] == null)
                        {
                            output.append("null") ;
                        }
                        else
                        {
                            output.append(o[i].toString()) ;
                        }
                    }
                    output.append('}') ;
                }
            }
            else
            {
                Object  o = convertToObject() ;
                if (o == null)
                {
                    output.append("null") ;
                }
                else
                {
                    output.append(o.toString()) ;
                }
            }
        } catch (Exception ex)
        {
            output.append("ERR: " + ex.getMessage()) ;
        }

        return output.toString() ;
    }

    /**
        Calculate a hashcode for the slots contents.
    */

    void calcHashcode(Hashcode h) throws Exception
    {
        // System.out.println("META OUT " + classname + "." + name + " (" + signature + ")") ;

        if (isArray())
        {
            switch (signature.charAt(1))
            {
                case 'C' :
                    {
                        char    v[] = charArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break;
                case 'B' :
                    {
                        byte    v[] = byteArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break;
                case 'F' :
                    {
                        float       v[] = floatArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'D' :
                    {
                        double  v[] = doubleArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'I' :
                    {
                        int     v[] = intArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'S' :
                    {
                        short       v[] = shortArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'J' :
                    {
                        long    v[] = longArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'Z' :
                    {
                        boolean     v[] = booleanArray() ;
                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                h.calc(v[i]) ;
                            }
                        }
                    }
                    break ;
                case 'L' :
                    {
                        Object v[] = getObjectArray() ;

                        if (v != null)
                        {
                            for (int i = 0 ; i < v.length ; i++)
                            {
                                if (v[i] != null)
                                {
                                    h.calc(v[i].hashCode()) ;
                                }
                                else
                                {
                                    h.calc(0) ;
                                }
                            }
                        }
                    }
                    break ;
                default :
                    throw new BadSlotType("Unsupported '" + signature + "'") ;
            }
        }
        else
        {
            switch (signature.charAt(0))
            {
                case 'C' :
                    h.calc(charValue()) ;
                    break;
                case 'B' :
                    h.calc(byteValue()) ;
                    break;
                case 'F' :
                    h.calc(floatValue()) ;
                    break ;
                case 'D' :
                    h.calc(doubleValue()) ;
                    break ;
                case 'I' :
                    h.calc(intValue()) ;
                    break ;
                case 'S' :
                    h.calc(shortValue()) ;
                    break ;
                case 'J' :
                    h.calc(longValue()) ;
                    break ;
                case 'Z' :
                    h.calc(booleanValue()) ;
                    break ;
                case 'L' :
                    {
                        Object o = getObject() ;

                        if (o == null)
                        {
                            h.calc(0) ;
                        }
                        else
                        {
                            h.calc(getObject().hashCode()) ;
                        }
                    }
                    break ;
                default :
                    throw new BadSlotType("Unsupported '" + signature + "'") ;
            }
        }
    }

    /**
        Given an object, set the value of the slot to the value
        of the object performing whatever conversions are
        required. Types and accessibility are checked.
    */
    public void convertFromObject(Object obj) throws BadSlotType, ClassNotFoundException
    {
        if (obj != null)
        {
            // System.out.println("META " + classname + "." + name + " (" + signature + ") = '" + obj + " # " + obj.hashCode() + "'" ) ;
        }

        switch (signature.charAt(0))
        {
            case 'C' :
                if (obj == null)
                {
                    set((char)0) ;
                }
                else
                {
                    set(((Character)obj).charValue()) ;
                }
                break;
            case 'B' :
                if (obj == null)
                {
                    set((byte)0) ;
                }
                else
                {
                    byte b = (byte)((Integer)obj).intValue() ;

                    set(b) ;
                }
                break;
            case 'F' :
                if (obj == null)
                {
                    set((float)0) ;
                }
                else
                {
                    set(((Float)obj).floatValue()) ;
                }
                break ;
            case 'D' :
                if (obj == null)
                {
                    set((double)0) ;
                }
                else
                {
                    set(((Double)obj).doubleValue()) ;
                }
                break ;
            case 'I' :
                if (obj == null)
                {
                    set((int)0) ;
                }
                else
                {
                    set(((Integer)obj).intValue()) ;
                }
                break ;
            case 'S' :
                if (obj == null)
                {
                    set((short)0) ;
                }
                else
                {
                    set((short)((Integer)obj).intValue()) ;
                }
                break ;
            case 'J' :
                if (obj == null)
                {
                    set((long)0) ;
                }
                else
                {
                    set(((Long)obj).longValue()) ;
                }
                break ;
            case 'Z' :
                if (obj == null)
                {
                    set((boolean)false) ;
                }
                else
                {
                    set(((Boolean)obj).booleanValue()) ;
                }
                break ;
            case 'L' :
                set( obj ) ;
                break ;
            default :
                throw new BadSlotType("Unsupported '" + signature + "'") ;
        }
    }

    /**
        Given an array of objects, set the value of the slot to the value
        of the array performing whatever conversions are
        required. Types and accessibility are checked.
    */
    public void convertFromObjectArray(Object obj[]) throws BadSlotType, ClassNotFoundException
    {
        if (signature.charAt(0) != '[')
        {
            throw new BadSlotType("Not an array slot") ;
        }

        // System.out.println("META [] " + classname + "." + name + " (" + signature + ") = '" + obj + "'" ) ;

        switch (signature.charAt(1))
        {
            case 'C' :
                if (obj == null)
                {
                    set((char[])null) ;
                }
                else
                {
                    char v[] ;

                    v = new char[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Character [])obj)[i].charValue() ;
                    }

                    set(v) ;
                }
                break;
            case 'B' :
                if (obj == null)
                {
                    set((byte[])null) ;
                }
                else
                {
                    byte v[] ;

                    v = new byte[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = (byte)((Integer [])obj)[i].intValue() ;
                    }

                    set(v) ;
                }
                break;
            case 'F' :
                if (obj == null)
                {
                    set((float[])null) ;
                }
                else
                {
                    float v[] ;

                    v = new float[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Float [])obj)[i].floatValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'D' :
                if (obj == null)
                {
                    set((double[])null) ;
                }
                else
                {
                    double v[] ;

                    v = new double[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Double [])obj)[i].doubleValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'I' :
                if (obj == null)
                {
                    set((int[])null) ;
                }
                else
                {
                    int v[] ;

                    v = new int[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Integer [])obj)[i].intValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'S' :
                if (obj == null)
                {
                    set((short[])null) ;
                }
                else
                {
                    short v[] ;

                    v = new short[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = (short)((Integer [])obj)[i].intValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'J' :
                if (obj == null)
                {
                    set((long[])null) ;
                }
                else
                {
                    long v[] ;

                    v = new long[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Long [])obj)[i].longValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'Z' :
                if (obj == null)
                {
                    set((boolean[])null) ;
                }
                else
                {
                    boolean v[] ;

                    v = new boolean[obj.length] ;

                    for (int i = 0 ; i < v.length ; i++)
                    {
                        v[i] = ((Boolean [])obj)[i].booleanValue() ;
                    }

                    set(v) ;
                }
                break ;
            case 'L' :
                set(obj) ;
                break ;
            default :
                throw new BadSlotType("Unsupported '" + signature + "'") ;
        }
    }

    /**
        Return an object holding the value of the slot. Scalar types are
        converted to their equivalent object types (eg int becomes java.lang.Integer).
        Scalar types 'short' and 'byte' are mapped to java.lang.Integer.
    */
    public Object convertToObject() throws BadSlotType
    {
        // System.out.println("META OUT " + classname + "." + name + " (" + signature + ")") ;

        Object o = null ;

        switch (signature.charAt(0))
        {
            case 'C' :
                o = new Character(charValue()) ;
                break;
            case 'B' :
                byte b = (byte)((int)byteValue() & 0xFF) ;
                o = new Integer(b) ;
                break;
            case 'F' :
                o = new Float(floatValue()) ;
                break ;
            case 'D' :
                o = new Double(doubleValue()) ;
                break ;
            case 'I' :
                o = new Integer(intValue()) ;
                break ;
            case 'S' :
                short s = (short)((int)shortValue() & 0xFFFF) ;
                o = new Integer(s) ;
                break ;
            case 'J' :
                o = new Long(longValue()) ;
                break ;
            case 'Z' :
                o = new Boolean(booleanValue()) ;
                break ;
            case 'L' :
                o = getObject() ;
                break ;
            default :
                throw new BadSlotType("Unsupported '" + signature + "'") ;
        }

        // System.out.println("META OUT " + classname + "." + name + " (" + signature + ") = " + o) ;

        return o ;
    }


    /**
        Return an object array holding the values of the slot. Scalar types are
        converted to their equivalent object types (eg int[] becomes '[java.lang.Integer').
        Scalar types 'short[]' and 'byte[]' are mapped to '[java.lang.Integer".
    */
    public Object[] convertToObjectArray() throws BadSlotType
    {
        // System.out.println("META OUT [] " + classname + "." + name + " (" + signature + ")") ;

        if (signature.charAt(0) != '[')
        {
            throw new BadSlotType("Not an array slot") ;
        }

        Object  o[] ;

        switch (signature.charAt(1))
        {
            case 'C' :
                {
                    char    v[] = charArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Character   obj[] = new Character[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Character(v[i]) ;
                        }

                        o = obj ;
                    }
                }
                break;
            case 'B' :
                {
                    byte    v[] = byteArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Integer obj[] = new Integer[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            byte b = (byte)((int)v[i] & 0xFF) ;

                            obj[i] = new Integer(b) ;
                        }

                        o = obj ;
                    }
                }
                break;
            case 'F' :
                {
                    float       v[] = floatArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Float       obj[] = new Float[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Float(v[i]) ;
                        }

                        o = obj ;
                    }
                }
                break ;
            case 'D' :
                {
                    double  v[] = doubleArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Double  obj[] = new Double[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Double(v[i]) ;
                        }

                        o = obj ;
                    }
                }
                break ;
            case 'I' :
                {
                    int     v[] = intArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Integer obj[] = new Integer[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Integer(v[i]) ;
                        }
                    o = obj ;
                    }
                }
                break ;
            case 'S' :
                {
                    short       v[] = shortArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Integer     obj[] = new Integer[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            short s = (short)((int)v[i] & 0xFFFF) ;

                            obj[i] = new Integer(s) ;
                        }

                        o = obj ;
                    }
                }
                break ;
            case 'J' :
                {
                    long    v[] = longArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Long    obj[] = new Long[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Long(v[i]) ;
                        }

                        o = obj ;
                    }
                }
                break ;
            case 'Z' :
                {
                    boolean     v[] = booleanArray() ;
                    if (v == null)
                    {
                        o = null ;
                    }
                    else
                    {
                        Boolean     obj[] = new Boolean[v.length] ;

                        for (int i = 0 ; i < v.length ; i++)
                        {
                            obj[i] = new Boolean(v[i]) ;
                        }

                        o = obj ;
                    }
                }
                break ;
            case 'L' :
                o = getObjectArray() ;
                break ;
            default :
                throw new BadSlotType("Unsupported '" + signature + "'") ;
        }

        return o ;
    }

    public synchronized char charValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("C"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getCharValue() ;
    }

    public synchronized byte byteValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("B"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getByteValue() ;
    }

    public synchronized float floatValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("F"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getFloatValue() ;
    }

    public synchronized double doubleValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("D"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getDoubleValue() ;
    }

    public synchronized int intValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("I"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getIntValue() ;
    }

    public synchronized short shortValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("S"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getShortValue() ;
    }

    public synchronized long longValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("J"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getLongValue() ;
    }

    public synchronized boolean booleanValue() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("Z"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getBooleanValue() ;
    }

    public synchronized Object getObject() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.startsWith("L"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getObjectValue() ;
    }


    public synchronized char[] charArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[C"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getCharArray() ;
    }

    public synchronized byte[] byteArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[B"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getByteArray() ;
    }

    public synchronized float[] floatArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[F"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getFloatArray() ;
    }

    public synchronized double[] doubleArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[D"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getDoubleArray() ;
    }

    public synchronized int[] intArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[I"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getIntArray() ;
    }

    public synchronized short[] shortArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[S"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getShortArray() ;
    }

    public synchronized long[] longArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[J"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getLongArray() ;
    }

    public synchronized boolean[] booleanArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.equals("[Z"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getBooleanArray() ;
    }

    public synchronized Object[] getObjectArray() throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Slot not accessible") ;
        }

        if (!signature.startsWith("[L"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        return getObjectArrayNative() ;
    }


    public synchronized void set(char v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("C"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setCharValue(v) ;
    }

    public synchronized void set(boolean v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("Z"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setBooleanValue(v) ;
    }

    public synchronized void set(long v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("J"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setLongValue(v) ;
    }


    public synchronized void set(short v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("S"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setShortValue(v) ;
    }


    public synchronized void set(int v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("I"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setIntValue(v) ;
    }


    public synchronized void set(float v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("F"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setFloatValue(v) ;
    }

    public synchronized void set(double v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("D"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setDoubleValue(v) ;
    }


    public synchronized void set(byte v) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("B"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setByteValue(v) ;
    }

    /**
        Sets a slot value to an object. Accessibility and object type
        are checked.
    */
    public synchronized void set(Object v) throws BadSlotType, ClassNotFoundException
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.startsWith("L"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }


        if (v != null)
        {
            //  Validate that the passed object is the
            //  correct type. It must be either an
            //  instance of the class, a sub-class or if the
            //  object is an interface must implement that
            //  interface.

            Class   decClass = Class.forName( signature.substring(1, signature.length()-1)) ;
            Class   varClass = v.getClass() ;

            boolean found = false ;

            //  loop until we reach the ultimate superclass or
            //  we find a match

            while (!found && varClass != null)
            {
                if (decClass.isInterface())
                {
                    // check to see if this class implements the
                    //  interface

                    Class   interfaces[] = varClass.getInterfaces() ;

                    for (int i = 0 ; !found && i < interfaces.length ; i++)
                    {
                        if (decClass.equals(interfaces[i]))
                        {
                            found = true ;
                        }
                    }
                }
                else if (decClass.equals(varClass))
                {
                    //  check to see if this class is the
                    //  required class

                    found = true ;
                }

                if (!found)
                {
                    //  if not the correct class then traverse
                    //  to the superclass

                    varClass = varClass.getSuperclass() ;
                }
            }

            if (!found)
            {
                throw new BadSlotType("Expected '" + signature + "'") ;
            }
        }

        setObjectValue(v);
    }


    public synchronized void set(char v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[C"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setCharArray(v) ;
    }

    public synchronized void set(boolean v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[Z"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setBooleanArray(v) ;
    }

    public synchronized void set(long v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[J"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setLongArray(v) ;
    }


    public synchronized void set(short v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[S"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setShortArray(v) ;
    }


    public synchronized void set(int v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[I"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setIntArray(v) ;
    }


    public synchronized void set(float v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[F"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setFloatArray(v) ;
    }

    public synchronized void set(double v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[D"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setDoubleArray(v) ;
    }


    public synchronized void set(byte v[]) throws BadSlotType
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.equals("[B"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        setByteArray(v) ;
    }

    public synchronized void set(Object v[]) throws BadSlotType, ClassNotFoundException
    {
        if (!isAccessible())
        {
            throw new BadSlotType("Cannot set static slot") ;
        }

        if (!signature.startsWith("[L"))
        {
            throw new BadSlotType("Signature is " + signature ) ;
        }

        if (v != null)
        {
            //  Validate that the passed object is the
            //  correct type. It must be either an
            //  instance of the class, a sub-class or if the
            //  object is an interface must implement that
            //  interface.
            //
            //  First get rid of the 'array' stuff

            String  varClassName = v.getClass().getName() ;

            Class   decClass = Class.forName( signature.substring(2, signature.length()-1)) ;
            Class   varClass = Class.forName( varClassName.substring(2, varClassName.length()-1)) ;

            boolean found = false ;

            //  loop until we reach the ultimate superclass or
            //  we find a match

            while (!found && varClass != null)
            {
                if (decClass.isInterface())
                {
                    // check to see if this class implements the
                    //  interface

                    Class   interfaces[] = varClass.getInterfaces() ;

                    for (int i = 0 ; !found && i < interfaces.length ; i++)
                    {
                        if (decClass.equals(interfaces[i]))
                        {
                            found = true ;
                        }
                    }
                }
                else if (decClass.equals(varClass))
                {
                    //  check to see if this class is the
                    //  required class

                    found = true ;
                }

                if (!found)
                {
                    //  if not the correct class then traverse
                    //  to the superclass

                    varClass = varClass.getSuperclass() ;
                }
            }

            //  ignore errors for now !

            if (!found)
            {
                // throw new BadSlotType("Expected '" + signature + "'") ;
            }
        }

        setObjectArray(v);
    }

    private native boolean getBooleanValue() ;
    private native long getLongValue() ;
    private native short getShortValue() ;
    private native int getIntValue() ;
    private native float getFloatValue() ;
    private native double getDoubleValue() ;
    private native byte getByteValue() ;
    private native char getCharValue() ;
    private native Object getObjectValue() ;

    private native boolean[] getBooleanArray() ;
    private native long[] getLongArray() ;
    private native short[] getShortArray() ;
    private native int[] getIntArray() ;
    private native float[] getFloatArray() ;
    private native double[] getDoubleArray() ;
    private native byte[] getByteArray() ;
    private native char[] getCharArray() ;
    private native Object[] getObjectArrayNative() ;

    private native void setCharValue(char v) ;
    private native void setBooleanValue(boolean v) ;
    private native void setLongValue(long v) ;
    private native void setShortValue(short v) ;
    private native void setIntValue(int v) ;
    private native void setFloatValue(float v) ;
    private native void setDoubleValue(double v) ;
    private native void setByteValue(byte v) ;
    private native void setObjectValue(Object o) ;

    private native void setCharArray(char v[]) ;
    private native void setBooleanArray(boolean v[]) ;
    private native void setLongArray(long v[]) ;
    private native void setShortArray(short v[]) ;
    private native void setIntArray(int v[]) ;
    private native void setFloatArray(float v[]) ;
    private native void setDoubleArray(double v[]) ;
    private native void setByteArray(byte v[]) ;
    private native void setObjectArray(Object o[]) ;
}

