/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

import  JOP.metadata.MetaClass ;
import  JOP.metadata.MetaSlot ;

import  java.util.Enumeration ;
import  java.util.Vector ;
import  java.util.Hashtable ;

public class NativeObjectDefinition implements ObjectDefinition
{
    MetaClass   meta ;
    Object      target ;
    Hashtable   classes ;
    boolean     create = false ;
    Class       targetClass ;

    NativeObjectDefinition(Class c)
    {
        // System.out.println("NATIVE MARSHALL for " + c.getName()) ;

        targetClass = c ;
    }

    public void setPersistManager( PersistManager p )
    {
    }

    /**
        Called by the persist manager before the object
        is read from the persist storage and before any
        call to 'aSlot.setValue()'
    */
    public void beginInput() throws Exception
    {
        setupMetaData() ;

        for (Enumeration e = this.classes() ; e.hasMoreElements() ; )
        {
            NodClassDefinition cdef = (NodClassDefinition)e.nextElement() ;

            cdef.beginInput() ;
        }
    }

    /**
        Called by the persist manager after the object has
        been read from the persist storage and after all
        calls to 'aSlot.setValue()' have been completed.
    */
    public void endInput() throws Exception
    {
        for (Enumeration e = this.classes() ; e.hasMoreElements() ; )
        {
            NodClassDefinition cdef = (NodClassDefinition)e.nextElement() ;

            cdef.endInput() ;
        }
    }

    /**
        Called by the persist manager before the object
        is written to the persist storage and before any
        call to 'aSlot.getValue()'
    */
    public void beginOutput() throws Exception
    {
        setupMetaData() ;

        for (Enumeration e = this.classes() ; e.hasMoreElements() ; )
        {
            NodClassDefinition cdef = (NodClassDefinition)e.nextElement() ;

            cdef.beginOutput() ;
        }
    }

    /**
        Setup the meta data to allow these classes to
        be used.
    */
    private void setupMetaData() throws Exception
    {
        if (create == true && target == null)
        {
            meta = MetaClass.forName(targetClass.getName()) ;

            target = meta.getObject() ;
        }
        else
        {
            // System.out.println("Native marshall " + target.getClass().getName()) ;

            meta = new MetaClass(target) ;
        }

        /*
            Calculate the classes vector
        */

        classes = new Hashtable() ;

        for (Enumeration e = meta.slots() ; e.hasMoreElements() ; )
        {
            MetaSlot    slot = (MetaSlot)e.nextElement() ;

            if (slot.isAccessible())
            {
                NodClassDefinition cdef = (NodClassDefinition)classes.get(slot.getClassName()) ;

                if (cdef == null)
                {
                    cdef = new NodClassDefinition() ;

                    cdef.classname = slot.getClassName() ;
                    cdef.slots = new Vector() ;

                    classes.put( cdef.classname, cdef ) ;
                }

                NodSlotDefinition sdef = new NodSlotDefinition( slot ) ;

                cdef.slots.addElement( sdef ) ;

                // System.out.println("Slot '" + slot.getClassName() + "." + slot.getSlotName() + " is mapped") ;
            }
            else
            {
                // System.out.println("Slot '" + slot.getClassName() + "." + slot.getSlotName() + " is not accessible") ;
            }
        }
    }

    /**
        Called by the persist manager after the object has
        been written to the persist storage and after all
        calls to 'aSlot.getValue()' have been completed.
    */
    public void endOutput()
    {
    }

    /**
        Set the object for which this definition manager
        must provide access.
    */
    public void setObject(Object obj)
    {
        target = obj ;
    }

    /**
        Tell the definition manager to create an instance
        of the object.
    */
    public void createObject()
    {
        create = true ;
    }

    /**
        Returns the object created by this controller
    */
    public Object getObject() throws Exception
    {
        return target ;
    }

    /**
        Return a checksum of the slots of the object. This
        value is used by the PersistManager to determine
        whether the object has changed. Do not return 0.
    */
    public long getChecksum() throws Exception
    {
        int val = meta.getHashcode() ;

        // System.out.println("Hashcode = " + val ) ;

        return val ;
    }

    /**
        Gets the name of the class of the object as a String.
        The classname is as returned by Class.getName().
    */
    public String getClassName()
    {
        return target.getClass().getName() ;
    }

    /**
        Gets the class definitions for this object. The
        class defintiions include the current class
        and all of the super classes. Classnames must be
        returned in order from the current class, the
        current classes superclass, up to java.lang.Object.
    */
    public Enumeration classes()
    {
        return classes.elements() ;
    }
}


class NodClassDefinition implements ClassDefinition
{
    Vector  slots ;
    String  classname ;

    /**
        Returns the classname as a String. This classname
        is the implementation class at this level in the
        inheretence graph of the object.
    */
    public String getClassName()
    {
        return classname ;
    }

    /**
        Returns the SlotDefinitions for each slot defined
        in this class. The slot definitions are returned
        in the same order as they are defined in the
        .java source file
    */
    public Enumeration slots()
    {
        return slots.elements() ;
    }

    /**
        Loop for each slot and invoke 'beginInput'
    */

    void beginInput() throws Exception
    {
        for (Enumeration e = this.slots() ; e.hasMoreElements() ; )
        {
            NodSlotDefinition sdef = (NodSlotDefinition)e.nextElement() ;

            sdef.beginInput() ;
        }
    }
    /**
        Loop for each slot and invoke 'endInput'
    */
    void endInput() throws Exception
    {
        for (Enumeration e = this.slots() ; e.hasMoreElements() ; )
        {
            NodSlotDefinition sdef = (NodSlotDefinition)e.nextElement() ;

            sdef.endInput() ;
        }
    }


    /**
        Loop for each slot and invoke 'beginOutput'
    */

    void beginOutput() throws Exception
    {
        for (Enumeration e = this.slots() ; e.hasMoreElements() ; )
        {
            NodSlotDefinition sdef = (NodSlotDefinition)e.nextElement() ;

            sdef.beginOutput() ;
        }
    }
}


class NodSlotArray
{
    Object              obj ;
    ObjectIdentifier    oid ;
}

class NodSlotDefinition implements SlotDefinition
{
    MetaSlot            slot ;
    ObjectIdentifier    refOid ;
    Vector              elements ;
    Object              slotobj ;
    boolean             nullArray ;

    NodSlotDefinition( MetaSlot s )
    {
        slot = s ;
    }

    /**
        Get the Slot's signature (Java signature as per
        signature.h)
    */
    public String getSignature()
    {
        if (slot.isArray())
        {
            return slot.getSignature().substring(1) ;
        }
        else
        {
            return slot.getSignature() ;
        }
    }

    public int getSignatureType()
    {
        return JavaTypes.getSignatureType( getSignature()) ;
    }

    /**
        Get the name of this slot
    */
    public String getName()
    {
        return slot.getSlotName() ;
    }

    /**
        Return an object that stores the slots value. Note
        that where the original type of the slot was an
        object reference the Object returned will be
        of type ObjectIdentifier - and not the object.
    */
    public Object getValue() throws Exception
    {
        if (slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is an array") ;
        }

        return slot.convertToObject() ;
    }

    /**
        Allows the slot's value to be set. If the slot
        is originally a reference to a mutable object
        then the Object must be of type ObjectIdentifier.
        <p>
        Can only called if the slot is not an array
    */
    public void setValue( Object obj ) throws Exception
    {
        if (slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is an array") ;
        }

        slotobj = obj ;
    }

    /**
        For object references then an object id will be
        assigned to the referenced object. The object id
        will be null for all immutable objects and for
        scalars.
        <p>
        Can only be called if the slot is not an array
    */
    public void setRefOid( ObjectIdentifier oid )
    {
        if (slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is an array") ;
        }

        refOid = oid ;
    }

    /**
        Test if the slot is an array
    */

    public boolean isArray()
    {
        return slot.isArray() ;
    }

    public boolean isNullArray()
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        return nullArray ;
    }

    public void setNullArray()
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        nullArray = true ;
    }

    /**
        Return the length of the array or -1 if the array itself is null.
        <p>
        Can only called if the slot is an array.
    */
    public int sizeArray()
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        if (elements == null)
        {
            return 0 ;
        }
        else
        {
            return elements.size() ;
        }
    }

    /**
        Fetch the 'nth' element. Can only be called
        if the slot is an array
    */

    public Object getValue( int n )
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        if (elements == null)
        {
            return null ;
        }

        NodSlotArray s = (NodSlotArray)elements.elementAt(n) ;

        if (s == null)
        {
            return null ;
        }

        return s.obj ;
    }

    /**
        Set the 'nth' element of an array.
    */

    public void setValue( int n, Object obj )
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        if (elements == null)
        {
            throw new RuntimeException("Array elements not initialised") ;
        }

        while (elements.size() <= n)
        {
            NodSlotArray element = new NodSlotArray() ;
            elements.insertElementAt( element, elements.size()) ;
        }

        NodSlotArray s = (NodSlotArray)elements.elementAt(n) ;

        s.obj = obj ;
    }

    /**
        For object references then an object id will be
        assigned to the referenced object. The object id
        will be null for all immutable objects and for
        scalars.
        <p>
        Can only be called if the slot is not an array
    */
    public void setRefOid( int n, ObjectIdentifier oid )
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        if (elements == null)
        {
            throw new RuntimeException("Array elements not initialised") ;
        }

        while (elements.size() <= n)
        {
            NodSlotArray element = new NodSlotArray() ;
            elements.insertElementAt( element, elements.size()) ;
        }

        NodSlotArray s = (NodSlotArray)elements.elementAt(n) ;

        s.oid = oid ;
    }

    /**
        Return the assigned RefOid for this object
    */
    public ObjectIdentifier getRefOid()
    {
        if (slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is an array") ;
        }

        return refOid ;
    }

    /**
        Return the assigned RefOid for each slot in an array
    */
    public ObjectIdentifier getRefOid(int n)
    {
        if (!slot.isArray())
        {
            throw new RuntimeException("Slot '" + slot.getSlotName() + "' is not an array") ;
        }

        if (elements == null)
        {
            return null ;
        }

        NodSlotArray s = (NodSlotArray)elements.elementAt(n) ;

        if (s == null)
        {
            return null ;
        }

        return s.oid ;
    }

    /**
        Reset the elemetns of the array
    */

    void beginInput() throws Exception
    {
        if (slot.isArray())
        {
            elements = new Vector() ;
        }
    }

    /**
        Save the elements of the array to the object
        instance.
    */

    void endInput() throws Exception
    {
        if (slot.isArray())
        {
            Object  obj[] = null;

            if (!nullArray && elements != null)
            {
                switch (slot.getSignature().charAt(1))
                {
                    case 'C' :
                        obj = new Character[elements.size()] ;
                        break ;
                    case 'B' :
                        obj = new Integer[elements.size()] ;
                        break ;
                    case 'F' :
                        obj = new Float[elements.size()] ;
                        break ;
                    case 'D' :
                        obj = new Double[elements.size()] ;
                        break ;
                    case 'I' :
                        obj = new Integer[elements.size()] ;
                        break ;
                    case 'S' :
                        obj = new Integer[elements.size()] ;
                        break ;
                    case 'J' :
                        obj = new Long[elements.size()] ;
                        break ;
                    case 'Z' :
                        obj = new Boolean[elements.size()] ;
                        break ;
                    case 'L' :
                        obj = new Object[elements.size()] ;
                        break ;
                    default :
                        throw new PersistException("Unsupported '" + slot.getSignature() + "'") ;
                }

                int i = 0 ;

                for (Enumeration e = elements.elements() ; e.hasMoreElements() ; i++ )
                {
                    NodSlotArray s = (NodSlotArray)e.nextElement() ;

                    obj[i] = s.obj ;
                }
            }

            slot.convertFromObjectArray(obj) ;
        }
        else
        {
            slot.convertFromObject( slotobj ) ;
        }
    }

    /**
        Read the value of an array slot into the 'elements'
        vector.
    */
    void beginOutput() throws Exception
    {
        if (slot.isArray())
        {
            Object obj[] = slot.convertToObjectArray() ;

            if (obj != null)
            {
                nullArray = false ;

                elements = new Vector(obj.length) ;

                for (int i = 0 ; i < obj.length ; i++)
                {
                    NodSlotArray    s = new NodSlotArray() ;

                    s.obj = obj[i] ;

                    elements.addElement( s ) ;
                }
            }
            else
            {
                nullArray = true ;
            }
        }
    }
}


