/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

/**
    A class used by the PersistManager to cache object
    values and references. The cache is cleared when
    a unit of work is commenced.
    <p>
    Note that multiple objects in the cache can have the
    same hashcode, and are therefore identical as far as
    Java is concerned (equals() returns true). But these
    objects may have been assigned different JOP object
    identifiers.
    <p>
    Different JOP identifiers are assigned when
    the two identical objects, that override hashCode() are
    saved in separate JOP units of work. Each object will
    receive a unique Object Identifier. However, if the
    objects are both later used within a single unit of
    work then JOP will see the two identical (value
    based) hashcodes but needs to maintain the unique
    object identifiers.
    <p>
    If an update is performed to the value in the database
    outside of JOP then all references to that value
    will change. Java will break if hashCode() is
    overidden for mutable objects.
    <p>
    For the above reason the objCache is a pair of arrays
    and comparisons are based upon the object reference
    not the objects hashcode, the alternative is
    to forbid changes to the database for any java object
    where hashCode() has been overridden.
*/
class ObjectCache
{
    private Object              objCacheVal[] ;
    private ObjectIdentifier    objCacheRef[] ;

    ObjectCache()
    {
        reset() ;
    }

    /**
        Reset the object cache
    */
    void reset()
    {
        objCacheVal = new Object[128] ;             // initial cache size
        objCacheRef = new ObjectIdentifier[128] ;   // initial cache size
    }

    /**
        If an object is in the cache assign a new identifier
        and save the object in the cache, otherwise return
        the existing identifier assigned to the object.
    */
    ObjectIdentifier assignIdentifier( Object obj ) throws Exception
    {
        if (obj == null)
        {
            return null ;
        }

        ObjectIdentifier id = getOid(obj) ;

        if (id == null)
        {
            id = new ObjectIdentifier( obj ) ;

            add( obj, id ) ;
        }

        return id ;
    }

    /**
        Put an identifier in the cache without an
        existing object
    */
    ObjectIdentifier assignIdentifier( long xid ) throws Exception
    {
        ObjectIdentifier id = getOid( xid ) ;

        if (id == null)
        {
            id = new ObjectIdentifier( xid ) ;

            add( null, id ) ;
        }

        // System.out.println("assignIdentifier(" + xid + ") = " + id) ;

        return id ;
    }

    /**
        Remove an object identifier from the cache
    */
    void remove( ObjectIdentifier oid )
    {
        ObjectIdentifier id ;

        removeOid( oid ) ;
    }

    /**
        Ensures that the object identifier can be referenced
        both by using the object id and by using the objects
        value
    */
    void update( ObjectIdentifier oid ) throws Exception
    {
        if (oid.getObject() == null)
        {
            return ;
        }

        add( oid.getObject(), oid ) ;
    }

    /**
        Returns the identifier from the cache for an
        existing object or returns null if no object
        identifier exists in the cache
    */
    ObjectIdentifier getIdentifier( Object obj )
    {
        if (obj == null)
        {
            return null ;
        }

        return getOid( obj ) ;
    }

    /**
        Return an identifier from the cache based upon an
        object id or returns nullif the no object identifier
        with the id exists in the cache.
    */
    ObjectIdentifier getIdentifier( long xid )
    {
        return getOid( xid ) ;
    }

    /******************************************************/

    final private int indexOfObject(Object val)
    {
        for (int i = 0 ; i < objCacheVal.length ; i++)
        {
            if (objCacheVal[i] == val)
            {
                return i ;
            }
        }

        return -1 ;
    }

    final private ObjectIdentifier getOid(long xid)
    {
        for (int i = 0 ; i < objCacheRef.length ; i++)
        {
            if (objCacheRef[i] != null && objCacheRef[i].getOid() == xid)
            {
                return objCacheRef[i] ;
            }
        }

        return null ;
    }


    final private ObjectIdentifier getOid(Object val)
    {
        int index = indexOfObject(val) ;

        if (index == -1)
        {
            return null ;
        }
        else
        {
            return objCacheRef[index] ;
        }
    }

    final private int indexOfOid(ObjectIdentifier ref)
    {
        for (int i = 0 ; i < objCacheRef.length ; i++)
        {
            if (objCacheRef[i] == ref)
            {
                return i ;
            }
        }

        return -1 ;
    }

    final private Object getObject(ObjectIdentifier ref)
    {
        int index = indexOfOid(ref) ;

        if (index == -1)
        {
            return null ;
        }
        else
        {
            return objCacheVal[index] ;
        }
    }

    final private void removeVal(Object val)
    {
        int index = indexOfObject(val) ;

        if (index != -1)
        {
            objCacheVal[index] = null ;
            objCacheRef[index] = null ;
        }
    }

    final private void removeOid(ObjectIdentifier ref)
    {
        int index = indexOfOid(ref) ;

        if (index != -1)
        {
            objCacheVal[index] = null ;
            objCacheRef[index] = null ;
        }
    }

    final private void add(Object val, ObjectIdentifier ref) throws PersistException
    {
        int valIndex = indexOfObject(val) ;
        int refIndex = indexOfOid(ref) ;

        ObjectIdentifier existRef = null ;
        Object existVal = null ;

        if (valIndex != -1)
        {
            existRef = objCacheRef[valIndex] ;
        }

        if (refIndex != -1)
        {
            existVal = objCacheVal[refIndex] ;
        }

        if (existRef != null && existRef != ref)
        {
            throw new PersistException("Cache reference inconsistent") ;
        }

        if (existVal != null && existVal != val)
        {
            throw new PersistException("Cache value inconsistent") ;
        }

        if (refIndex == -1 && valIndex == -1)
        {
            //  insert a new entry

            insert(val, ref) ;
        }
        else if (refIndex != -1 && valIndex == -1)
        {
            objCacheVal[refIndex] = val ;
        }
        else if (refIndex == -1 && valIndex != -1)
        {
            objCacheRef[valIndex] = ref ;
        }
        else if (refIndex != valIndex)
        {
            throw new PersistException("Inconsistent reference and value") ;
        }
    }

    private int insert( Object val, ObjectIdentifier ref )
    {
        //  try to find a null entry

        for (int i = 0 ; i < objCacheVal.length ; i++)
        {
            if (objCacheVal[i] == null && objCacheRef[i] == null)
            {
                objCacheVal[i] = val ;
                objCacheRef[i] = ref ;

                return i ;
            }
        }

        // no space in existing array...extend the array

        Object newVal[] = new Object[objCacheVal.length + 128] ;
        ObjectIdentifier newRef[] = new ObjectIdentifier[objCacheRef.length + 128] ;

        // copy ole elements to new array

        int j ;

        for (j = 0 ; j < objCacheVal.length ; j++)
        {
            newVal[j] = objCacheVal[j] ;
            newRef[j] = objCacheRef[j] ;
        }

        newVal[j] = val ;
        newRef[j] = ref ;

        objCacheVal = newVal ;
        objCacheRef = newRef ;

        return j;
    }
}

