/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

import  java.util.Vector ;
import  java.util.Hashtable ;
import  java.util.Enumeration ;

/**
    This class encapsulates the concept of an object
    identifier. The class includes both the identifier
    and the object to which it refers to.
    <p>
    There are several utility methods that support formatting
    of object identifiers.
*/

public class ObjectIdentifier
{
    private long    oid ;           // the id of the object
    private Object  obj ;           // the actual object
    private long    checksum = 0 ; // checksum of the slots
    private boolean onDatabase ;    // set by PersistManager
    private static Hashtable map = new Hashtable() ;

    /**
        Low limit of OID's suitable for use by
        user applications.
    */
    public static final long USER_LOW = OidGenerator.USER_LOW ;

    /**
        High limit of OID's suitable for use by
        user applications.
    */
    public static final long USER_HIGH = OidGenerator.USER_HIGH ;

    /**
        This method allows a user program pre-define
        an OID for an object. This must be called before the
        object has been streamed out, otherwise an
        automatically assigned random ID will be generated.
        <p>
        User identifiers must be in the range JOP.persist.OidGenerator.USER_LOW
        to JOP.persist.OidGenerator.USER_HIGH
        <p>
        Caution: This method will be moved to PersistManager.
    */
    public static void defineUserOid( Object o, long oid ) throws PersistException
    {
        if (oid < OidGenerator.USER_LOW || oid > OidGenerator.USER_HIGH)
        {
            throw new PersistException("User defined object identifier must be in range " + OidGenerator.USER_LOW + " to " + OidGenerator.USER_HIGH ) ;
        }

        if (map.containsKey(o))
        {
            throw new PersistException("Already mapped") ;
        }

        map.put( o, new Long(oid)) ;
    }

    /**
        Method used by JOP to define objects for internal use.
    */
    static void defineJOPOid( Object o, long oid ) throws PersistException
    {
        if (oid >= OidGenerator.USER_LOW)
        {
            throw new PersistException("JOP defined object identifier must be in range 0 to " + OidGenerator.USER_LOW ) ;
        }

        if (map.containsKey(o))
        {
            throw new PersistException("Already mapped") ;
        }

        map.put( o, new Long(oid)) ;
    }


    /**
        INTERNAL METHOD
    */
    public ObjectIdentifier(Object o)
    {
        Long id = (Long)map.get(o) ;

        if (id == null)
        {
            oid = OidGenerator.newOid() ;
        }
        else
        {
            oid = id.longValue() ;
        }

        obj = o ;
        checksum = 0 ;
        onDatabase = false ;
    }

    /**
        INTERNAL METHOD
    */
    public ObjectIdentifier(Object o, long i)
    {
        obj = o ;
        oid = i ;
        checksum = 0 ;
        onDatabase = false ;
    }

    /**
        INTERNAL METHOD
    */
    public ObjectIdentifier(long xid)
    {
        oid = xid ;
        obj = null ;
        checksum = 0 ;
        onDatabase = false ;
    }

    private static String toHex(int val)
    {
        if (val < 0)
        {
            val = val * -1 ;
        }

        String hex = Integer.toString(val, 16) ;

        while (hex.length() < 4)
        {
            hex = "0" + hex ;
        }

        return hex ;
    }

    private static int fromHex(String hex)
    {
        return Integer.parseInt( hex, 16 ) ;
    }

    /**
        A useful method that takes an object identifier and returns
        a string in xxxx-xxxx-xxxx-xxxx format.
    */

    public static String oid( long oid )
    {
        StringBuffer buffer = new StringBuffer() ;

        for (int i = 0 ; i < 4 ; i++)
        {
            int p = (int)(oid % 65536) ;

            if (i > 0)
            {
                buffer.append("-") ;
            }

            buffer.append(toHex(p)) ;

            oid = oid / 65536L ;
        }

        return buffer.toString() ;
    }

    /*
        A useful method that takes an object identifier in
        xxxx-xxxx-xxxx-xxxx format and returns the long integer
        equivalent.
    */
    public static long oid( String oid )
    {
        long    id = 0 ;

        // System.out.println("Convert <" + oid + ">") ;

        for (int i = 15 ; i >= 0 ; i -= 5 )
        {
            int p = fromHex( oid.substring(i, i+4 )) ;

            id = id * 65536L + p ;
        }

        if (id < 0)
        {
            id = id * -1L ;
        }

        return id ;
    }

    public String toString()
    {
        return oid(oid) ;
    }

    public boolean equals(Object o)
    {
        if (!(o instanceof ObjectIdentifier))
        {
            return false ;
        }

        ObjectIdentifier t = (ObjectIdentifier)o ;

        if (t.oid != oid || t.obj != obj)
        {
            return false ;
        }

        return true ;
    }

    public int hashCode()
    {
        if (obj == null)
        {
            return (int)oid ;
        }
        else
        {
            return (int)oid + obj.hashCode() ;
        }
    }

    /**
        Get the 'long' representation of the object's identifier
    */
    public long getOid()
    {
        return oid ;
    }

    /**
        Set the object's value and then spin through all of
        the queued definitions and set them to the appropriate
        value.
        <p>
        INTERNAL METHOd
    */
    public void setObject(Object o) throws Exception
    {
        obj = o ;
    }

    /**
        INTERNAL METHOD
    */
    public Object getObject()
    {
        return obj ;
    }

    /**
        INTERNAL METHOD
    */
    public long getChecksum()
    {
        return checksum ;
    }

    /**
        This method remembers the checksum of the object
        so that it can be used to optimise updates
        if the object has not changed.
        <p>
        INTERNAL METHOD
    */
    public void setChecksum(long c)
    {
        checksum = c ;
    }

    /**
        INTERNAL METHOD
    */
    public void setOnDatabase()
    {
        onDatabase = true ;
    }

    /**
        INTERNAL METHOD
    */
    public boolean isOnDatabase()
    {
        return onDatabase ;
    }
}
