/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

import java.util.Date ;

/**
    A PersistController is passed to objects that implement
    JOP.persist.Persistable or JOP.persist.PersistMarshall.
    The purpose of the PersistController is to allow the object to...
    <p>
    <ol>
    <li>define its slots and for each slot its type
    <li>save the values of each slot
    <li>load values from the PersistController and save them in the
           objects slot's.
    </ol>
    <p>
    The methods '<b>streamIn()</b>', '<b>streamOut()</b>' and
    '<b>defineObject()</b>' are invoked automatically by the PersistManager.
    The object's implementation of these methods should call methods
    in the PersistController to access slot values or definitions.
    <p>
    The documentation for Persistable contains a complete example of use
    of the Persistable interface.
    <p>
    JOP includes support for all Java scalar types and the Java objects for
    each of those scalar types plus String and java.util.Date. Slots (fields)
    are defined in terms of these types. TYPE_OBJECT indicates that the field
    contains a reference to another object. Arrays (single dimension only)
    are supported.
    <p>
    *@see   JOP.persist.Persistable
*/

public interface PersistController
{
    /**
        Slot contains a 'float' or a 'Float'
    */
    final static PersistSlotType TYPE_FLOAT = new PersistSlotType(JavaTypes.TYPE_FLOAT);
    /**
        Slot contains a 'double' or a 'Double'
    */
    final static PersistSlotType TYPE_DOUBLE = new PersistSlotType(JavaTypes.TYPE_DOUBLE);
    /**
        Slot contains a 'int' or a 'Integer'
    */
    final static PersistSlotType TYPE_INTEGER = new PersistSlotType(JavaTypes.TYPE_INTEGER);
    /**
        Slot contains a 'short'
    */
    final static PersistSlotType TYPE_SHORT = new PersistSlotType(JavaTypes.TYPE_SHORT);
    /**
        Slot contains a 'byte'
    */
    final static PersistSlotType TYPE_BYTE = new PersistSlotType(JavaTypes.TYPE_BYTE);
    /**
        Slot contains a 'long' or a 'Long'
    */
    final static PersistSlotType TYPE_LONG = new PersistSlotType(JavaTypes.TYPE_LONG);
    /**
        Slot contains a 'boolean' or a 'Boolean'
    */
    final static PersistSlotType TYPE_BOOLEAN = new PersistSlotType(JavaTypes.TYPE_BOOLEAN);
    /**
        Slot contains a 'char' or a 'Character'
    */
    final static PersistSlotType TYPE_CHARACTER = new PersistSlotType(JavaTypes.TYPE_CHARACTER);
    /**
        Slot contains a 'String'
    */
    final static PersistSlotType TYPE_STRING = new PersistSlotType(JavaTypes.TYPE_STRING);
    /**
        Slot contains a 'Date'
    */
    final static PersistSlotType TYPE_DATE = new PersistSlotType(JavaTypes.TYPE_DATE);
    /**
        Slot contains an Object of undefined type
    */
    final static PersistSlotType TYPE_OBJECT = new PersistSlotType(JavaTypes.TYPE_OBJECT);

    /**
        Defines the type of a slot. This method can only
        be invoked from within the defineObject() method of
        an object that implements either Persistable
        or PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#defineObject
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param classname   The full name of this class. It should include the
        name of the package.
        *@param slotname    The external (ie the PersistManager's) name of the slot. This
        need not be the same as the internal slot name. Slotnames must be unique for
        a class and all it's superclasses.
        *@param type        One of the value types for the slot.
        *@throws    Exception    The classname, slotname, or slot type were
        incorrect, inconsistent.
    */
    public void defineSlot(String classname, String slotname, PersistSlotType type ) throws Exception ;

    /**
        Defines an array slot where each element of the
        array has the type.
    */

    public void defineSlotArray(String classname, String slotname, PersistSlotType type ) throws Exception ;

    /**
        Retrieve a slot of type 'int' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'int' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public int intValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'char' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'char' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public char charValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'byte' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'byte' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public byte byteValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'short' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'short' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public short shortValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'long' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'long' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public long longValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'float' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'float' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public float floatValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'double' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'double' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public double doubleValue(String name) throws Exception ;

    /**
        Retrieve a slot of type 'boolean' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'boolean' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public boolean booleanValue(String name) throws Exception ;


    /**
        Retrieve an array of type 'int' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'int' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public int[] intArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'char' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'char' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public char[] charArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'byte' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'byte' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public byte[] byteArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'short' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'short' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public short[] shortArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'long' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'long' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public long[] longArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'float' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'float' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public float[] floatArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'double' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'double' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public double[] doubleArray(String name) throws Exception ;

    /**
        Retrieve an array of type 'boolean' with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'boolean' value
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public boolean[] booleanArray(String name) throws Exception ;


    /**
        Retrieve a reference to a 'Date' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Date' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Date getDate(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Character' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Character' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Character getCharacter(String name) throws Exception ;

    /**
        Retrieve a reference to a 'String' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'String' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public String getString(String name) throws Exception ;

    /**
        Retrieve a reference to an 'Integer' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Integer' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Integer getInteger(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Long' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Long' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Long getLong(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Float' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Float' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Float getFloat(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Double' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Double' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Double getDouble(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Boolean' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Boolean' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Boolean getBoolean(String name) throws Exception ;

    /**
        Retrieve a reference to an 'Object' for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        The object will be of the same type as the object
        that was saved in the corresponding 'put()' call.
        The calling method must cast the returned object to
        an object of the appropriate type before assigning it
        to a slot value.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Object'
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Object getObject(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Date array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Date array' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Date[] getDateArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Character array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Character array' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Character[] getCharacterArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'String array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'String array' object reference
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public String[] getStringArray(String name) throws Exception ;

    /**
        Retrieve a reference to an 'Integer array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Integer array' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Integer[] getIntegerArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Long array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Long array' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Long[] getLongArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Float array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Float array' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Float[] getFloatArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Double array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Double array' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Double[] getDoubleArray(String name) throws Exception ;

    /**
        Retrieve a reference to a 'Boolean array' object for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Boolean array' object
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Boolean[] getBooleanArray(String name) throws Exception ;

    /**
        Retrieve a reference to an 'Object' for the slot with the given name. This
        method can only be invoked from within streamIn() of an
        object that implements Persistable or from a class that
        implements PersistMarshall.
        <p>
        The object will be of the same type as the object
        that was saved in the corresponding 'put()' call.
        The calling method must cast the returned object to
        an object of the appropriate type before assigning it
        to a slot value.
        <p>
        *@see   JOP.persist.Persistable#streamIn
        *@see   JOP.persist.PersistMarshall
        <p>
        *@returns   the 'Object'
        *@param     name    The name of the slot. This name must correspond
                            to one of the slot names and types defined
                            in the user provided 'defineObject()' method by a
                            call to 'defineSlot()'.
        *@throws    Exception    If the name is oncorrect or the type is
        incorrect for the name or there is an error detected in the implementation
        of the PersistManager.
    */
    public Object[] getObjectArray(String name) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, int value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, char value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, byte value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, short value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, long value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, float value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, double value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, boolean value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, String value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Date value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Integer value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Character value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Long value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Float value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Double value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Boolean value) throws Exception ;

    /**
        Save a slot value. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The value to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Object value) throws Exception ;


    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, int value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, char value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, byte value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, short value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, long value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, float value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, double value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, boolean value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, String value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Date value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Integer value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Character value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Long value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Float value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Double value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Boolean value[]) throws Exception ;

    /**
        Save a slot array. This method can only be invoked
        from within streamOut() of an object that implements
        Persistable or from a class that implements PersistMarshall.
        The type of the argument and the slotname must match
        a slot defined in the user provided 'defineObject()' method.
        <p>
        *@see   JOP.persist.Persistable#streamOut
        *@see   JOP.persist.PersistMarshall
        <p>
        *@param name    The name of the slot. Must have been
                        previously defined.
        *@param value   The array to be saved.
        *@throws Exception   If the name is incorrect or the
        type does not match the defined type or the PersistManager
        implementation detects an error.
    */
    public void put(String name, Object value[]) throws Exception ;

    /**
        Returns the PeristManager associated with this PersistController.
        <p>
        *@returns   The persist manager
        *@throws    Exception    PersistManager is not defined or
        is not availabe.
    */
    PersistManager getPersistManager() throws Exception ;
}


