/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

import  java.util.Vector;
import  java.util.Hashtable ;
import  java.util.Enumeration ;

import  JOP.util.HashtableDB ;


/**
    The PersistManager is the class that provides JOP's basic functionality.
    Specialisations of this class exist for each type of database (eg
    JDBC or RandomAccessFile).
    <p>
    To save an object a developer needs to do two things. First open
    a persist manager somewhere in the mainline of your program:
    </p>
    <pre>    String  <em>driver</em> = &quot;jdbc.odbc.JdbcOdbcDriver&quot; ;
    String  <em>url</em> = &quot;jdbc:odbc:mydatabase&quot; ;

    PersistManager <em>pmgr</em> = <font color="#0000FF">new</font> JdbcPersistManager( <em>driver</em>,
                                                  <em>url</em>,
                                                  &quot;user&quot;,
                                                  &quot;password&quot;) ;
    <em>pmgr</em>.open() ;</pre>

    <p>and then use the persist manager to save your object. This will also automatically
    save all the objects that are referenced by your object.</p>

    <pre>    YourClass <em>myObject</em> = ... ;
    <font color="#0000FF">long</font> <em>objectId</em> = <em>pmgr</em>.streamOutObject( <em>myObject</em> ) ;</pre>

    <p>The <em>objectId</em> is a unique identifier for your object returned by JOP.
    To retrieve the saved object (and all of its references) in another program or
    in a later session simply pass the <em>objectId</em> as follows:</p>

    <pre>    <em>aNewObject</em> = (YourClass)<em>pmgr</em>.streamInObject( <em>objectId</em> )</pre>

    <p>In the above example <em>aNewObject</em> will have identical content
    to <em>myObject</em>, including all the references of <em>myObject.</em>
    </p><p>
    JOP also supports the saving and retrieval of named objects.</p>
    <p>
    JOP's internal cache ensures that <em>aNewObject</em> and
    <em>myObject</em> will be references to the exact same object if the
    above operations are performed within a unit of work
*/
public abstract class PersistManager
{
    private ObjectCache cache ;
    int                 transLevel = 0 ;
    int                 transIdent = (int)System.currentTimeMillis() ;
    private HashtableDB root ;

    protected PersistManager()
    {
        cache = new ObjectCache() ;
    }


    /**
        Save the object to the persistent database. The object should
        either implement Persistable or have a custom marshall
        class or it will be marshalled automatically using native
        methods. The object and all of the objects referenced by the object
        are saved. Loops are detected and avoided.
        <p>
        To retrieve the object call 'streamInObject()'
        <p>
        *@see   JOP.persist.Persistable
        *@see   JOP.persist.PersistMarshall

        *@param     o   The object to be saved
        *@returns       The unique identifier of the object
        *@throws    java.lang.Exception
    */

    public abstract long streamOutObject(Object o) throws Exception ;

    /**
        Retrieves an object previously saved by 'streamOutObject()'.
        The unique object id is passed and an object and all
        of the objects referenced by the object are
        returned. The object will be an instance of the same type
        that was saved and will have the same context. The
        caller of this method must explicitly
        cast the returned object before using it.
        <p>
        Within a unit of work multiple reads of an object
        either directly (via this method) or indirectly because
        the object is referenced by another object will not
        create secondary instances.
        <p>
        *@param     id  Unique indentifier of the object
        *@returns   The object or throws ObjectNotFoundException
        *@throws    JOP.persist.ObjectNotFoundException if the object id
                    does not correspond to a known object
    */

    public abstract Object streamInObject(long oid) throws Exception ;


    /**
        Save a named object to the persistent database. Named objects
        are useful for well-defined objects to start an application (typically
        a hashtable or vector). The object should
        either implement Persistable or have a custom marshall
        class or it will be marshalled automatically using native
        methods. The object and all of the objects referenced by the object
        are saved. Loops are detected and avoided.
        <p>
        To retrieve the object call 'streamInRoot()' using the same name
        as was used to save the object.
        <p>
        *@see   JOP.persist.Persistable
        *@see   JOP.persist.PersistMarshall

        *@param     o   The object to be saved
        *@throws        java.lang.Exception
    */

    public void streamOutRoot(String name, Object o) throws Exception
    {
        if (root == null)
        {
            try
            {
                root = (HashtableDB)streamInObject( OidGenerator.JOP_ROOT ) ;
            } catch (JOP.persist.ObjectNotFoundException ex)
            {
                root = new HashtableDB() ;

                ObjectIdentifier.defineJOPOid( root, OidGenerator.JOP_ROOT ) ;
            }
        }

        if (root.containsKey( name ))
        {
            root.remove( name ) ;
        }

        root.put( name, o ) ;

        streamOutObject( root ) ;
    }

    /**
        Retrieves an object previously saved by 'streamOutRoot()'.
        The name used to save the object is used to retrieve the
        object. The object and all of the objects referenced by
        the object are returned. The object will be an instance of
        the same type that was saved and will have the same context. The
        caller of this method must explicitly cast the returned object
        before using it.
        <p>
        *@param     id  Unique indentifier of the object
        *@returns   The object or throws ObjectNotFoundException
        *@throws    java.lang.Exception JOP.persist.ObjectNotFoundException if the object id
                    does not correspond to a known object
    */

    public Object streamInRoot(String name ) throws Exception
    {
        if (root == null)
        {
            root = (HashtableDB)streamInObject( OidGenerator.JOP_ROOT ) ;
        }

        if (root.containsKey(name))
        {
            return root.get(name) ;
        }
        else
        {
            throw new ObjectNotFoundException("Object not found") ;
        }
    }

    /**
        Start a unit of work. If the same unique object id
        is referenced multiple times within a unit of work
        then the exact same object instance will be returned each time.
        Once outside of the unit of work a different instance
        may be returned (therefore there may be more than one
        object instance in the application that has the same 'id').
        Objects saved within a unit of work and then read within
        the same unit of work will return the 'dirty' value of
        the object.
        <p>
        When a transactional resource manager (eg a relational
        database) implements this method, it corresponds to the
        start of a transaction. The exact semantics of the unit of
        work are therefore defined for each implementation of the
        PersistManager.
        <p>
        Consistency units can be nested.
        <p>
        *@throws PersistException
    */

    public final void beginWork() throws Exception
    {
        if (transLevel == 0)
        {
            cache.reset() ;

            startTrans() ;
        }

        transLevel++ ;
        transIdent++ ;
    }

    /**
        This method is overridden by the sub-class to start
        a persist manager transaction
        <p>
        <b>INTERNAL METHOD</b>
    */
    protected void startTrans() throws Exception
    {
    }

    /**
        Ends a unit of work. If the resource manager
        supports transactions then the transaction is
        committed (when the nesting level is zero).
        <p>
        *@throws PersistException   If a consistency unit has not
        been started or a PersistManager implemenation specific
        exception.
    */
    public final void commitWork() throws Exception
    {
        if (transLevel == 0)
        {
            throw new PersistException("Not in work");
        }

        if (transLevel == 1)
        {
            commitTrans() ;
        }

        transLevel-- ;
    }

    /**
        This method is overridden by the sub-class to commit
        a persist manager transaction
        <p>
        <b>INTERNAL METHOD</b>
    */
    protected void commitTrans() throws Exception
    {
    }

    /**
        Aborts a unit of work and resets any information
        used to maintain consistency.
    */
    public final void abortWork() throws Exception
    {
        if (transLevel == 0)
        {
            throw new PersistException("Not in work");
        }

        abortTrans() ;

        transLevel = 0 ;
    }

    /**
        This method is overridden by the sub-class to abort
        a persist manager transaction
        <p>
        <b>INTERNAL METHOD</b>
    */
    protected void abortTrans() throws Exception
    {
    }

    /**
        Each consistency unit is assigned an identifier (within the
        current process). This identifier can be useful to
        allow external objects to synchronize their cache with
        the PersistManager.
        <p>
        *@returns   Consistency unit identifier
        *@throws    PersistException   if not in a consistency unit
    */
    public int idWork() throws Exception
    {
        return transIdent ;
    }

    /**
        The open() method is used to commence persistent
        operations. It should be called before any
        other method is called.
    */
    public abstract void open() throws Exception ;

    /**
        The close() method is used to signal the end
        of use of the PersistManager. Once closed a
        PersistManager cannot be used again.
    */
    public abstract void close() throws Exception ;

    /**
        Save the object on the stream. The object must either
        implement Persistable or must have a custom marshall
        class available. The object and all of the objects
        referenced by the object are saved. Loops are detected and
        avoided.
        <p>
        <b>INTERNAL METHOD</b>
        <p>
        To retrieve the object call 'streamInObject()'
        <p>
        *@see   JOP.persist.Persistable
        *@see   JOP.persist.PersistMarshall
        *@see   #streamInObject

        *@param     o   The object to be saved
        *@returns       The unique identifier of the object
        *@throws    java.lang.Exception
    */

    protected ObjectIdentifier streamOutObject(Object obj, Vector history, OutputController cntl, boolean explicit) throws Exception
    {
        // System.out.println("STREAM OUT " + obj.hashCode() + " = " + obj.getClass().getName() ) ;

        if (obj == null)
        {
            throw new PersistException("Can't save NULL object") ;
        }

        ObjectIdentifier oid = cache.getIdentifier( obj ) ;

        /*
            If this object does not have an identifier, then
            create an identifier and assign it to this
            object.
        */
        if (oid == null)
        {
            oid = cache.assignIdentifier( obj ) ;
        }

        if (!history.contains(obj))
        {
            /*
                Add this object to the history to prevent
                loops occuring
            */
            history.addElement(obj) ;

            /*
                Get the objects definition handler. The definition
                handler is our way of accessing the object, ie
                the marshalling mechanism. The definition handler
                can choose to early or late fetch information about
                the object.
            */
            ObjectDefinition defObject = DefinitionManager.getObjectDefinition( obj ) ;

            /*
                Allow the marshalling mechanism to access this
                class if necessary
            */
            defObject.setPersistManager( this ) ;

            /*
                Indicate that we are about to commence output
                of the object
            */
            defObject.beginOutput() ;

            /*
                Walk the object graph and resolve all references
                to other objects by storing the OID for those
                objects in slots that reference other objects.
            */
            resolveReferences( history, defObject, cntl ) ;

            long checksum = defObject.getChecksum() ;

            // System.out.println("OBJ (" + obj.getClass().getName() + ") CSUM = " + checksum + " SAVED CSUM = " + oid.getChecksum()) ;

            if (oid.getChecksum() != checksum)
            {
                /*
                    Test to see if the object has been read, in
                    which case perform an update, otherwise
                    perform an insert.
                */
                if (oid.isOnDatabase())
                {
                    // System.out.println("*** UPDATE ***" + oid ) ;

                    cntl.updateObject( oid, defObject, explicit) ;
                }
                else
                {
                    // System.out.println("*** INSERT ***" + oid) ;

                    cntl.insertObject( oid, defObject, explicit ) ;
                }

                /*
                    Indicate that we have completed output
                    of the object
                */

                oid.setChecksum(checksum) ;
                oid.setOnDatabase() ;
            }

            defObject.endOutput() ;
        }

        return oid ;
    }

    /**
        This method spins the objects definition and
        resolves references to other objects by recursively
        streaming those objects out
        <p>
        <b>INTERNAL METHOD</b>
    */
    void resolveReferences( Vector history, ObjectDefinition defObject, OutputController cntl ) throws Exception
    {
        for (Enumeration eClass = defObject.classes() ; eClass.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)eClass.nextElement() ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                int signature = defSlot.getSignatureType() ;

                if (signature == JavaTypes.TYPE_OBJECT)
                {
                    if (defSlot.isArray())
                    {
                        for (int n = 0 ; n < defSlot.sizeArray() ; n++)
                        {
                            Object  obj = defSlot.getValue(n) ;

                            if (obj != null)
                            {
                                defSlot.setRefOid(n, streamOutObject(obj, history, cntl, false)) ;
                            }
                        }
                    }
                    else
                    {
                        Object  obj = defSlot.getValue() ;

                        if (obj != null)
                        {
                            defSlot.setRefOid(streamOutObject( obj, history, cntl, false)) ;
                        }
                    }
                }
            }
        }
    }

    /**
        Retrieves an object previously saved by 'streamOutObject()'.
        The unique object id is passed and an object and all
        of the objects referenced by the object are
        returned. The object will be an instance of the same type
        that was saved and will have the same context. The
        caller of this method must explicitly
        cast the returned object before using it.
        <p>
        Within a consistency unit multiple reads of an object
        either directly (via this method) or indirectly because
        the object is referenced by another object will not
        create secondary instances.
        <p>
        <b>INTERNAL METHOD</b>

        *@param     id  Unique indentifier of the object
        *@returns       The object
    */

    protected ObjectIdentifier streamInObject(long xid, InputController cntl ) throws Exception
    {
        ObjectIdentifier oid = cache.getIdentifier( xid ) ;

        if (oid == null)
        {
            try
            {
                /*
                    Create an object identifier with a 'null'
                    object reference.
                */
                oid = cache.assignIdentifier( xid ) ;

                /*
                    The object does not exist in the cache so it
                    must be read from the persistent store. If it
                    is not found on the persistent store then an
                    exception will be thrown.
                */
                String classname = cntl.readClassname( oid ) ;

                /*
                    Given the classname, get a object definition
                    controller for the class. Assigning a definition
                    manager can late create an instance of the
                    object if one doesn't exist.
                */
                ObjectDefinition defObject = DefinitionManager.getObjectDefinition( classname ) ;


                /*
                    Allow the marshalling mechanism to access this
                    class if necessary
                */
                defObject.setPersistManager( this ) ;

                /*
                    Start the processing of this object

                    There is an interesting dilemma here...the
                    object can't be completely setup until
                    all references are resolved, however if the
                    references are circular the reference
                    can't be resolved until the object is created.

                    To get around this dilemma the object
                    must be available before its references
                    are expanded. This will cause a problem if the
                    object cannot be modified once created. This is
                    not an issue for Persistable or natively marshalled
                    objects.

                    See the canModify() method for the PersistMarshall
                    interface.
                */
                defObject.beginInput() ;

                cntl.readObject( oid, defObject ) ;

                Object ox = defObject.getObject() ;

                // System.out.println("StreamIn " + xid + " obj = " + ox ) ;

                oid.setObject(ox) ;

                cache.update( oid ) ;

                expandReferences( defObject, cntl ) ;

                /*
                    Finish the processing of this object
                */
                defObject.endInput() ;

                oid.setOnDatabase() ;
                oid.setChecksum(defObject.getChecksum()) ;

                // System.out.println("IN (" + ox.getClass().getName() + ") CSUM SET TO " + oid.getChecksum()) ;
            }
            catch (JOP.persist.ObjectNotFoundException ex)
            {
                cache.remove( oid ) ;

                throw ex ;
            }
        }

        return oid ;
    }

    /**
        This method spins the objects definition and
        resolves references to other objects by recursively
        streaming those objects in. This is invoked
        after an object is streamed in.
        <p>
        <b>INTERNAL METHOD</b>
    */
    void expandReferences( ObjectDefinition defObject, InputController cntl ) throws Exception
    {
        for (Enumeration eClass = defObject.classes() ; eClass.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)eClass.nextElement() ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                int signature = defSlot.getSignatureType() ;

                if (signature == JavaTypes.TYPE_OBJECT)
                {
                    if (defSlot.isArray())
                    {
                        for (int n = 0 ; n < defSlot.sizeArray() ; n++)
                        {
                            ObjectIdentifier i = defSlot.getRefOid( n ) ;

                            if (i == null)
                            {
                                defSlot.setValue(n, null) ;
                            }
                            else
                            {
                                ObjectIdentifier oid = streamInObject(i.getOid(), cntl ) ;

                                defSlot.setRefOid( n, oid ) ;
                                defSlot.setValue( n, oid.getObject()) ;
                            }
                        }
                    }
                    else
                    {
                        ObjectIdentifier i = defSlot.getRefOid() ;

                        if (i == null)
                        {
                            defSlot.setValue(null) ;
                        }
                        else
                        {
                            ObjectIdentifier oid = streamInObject(i.getOid(), cntl ) ;

                            defSlot.setRefOid( oid ) ;
                            defSlot.setValue( oid.getObject()) ;
                        }
                    }
                }
            }
        }
    }

}

