/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

/**
    An interface implemented by classes that are saved or read
    by a PeristManager.
    <p>
    Clases that implement Persistable must:
    <p>
    <ul>
    <li>implement a constructor that takes no arguments
    <p>
    <li>be public
    <p>
    <li>constructor must be public
    </ul>
    <p>
    The PersistManager works by invoking each of these methods
    to define the structure of a class and to marshall the
    objects slots to/from the PersistManager. PersistManager's use
    PeristController to do the actual work.
    <p>
    You do not need to implement JOP.persist.Persistable to allow
    your class to be saved and restored. You can also create a
    custom marshall class or rely upon the automatic (native) marshalling.
    <p>
    The following is an example of the use of the Persistable interface:
    <p>
    <pre>
    <font color="#0000FF">package</font> COM.mycompany ;

    <font color="#0000FF">import</font>  JOP.persist.Persistable ;
    <font color="#0000FF">import</font>  JOP.persist.PersistController ;

    <font color="#0000FF">public</font> MyClass <font color="#0000FF">extends</font> AnotherClass <font color="#0000FF">implements</font> Persistable
    {
        <font color="#0000FF">private</font>     <font color="#0000FF">char</font>    c1 ;
        <font color="#0000FF">private</font>     <font color="#0000FF">String</font>  s2[] ;

        <font color="#0000FF">public</font> MyClass()
        {
            c1 = 'A' ;
            s1 = new String[1] ;

            s1[0] = "Hello world" ;
        }

        <font color="#0000FF">public</font> <font color="#0000FF">void</font> defineObject( PersistController c ) <font color="#0000FF">throws</font> Exception
        {
            <font color="#0000FF">super</font>.defineObject( c ) ;

            c.defineSlot("COM.mycompany.MyClass", "c1", PersistController.TYPE_CHARACTER ) ;
            c.defineSlotArray("COM.mycompany.MyClass", "s2", PersistController.TYPE_STRING ) ;
        }

        <font color="#0000FF">public</font> <font color="#0000FF">void</font> streamOut( PersistController c ) <font color="#0000FF">throws</font> Exception
        {
            <font color="#0000FF">super</font>.streamOut( c ) ;

            c.put( "c1", c1 ) ;
            c.put( "s2", s2 ) ;
        }

        <font color="#0000FF">public</font> <font color="#0000FF">void</font> streamIn( PersistController c ) <font color="#0000FF">throws</font> Exception
        {
            <font color="#0000FF">super</font>.streamIn( c ) ;

            c1 = c.charValue( "c1" ) ;
            s2 = c.getStringArray( "s2" ) ;
        }
    }
    </pre>

    *@see JOP.persist.PersistController
    *@see JOP.persist.PersistMarshall
*/

public interface Persistable
{
    /**
        An object that implements Persistable must define this method. This
        method is called by the PersistManager the first time an object
        of this class is to be saved or retrieved. The purpose of this
        method is to call-back to the PersistController to define the
        object's slots.
        <p>
        Make sure the superclasses defineObject() method is called before
        the body of this method.
        <p>
        Make sure the class name passed to defineSlot is the fully qualified
        classname.
        <p>
        *@see   JOP.persist.PersistManager
        <p>
        *@param c   passed by the PersistManager
        *@throws    java.lang.Exception     Can be thrown
    */
    public void defineObject( PersistController c ) throws Exception ;

    /**
        This method must be implemented by all objects that implement the
        Persistable interface. The method is called by the PersistManager
        when the object is being saved to the persistent database.
        <p>
        Call the superclasses streamOut() method before the bofy of this
        method.
        <p>
        *@see   JOP.persist.PersistManager
        <p>
        *@param c   passed by the PersistManager
        *@throws    java.lang.Exception     Can be thrown
    */
    public void streamOut( PersistController c ) throws Exception ;

    /**
        This method must be implemented by all objects that implement the
        Persistable interface. The method is called by the PersistManager
        when the object is being retrieved from the persistent database.
        <p>
        Call the superclasses streamIn() method before the bofy of this
        method.
        <p>
        *@see   JOP.persist.PersistManager
        <p>
        *@param c   passed by the PersistManager
        *@throws    java.lang.Exception     Can be thrown
    */
    public void streamIn( PersistController c ) throws Exception ;
}

