/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist ;

import  java.util.Enumeration ;
import  java.util.Vector ;
import  java.util.Hashtable ;
import  java.util.Date ;

/**
    This class supports object definitions for objects
    that support the Persistable interface. This class
    and the accompanying PodXxxxx classes act to provide
    a standard interface to objects that implement
    Persistable.
    <p>
    *@see   JOP.persist.MarshallObjectDefinition
    *@see   JOP.persist.NativeObjectDefinition
*/

public class PersistableObjectDefinition implements ObjectDefinition
{
    private     Persistable target ;
    private     Class       targetClass ;
    protected   Hashtable   classes ;
    PersistManager          persistManager ;

    PersistableObjectDefinition( Class c )
    {
        // System.out.println("PERSISTABLE MARSHALL FOR " + c.getName()) ;

        targetClass = c ;
    }

    public void setPersistManager( PersistManager p )
    {
        persistManager = p ;
    }

    /**
        Called by the persist manager before the object
        is read from the persist storage and before any
        call to 'aSlot.setValue()'
    */
    public void beginInput() throws Exception
    {
        // reset the array of slot values. Each slot value
        // references a slot definition

        reset() ;
    }

    /**
        Called by the persist manager after the object has
        been read from the persist storage and after all
        calls to 'aSlot.setValue()' have been completed.
    */
    public void endInput() throws Exception
    {
        //  call target.streamIn( controller ) using the
        //  array of slot values and then resets the
        //  array of slot values.

        if (target == null)
        {
            throw new PersistException("No object") ;
        }

        PodInCntl cntl = new PodInCntl( this ) ;

        target.streamIn( cntl ) ;
    }

    /**
        Called by the persist manager before the object
        is written to the persist storage and before any
        call to 'aSlot.getValue()'
    */
    public void beginOutput() throws Exception
    {
        //  reads the object by calling target.streamOut()
        //  into a temporary array of values, each value
        //  has references to the slot definition.

        if (target == null)
        {
            throw new PersistException("No object") ;
        }

        PodOutCntl cntl = new PodOutCntl( this ) ;

        target.streamOut( cntl ) ;
    }

    /**
        Called by the persist manager after the object has
        been written to the persist storage and after all
        calls to 'aSlot.getValue()' have been completed.
    */
    public void endOutput()
    {
        //  resets the array of slot values

        reset() ;
    }

    private void reset()
    {
        for (Enumeration e = this.classes() ; e.hasMoreElements() ; )
        {
            PodClassDefinition pcdef = (PodClassDefinition)e.nextElement() ;

            pcdef.reset() ;
        }
    }

    /**
        Set the object for which this definition manager
        must provide access. This will occur if we are
        reading the object or updating an existing
        object.
    */
    public void setObject(Object obj) throws Exception
    {
        target = (Persistable)obj ;
    }

    /**
        Tell the definition manager to create an instance
        of the object.
    */
    public void createObject() throws Exception
    {
    }

    /**
        Returns the object created by this controller
    */
    public Object getObject() throws InstantiationException, IllegalAccessException, Exception
    {
        if (target == null)
        {
            target = (Persistable)targetClass.newInstance() ;
        }

        return target ;
    }

    /**
        Return a checksum of the slots of the object. This
        value is used by the PersistManager to determine
        whether the object has changed. Should not ever return
        a value of 0.
    */
    public long getChecksum()
    {
        Hashcode h = new Hashcode() ;

        for (Enumeration e = this.classes() ; e.hasMoreElements() ; )
        {
            PodClassDefinition pcdef = (PodClassDefinition)e.nextElement() ;

            pcdef.calcChecksum(h) ;
        }

        // if (target == null)
        // {
        //    System.out.println("No target checksum = " + h.getHashcode() ) ;
        // }
        // else
        // {
        //    System.out.println(target.getClass().getName() + ":" + target.hashCode() + " checksum = " + h.getHashcode()) ;
        // }

        return h.getHashcode() ;
    }

    /**
        Gets the name of the class of the object as a String.
        The classname is as returned by Class.getName().
    */
    public String getClassName()
    {
        return targetClass.getName() ;
    }

    /**
        Gets the class definitions for this object. The
        class defintiions include the current class
        and all of the super classes. Classnames must be
        returned in order from the current class, the
        current classes superclass, up to java.lang.Object.
        <p>
        The definition is stored as a vector of class
        defintions. Each class definition is a vector
        containing slot definitions.
    */
    public Enumeration classes()
    {
        if (classes == null)
        {
            classes = new Hashtable() ;

            try
            {
                Hashtable cdefs = PodDefnCntl.getDefinition((Persistable)this.getObject()) ;

                for (Enumeration e = cdefs.elements() ; e.hasMoreElements() ; )
                {
                    PodDefnCntlClass cdef = (PodDefnCntlClass)e.nextElement() ;

                    classes.put(cdef.classname, new PodClassDefinition(cdef)) ;
                }
            }
            catch (Exception ex)
            {
                return classes.elements() ;
            }

        }
        return classes.elements() ;
    }
}

/**
    For each class declared of an object, this
    interface allows the PersistManager to marshall
    the slots of the class
*/
class PodClassDefinition implements ClassDefinition
{
    /**
        A reference to the definition information for the
        class.
    */
    PodDefnCntlClass    cref ;
    Hashtable           slots ;

    PodClassDefinition( PodDefnCntlClass c )
    {
        cref = c ;

        slots = new Hashtable() ;

        for (Enumeration e = cref.slots.elements() ; e.hasMoreElements() ; )
        {
            PodDefnCntlSlot sref = (PodDefnCntlSlot)e.nextElement() ;

            slots.put(sref.slotname, new PodSlotDefinition(sref)) ;
        }
    }

    void reset()
    {
        // reset the array of slot values. Each slot value
        // references a slot definition

        for (Enumeration e = slots.elements() ; e.hasMoreElements() ; )
        {
            PodSlotDefinition psdef = (PodSlotDefinition)e.nextElement() ;

            psdef.reset() ;
        }
    }

    void calcChecksum(Hashcode h)
    {
        // reset the array of slot values. Each slot value
        // references a slot definition

        for (Enumeration e = slots.elements() ; e.hasMoreElements() ; )
        {
            PodSlotDefinition psdef = (PodSlotDefinition)e.nextElement() ;

            psdef.calcChecksum(h) ;
        }
    }


    /**
        Returns the classname as a String. This classname
        is the implementation class at this level in the
        inheretence graph of the object.
    */
    public String getClassName()
    {
        return cref.classname ;
    }

    /**
        Returns the SlotDefinitions for each slot defined
        in this class. The slot definitions are returned
        in the same order as they are defined in the
        .java source file
    */
    public Enumeration slots()
    {
        return slots.elements() ;
    }
}


/**
    For each slot an interface to save and fetch the
    slot value.
*/
class PodSlotDefinition implements SlotDefinition
{
    PodDefnCntlSlot     sref ;
    Object              value ;
    ObjectIdentifier    refOid ;
    boolean             nullArray ;

    PodSlotDefinition( PodDefnCntlSlot s)
    {
        sref = s ;
    }

    void reset()
    {
        // reset the array of slot values. Each slot value
        // references a slot definition

        if (sref.array)
        {
            value = new PersistableArray(sref.type) ;
        }
        else
        {
            value = null ;
            refOid = null ;
        }
    }

    /**
        Get the Slot's signature (Java signature as per
        signature.h)
    */
    public String getSignature()
    {
        return  sref.signature ;
    }

    public int getSignatureType()
    {
        return sref.signatureType ;
    }

    /**
        Get the name of this slot
    */
    public String getName()
    {
        return sref.slotname ;
    }

    /**
        Calculate and return the checksum of this slot.
    */

    void calcChecksum(Hashcode h)
    {
        if (value == null)
        {
            h.calc(0) ;
        }
        else if (sref.array)
        {
            int size = ((PersistableArray)value).size() ;

            for (int i = 0 ; i < size ; i++)
            {
                Object o = ((PersistableArray)value).getValue(i) ;

                if (o != null)
                {
                    h.calc(o.hashCode()) ;
                }
                else
                {
                    h.calc(0) ;
                }
            }
        }
        else
        {
            h.calc(value.hashCode()) ;
        }

        // System.out.println(sref.slotname + " (" + sref.signature + ") checksum = " + checksum ) ;
    }

    /**
        Return an object that stores the slots value. Note
        that where the original type of the slot was an
        object reference the Object returned will be
        of type ObjectIdentifier - and not the object.
    */
    public Object getValue()
    {
        return value ;
    }

    /**
        If the signature is a class signature, return true
        if the object is the same type as the
        signature, or is a subclass of the signature. If the
        signature is an interface, return true if the
        object implements the interface
    */
    private boolean typeOfObject( String signature, Object obj ) throws Exception
    {
        Class   decClass = Class.forName( signature.substring(1, signature.length())) ;
        Class   varClass = obj.getClass() ;

        //  loop until we reach the ultimate superclass or
        //  we find a match

        while (varClass != null)
        {
            if (decClass.isInterface())
            {
                // check to see if this class implements the
                //  interface

                Class   interfaces[] = varClass.getInterfaces() ;

                for (int i = 0 ; i < interfaces.length ; i++)
                {
                    if (decClass.equals(interfaces[i]))
                    {
                        return true ;
                    }
                }
            }
            else if (decClass.equals(varClass))
            {
                //  check to see if this class is the
                //  required class

                return true ;
            }

            //  if not the correct class then traverse
            //  to the superclass

            varClass = varClass.getSuperclass() ;
        }

        throw new PersistException("Expected '" + signature + "'") ;
    }

    /**
        Allows the slot's value to be set. If the slot
        is originally a reference to a mutable object
        then the Object must be of type ObjectIdentifier.
        <p>
        Can only called if the slot is not an array
    */
    public void setValue( Object obj ) throws Exception
    {
        if (sref.array)
        {
            if (!(obj instanceof PersistableArray))
            {
                throw new RuntimeException("Slot '" + sref.slotname + "' is an array") ;
            }
        }
        else
        {
            if (obj != null)
            {
                if (!typeOfObject( sref.signature, obj ))
                {
                    throw new PersistException("Bad object type") ;
                }
            }
        }

        value = obj ;
    }

    /**
        For object references then an object id will be
        assigned to the referenced object. The object id
        will be null for all immutable objects and for
        scalars.
        <p>
        Can only be called if the slot is not an array
    */
    public void setRefOid( ObjectIdentifier o )
    {
        if (sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is an array") ;
        }

        refOid = o ;
    }

    /**
        Test if the slot is an array
    */

    public boolean isArray()
    {
        return sref.array ;
    }

    public boolean isNullArray()
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }


        return nullArray ;
    }

    public void setNullArray()
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        nullArray = true ;
    }

    /**
        Return the length of the array.
        <p>
        Can only called if the slot is an array.
    */
    public int sizeArray()
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        if (nullArray)
        {
            return 0 ;
        }
        else if (value == null)
        {
            nullArray = true ;

            return 0 ;
        }
        else
        {
            return ((PersistableArray)value).size() ;
        }
    }

    /**
        Fetch the 'nth' element. Can only be called
        if the slot is an array
    */

    public Object getValue( int n )
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        return ((PersistableArray)value).getValue(n) ;
    }

    /**
        Set the 'nth' element of an array.
    */

    public void setValue( int n, Object obj )
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        ((PersistableArray)value).setValue(n, obj) ;
    }


    /**
        For object references then an object id will be
        assigned to the referenced object. The object id
        will be null for all immutable objects and for
        scalars.
        <p>
        Can only be called if the slot is not an array
    */
    public void setRefOid( int n, ObjectIdentifier oid )
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        ((PersistableArray)value).setRefOid(n, oid) ;
    }

    /**
        Return the assigned RefOid for this object
    */
    public ObjectIdentifier getRefOid()
    {
        if (sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is an array") ;
        }

        return refOid ;
    }

    /**
        Return the assigned RefOid for each slot in an array
    */
    public ObjectIdentifier getRefOid(int n)
    {
        if (!sref.array)
        {
            throw new RuntimeException("Slot '" + sref.slotname + "' is not an array") ;
        }

        return ((PersistableArray)value).getRefOid(n) ;
    }
}


class PodDefnCntlClass
{
    String      classname ;
    Hashtable   slots ;
}

class PodDefnCntlSlot
{
    String              slotname ;
    String              signature ;
    int                 signatureType ;
    boolean             array;
    PersistSlotType     type ;
}

class PodDefn implements PersistController
{
    PersistableObjectDefinition odef ;

    PodDefn()
    {
    }

    PodDefn( PersistableObjectDefinition c)
    {
        odef = c ;
    }

    /**
        Returns the slot definition with a given name. Assumes
        that no two slots have the same name.
    */
    protected PodSlotDefinition getSlotDefinition( String name ) throws PersistException
    {
        for (Enumeration e = odef.classes() ; e.hasMoreElements() ; )
        {
            PodClassDefinition cdef = (PodClassDefinition)e.nextElement() ;

            PodSlotDefinition sdef = (PodSlotDefinition)cdef.slots.get( name ) ;

            if (sdef != null)
            {
                // found what we are looking for

                return sdef ;
            }
        }

        throw new PersistException("Bad slot name '" + name + "'") ;
    }

    public void defineSlotArray(String slotClassName, String slotName, PersistSlotType t) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void defineSlot(String slotClassName, String slotName, PersistSlotType t) throws Exception
    {
        throw new PersistException("No method") ;
    }

    /************ throw errors on illegal calls ************/

    public int intValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public short shortValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public char charValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public byte byteValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public long longValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public float floatValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Float getFloat(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public double doubleValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Double getDouble(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public boolean booleanValue(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Date getDate(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Character getCharacter(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public String getString(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Integer getInteger(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Long getLong(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Boolean getBoolean(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Object getObject(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public int[] intArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public short[] shortArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public char[] charArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public byte[] byteArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public long[] longArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public float[] floatArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public double[] doubleArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public boolean[] booleanArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Float[] getFloatArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Double[] getDoubleArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Date[] getDateArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Character[] getCharacterArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public String[] getStringArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Integer[] getIntegerArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Long[] getLongArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Boolean[] getBooleanArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public Object[] getObjectArray(String name) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, int value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, char value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, byte value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, short value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Character value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, long value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, float value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, double value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, boolean value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, String value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Date value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Integer value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Long value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Float value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Double value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Boolean value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Object value) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, int value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, char value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, byte value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, short value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Character value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, long value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, float value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, double value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, boolean value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, String value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Date value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Integer value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Long value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Float value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Double value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Boolean value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public void put(String name, Object value[]) throws Exception
    {
        throw new PersistException("No method") ;
    }

    public PersistManager getPersistManager() throws Exception
    {
        return odef.persistManager ;
    }
}

class PodDefnCntl extends PodDefn
{
    private static Hashtable    definitions = new Hashtable() ;
    private Hashtable           classes ;
    private String              classname ;

    private PodDefnCntl()
    {
    }

    static Hashtable getDefinition( Persistable o ) throws Exception
    {
        Hashtable classes = (Hashtable)definitions.get(o.getClass().getName()) ;

        if (classes == null)
        {
            PodDefnCntl c = new PodDefnCntl() ;

            classes = c.readDefinition(o) ;

            definitions.put(o.getClass().getName(), classes ) ;
        }

        return classes ;
    }

    private Hashtable readDefinition( Persistable o ) throws Exception
    {
        classname = o.getClass().getName() ;

        classes = new Hashtable() ;

        /*
            Invoke the Persistable object to define itself
        */

        o.defineObject( this ) ;

        return classes ;
    }

    public void defineSlot(String slotClassName, String slotName, PersistSlotType t) throws Exception
    {
        _defineSlot( slotClassName, slotName, t, false ) ;
    }

    public void defineSlotArray(String slotClassName, String slotName, PersistSlotType t) throws Exception
    {
        _defineSlot( slotClassName, slotName, t, true ) ;
    }

    private void _defineSlot(String slotClassName, String slotName, PersistSlotType t, boolean array) throws Exception
    {
        /*
            Validate that the classname passed by the
            object is the objects class or the classname
            of one of the object's superclasses.
        */

        Class   defnClass = null ;
        Class   slotClass = null ;

        defnClass = Class.forName( classname ) ;
        slotClass = Class.forName( slotClassName ) ;

        boolean found = false ;

        while (!found && defnClass != null)
        {
            if (defnClass.equals(slotClass))
            {
                found = true ;
            }
            else
            {
                defnClass = defnClass.getSuperclass() ;
            }
        }

        if (!found)
        {
            throw new PersistException("Class '" + slotClassName + "' is not correct (slot = '" + slotName + "'") ;
        }

        /*
            Get the class definition for this class
        */

        PodDefnCntlClass classdef = (PodDefnCntlClass)classes.get(slotClassName) ;

        if (classdef == null)
        {
            classdef = new PodDefnCntlClass() ;

            classdef.classname = slotClassName ;

            classdef.slots = new Hashtable() ;

            classes.put( slotClassName, classdef ) ;
        }

        if (classdef.slots.containsKey(slotName))
        {
            throw new PersistException("Slot '" + slotName + "' already defined") ;
        }

        PodDefnCntlSlot  slotdef = new PodDefnCntlSlot() ;

        slotdef.slotname = slotName ;
        slotdef.signature = t.getSignature() ;
        slotdef.signatureType = t.getSignatureType() ;
        slotdef.type = t ;

        if (array)
        {
            slotdef.array = true ;
        }
        else
        {
            slotdef.array = false ;
        }

        classdef.slots.put( slotName, slotdef ) ;
    }
}


class PodOutCntl extends PodDefn
{
    PodOutCntl( PersistableObjectDefinition c)
    {
        super(c) ;
    }

    /*********** Methods to access the object's slots **********/

    public void put(String name, int value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Integer(value) ) ;
    }

    public void put(String name, char value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Character(value) ) ;
    }

    public void put(String name, byte value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Integer(value) ) ;
    }

    public void put(String name, short value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Integer(value) ) ;
    }

    public void put(String name, Character value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, long value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Long(value) ) ;
    }

    public void put(String name, float value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Float(value) ) ;
    }

    public void put(String name, double value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Double(value) ) ;
    }

    public void put(String name, boolean value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( new Boolean(value) ) ;
    }

    public void put(String name, String value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Date value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Integer value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Long value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Float value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Double value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Boolean value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, Object value) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        sdef.setValue( value ) ;
    }

    public void put(String name, byte value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, short value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, int value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, long value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, float value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, double value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, boolean value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, char value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Integer value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Long value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Boolean value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Float value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Double value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Character value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, String value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }

    public void put(String name, Date value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }


    public void put(String name, Object value[]) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        PersistableArray    par = new PersistableArray( sdef.sref.type ) ;

        if (value == null)
        {
            sdef.setNullArray() ;
        }
        else
        {
            par.putArray( value ) ;

            sdef.setValue( par ) ;
        }
    }


    /********* end of code to get slot value **************/
}


class PodInCntl extends PodDefn
{
    PodInCntl( PersistableObjectDefinition c)
    {
        super(c) ;
    }

    /************ methods to access the persist info ************/

    public int intValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return ((Integer)sdef.getValue()).intValue() ;
        }
    }

    public short shortValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return (short)((Integer)sdef.getValue()).intValue() ;
        }
    }

    public char charValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return ((Character)sdef.getValue()).charValue() ;
        }
    }

    public byte byteValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return (byte)((Integer)sdef.getValue()).intValue() ;
        }
    }

    public long longValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return ((Long)sdef.getValue()).longValue() ;
        }
    }

    public float floatValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return ((Float)sdef.getValue()).floatValue() ;
        }
    }

    public Float getFloat(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Float)sdef.getValue() ;
    }

    public double doubleValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return 0 ;
        }
        else
        {
            return ((Double)sdef.getValue()).doubleValue() ;
        }
    }

    public Double getDouble(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Double)sdef.getValue() ;
    }

    public boolean booleanValue(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.getValue() == null)
        {
            return false ;
        }
        else
        {
            return ((Boolean)sdef.getValue()).booleanValue() ;
        }
    }

    public Date getDate(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Date)sdef.getValue() ;
    }

    public Character getCharacter(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Character)sdef.getValue() ;
    }

    public String getString(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (String)sdef.getValue() ;
    }

    public Integer getInteger(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Integer)sdef.getValue() ;
    }

    public Long getLong(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Long)sdef.getValue() ;
    }

    public Boolean getBoolean(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return (Boolean)sdef.getValue() ;
    }

    public Object getObject(String name) throws PersistException
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        return sdef.getValue() ;
    }


    public Integer[] getIntegerArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getIntegerArray() ;
        }
    }

    public Long[] getLongArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getLongArray() ;
        }
    }

    public Boolean[] getBooleanArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getBooleanArray() ;
        }
    }

    public Float[] getFloatArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getFloatArray() ;
        }
    }

    public Double[] getDoubleArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getDoubleArray() ;
        }
    }

    public Character[] getCharacterArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getCharacterArray() ;
        }
    }

    public String[] getStringArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getStringArray() ;
        }
    }

    public Date[] getDateArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getDateArray() ;
        }
    }

    public Object[] getObjectArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).getObjectArray() ;
        }
    }

    public byte[] byteArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).byteValueArray() ;
        }
    }

    public short[] shortArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).shortValueArray() ;
        }
    }

    public int[] intArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).intValueArray() ;
        }
    }

    public long[] longArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).longValueArray() ;
        }
    }

    public boolean[] booleanArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).booleanValueArray() ;
        }
    }

    public float[] floatArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).floatValueArray() ;
        }
    }

    public double[] doubleArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).doubleValueArray() ;
        }
    }

    public char[] charArray(String name) throws Exception
    {
        PodSlotDefinition   sdef = getSlotDefinition( name ) ;

        if (sdef.isNullArray())
        {
            return null ;
        }
        else
        {
            return ((PersistableArray)sdef.getValue()).charValueArray() ;
        }
    }
}

