/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.file ;

import  java.io.RandomAccessFile ;

/**
    A class that allows variable length blocks to be
    inserted, updated and read from a RandomAccessFile. The
    structure of each block is:

        long    offset
        int     allocated_length
        boolean free_block_indicator (true if block is free)
        int     used_length
        ...     user data ...
*/
class FileBlock
{
    RandomAccessFile    rfa ;

    FileBlock( RandomAccessFile f )
    {
        rfa = f ;
    }

    /**
        Inserts a new block and returns the address of the
        new block
    */
    long insertBlock( byte buf[], int len ) throws Exception
    {
        long    offset ;

        synchronized(rfa)
        {
            offset = rfa.length() ;

            rfa.seek(offset) ;

            rfa.writeLong(offset) ;     // current position
            rfa.writeInt( len ) ;       // allocated length
            rfa.writeBoolean(false) ;   // not free
            rfa.writeInt( len ) ;       // used length
            rfa.write( buf, 0, len ) ;
        }

        // System.out.println("BLOCK INSERT " + offset + " (" + len + ")") ;

        return offset;
    }

    /**
        Tries to re-write an existing block. If insufficient
        space is available then an insert is performed
        and the old block is marked as 'free'.
    */
    void updateBlock( long offset, byte buf[], int len ) throws Exception
    {
        synchronized(rfa)
        {
            rfa.seek(offset) ;

            if (rfa.readLong() != offset)
            {
                throw new FilePersistException("Corrupt at " + offset ) ;
            }

            if (rfa.readInt() < len)
            {
                throw new FileBlockTooShortException() ;
            }

            if (rfa.readBoolean() == true)
            {
                throw new FilePersistException("Can't update free block") ;
            }

            rfa.writeInt( len ) ;       // used length
            rfa.write( buf, 0, len ) ;  // block data
        }

        // System.out.println("BLOCK UPDATE " + offset + " (" + len + ")") ;

        readBlock( offset ) ;
    }

    /**
        Reads a block
     */

    byte[] readBlock( long offset ) throws Exception
    {
        int     length ;
        byte    buf[] ;

        synchronized(rfa)
        {
            rfa.seek(offset) ;

            if (rfa.readLong() != offset)
            {
                throw new FilePersistException("Corrupt at " + offset ) ;
            }

            length = rfa.readInt() ;

            if (rfa.readBoolean() == true)
            {
                throw new FilePersistException("Can't read free block") ;
            }

            length = rfa.readInt() ;

            buf = new byte[length] ;

            rfa.read( buf, 0, length ) ;
        }

        return buf ;
    }

    /**
        Marks a block as free
    */
    void freeBlock( long offset ) throws Exception
    {
        long    length ;

        synchronized(rfa)
        {
            rfa.seek(offset) ;

            if (rfa.readLong() != offset)
            {
                throw new FilePersistException("Corrupt at " + offset ) ;
            }

            length = rfa.readInt() ;

            rfa.writeBoolean(true) ;
            rfa.writeInt(0) ;
        }

        // System.out.println("BLOCK FREE " + offset + " (" + length + ")") ;
    }
}

