/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.file ;

import  java.io.ByteArrayOutputStream ;
import  java.io.ByteArrayInputStream ;
import  java.io.DataOutputStream ;
import  java.io.DataInputStream ;

import  JOP.persist.ObjectIdentifier ;

/**
    This class supports the index for object id's in the
    random access file. The index is interspersed
    throughout the file.
    <p>
    This is a simplified version of the index. Eventually
    I will support either a B-Tree or a more sophisticated
    hashtable.
    <p>
    The current implementation is pretty straight-forward. There
    is an index control block. This must be at offset 0 in the
    file. When the index control block is created it allocates
    INDEX_BUCKETS index blocks. Each index block is initially
    BLOCK_ALLOC bytes in length. Each fixed length key is
    KEY_LENGTH + 8 bytes in length.
    Keys are hashed into the index blocks
    using a straightforward divisor algorithm, ie I assume
    that the OID generator generates random values. If an
    index block grows it is re-written and the index control
    block is updated.
    <p>
    This class will be replaced very soon !
*/
class FileIndex
{
    FileBlock   fbk ;

    static final int INDEX_BUCKETS = 127 ;
    static final int KEY_ALLOC = 64 ;
    static final int KEY_EXTEND = 16 ;
    static final int KEY_LENGTH = 8 ;
    int     blockCount ;
    long    blockOffset[] ;
    int     keyCount = 0 ;
    boolean changed = false ;

    FileIndex( FileBlock f ) throws Exception
    {
        /*
            Allocate space for the index buckets
        */

        blockOffset = new long[INDEX_BUCKETS] ;

        fbk = f ;

        try
        {
            byte buf[] = fbk.readBlock( 0 ) ;

            DataInputStream in = new DataInputStream(new ByteArrayInputStream(buf)) ;

            if (in.readInt() != KEY_LENGTH)
            {
                throw new FilePersistException("Bad index key length") ;
            }

            blockCount = in.readInt() ;
            keyCount = in.readInt() ;

            int blocksUsed = 0 ;

            for (int i = 0 ; i < blockCount ; i++)
            {
                blockOffset[i] = in.readLong() ;
                if (blockOffset[i] != 0)
                {
                    blocksUsed++ ;
                }
            }

            // System.out.println("Key count = " + keyCount + "  Blocks used = " + blocksUsed ) ;

        } catch (java.io.EOFException ex)
        {
            blockCount = INDEX_BUCKETS ;

            writeIndex() ;
        }
    }

    void close() throws Exception
    {
        if (changed)
        {
            writeIndex() ;
        }
    }

    /**
        Writes the index control structure
    */
    void writeIndex() throws Exception
    {
        ByteArrayOutputStream outb = new ByteArrayOutputStream() ;
        DataOutputStream out = new DataOutputStream(outb) ;

        out.writeInt( KEY_LENGTH ) ;
        out.writeInt( blockCount ) ;
        out.writeInt( keyCount ) ;

        for (int i = 0 ; i < blockCount ; i++ )
            out.writeLong( blockOffset[i] ) ;

        try
        {
            fbk.updateBlock( 0, outb.toByteArray(), outb.size()) ;
        } catch (java.io.EOFException ex)
        {
            long offset = fbk.insertBlock( outb.toByteArray(), outb.size()) ;

            if (offset != 0)
            {
                throw new FilePersistException("Index block cannot be created at 0") ;
            }
        }

        changed = false ;
    }

    /**
        Hash the key value and return a block number
     */
    private int hashKey( ObjectIdentifier oid )
    {
        long    v = oid.getOid() ;
        int     hash = 0 ;

        hash = hash + (int)(v >>> 48) & 0xFFFF ;
        hash = hash + (int)(v >>> 32) & 0xFFFF ;
        hash = hash + (int)(v >>> 16) & 0xFFFF ;
        hash = hash + (int)(v >>> 0) & 0xFFFF ;

        hash = hash % blockCount ;

        // System.out.println("HASH = " + hash ) ;

        if (hash < 0)
        {
            return hash * -1 ;
        }
        else
        {
            return hash ;
        }
    }

    void insertEntry( ObjectIdentifier oid, long offset ) throws Exception
    {
        int block = hashKey( oid ) ;

        IndexBlock blk = new IndexBlock( this, block ) ;

        blk.insertEntry( oid, offset ) ;

        keyCount++ ;

        changed = true ;
    }

    void updateEntry( ObjectIdentifier oid, long newOffset ) throws Exception
    {
        int block = hashKey( oid ) ;

        IndexBlock blk = new IndexBlock( this, block ) ;

        blk.updateEntry( oid, newOffset ) ;
    }

    long findEntry( ObjectIdentifier oid ) throws Exception
    {
        if (oid.getOid() == 0)
        {
            throw new KeyNotFoundException() ;
        }

        int block = hashKey( oid ) ;

        if (blockOffset[block] == 0)
        {
            throw new KeyNotFoundException() ;
        }

        IndexBlock blk = new IndexBlock( this, block ) ;

        return blk.findEntry( oid ) ;
    }
}

class IndexBlock
{
    int                     block ;
    FileIndex               fidx ;

    IndexBlock( FileIndex f, int b) throws Exception
    {
        block = b ;
        fidx = f ;
    }

    void insertEntry( ObjectIdentifier oid, long offset ) throws Exception
    {
        int keyCount = 0 ;                  // keys used
        int keyAlloc = fidx.KEY_ALLOC ;     // space available

        ByteArrayOutputStream outb = new ByteArrayOutputStream() ;
        DataOutputStream out = new DataOutputStream(outb) ;
        DataInputStream in = null ;

        if (fidx.blockOffset[block] != 0)
        {
            /*
                No index block exists. Create an empty index
                block.
            */

            byte inbuf[] = fidx.fbk.readBlock( fidx.blockOffset[block] ) ;

            in = new DataInputStream(new ByteArrayInputStream(inbuf)) ;

            keyAlloc = in.readInt() ;
            keyCount = in.readInt() ;
        }

        if (keyCount == keyAlloc)
        {
            keyAlloc = keyAlloc + fidx.KEY_EXTEND ;
        }

        out.writeInt( keyAlloc ) ;
        out.writeInt( keyCount + 1 ) ;

        while (keyCount > 0)
        {
            out.writeLong(in.readLong()) ;  // key
            out.writeLong(in.readLong()) ;  // offset

            keyCount-- ;
            keyAlloc-- ;
        }

        out.writeLong(oid.getOid()) ;
        out.writeLong(offset) ;

        keyAlloc-- ;

        while (keyAlloc > 0)
        {
            out.writeLong(0) ;
            out.writeLong(0) ;

            keyAlloc-- ;
        }


        byte outbuf[] = outb.toByteArray() ;
        int outsiz = outb.size() ;

        if (fidx.blockOffset[block] != 0)
        {
            try
            {
                fidx.fbk.updateBlock(fidx.blockOffset[block], outbuf, outsiz) ;
            } catch (FileBlockTooShortException ex)
            {
                long newOffset = fidx.fbk.insertBlock( outbuf, outsiz) ;

                long oldOffset = fidx.blockOffset[block] ;

                fidx.blockOffset[block] = newOffset ;

                fidx.writeIndex() ;

                fidx.fbk.freeBlock( oldOffset ) ;
            }
        }
        else
        {
            long newOffset = fidx.fbk.insertBlock( outbuf, outsiz) ;

            fidx.blockOffset[block] = newOffset ;

            fidx.writeIndex() ;
        }
    }

    void updateEntry( ObjectIdentifier oid, long newOffset ) throws Exception
    {
        int keyCount = 0 ;                  // keys used
        int keyAlloc = fidx.KEY_ALLOC ;     // space available

        ByteArrayOutputStream outb = new ByteArrayOutputStream() ;
        DataOutputStream out = new DataOutputStream(outb) ;

        byte inbuf[] = fidx.fbk.readBlock( fidx.blockOffset[block] ) ;

        DataInputStream in = new DataInputStream(new ByteArrayInputStream(inbuf)) ;

        if (fidx.blockOffset[block] == 0)
        {
            throw new FilePersistException("No index block to update") ;
        }

        keyAlloc = in.readInt() ;
        keyCount = in.readInt() ;

        out.writeInt(keyAlloc) ;
        out.writeInt(keyCount) ;

        long    xid = oid.getOid() ;

        while (keyAlloc-- > 0)
        {
            long id = in.readLong() ;

            out.writeLong(id) ;

            if (id == xid)
            {
                in.readLong() ;
                out.writeLong( newOffset ) ;
            }
            else
            {
                out.writeLong( in.readLong()) ;
            }
        }

        byte outbuf[] = outb.toByteArray() ;
        int outsiz = outb.size() ;

        fidx.fbk.updateBlock(fidx.blockOffset[block], outbuf, outsiz) ;
    }

    long findEntry( ObjectIdentifier oid ) throws Exception
    {
        byte buf[] = fidx.fbk.readBlock( fidx.blockOffset[block] ) ;

        DataInputStream in = new DataInputStream(new ByteArrayInputStream(buf)) ;

        in.readInt() ;                  // key allocation

        int keyCount = in.readInt() ;   // key values

        long    xid = oid.getOid() ;

        while (keyCount-- > 0)
        {
            if (in.readLong() == xid)
            {
                return in.readLong() ;
            }

            in.readLong() ;
        }

        throw new KeyNotFoundException() ;
    }
}
