/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.file ;

import  JOP.persist.* ;

import  java.util.Vector ;
import  java.util.Date ;
import  java.util.Enumeration ;
import  java.io.RandomAccessFile ;
import  java.io.DataOutputStream ;
import  java.io.ByteArrayOutputStream ;
import  java.io.DataOutput ;
import  java.io.DataInputStream ;
import  java.io.ByteArrayInputStream ;
import  java.io.DataInput ;

/**
    The class saves and read objects to a random access file.
    <p>
    *@see JOP.persist.PersistManager
*/

public class FilePersistManager extends PersistManager
{
    private boolean                 open = false ;
    private boolean                 closed = false ;
    private String                  filename = null ;
    private FileInputController     inCntl ;
    private FileOutputController    outCntl ;
    private RandomAccessFile        fid ;
    private FileBlock               fbk ;
    FileIndex                       fdx ;

    /**
        Initialise a FilePeristManager with the name of
        a RandomAccessFile.
    */

    public FilePersistManager( String f )
    {
        filename = f ;
    }


    /**
        Begin using the perist manager
        <p>
        *@see JOP.persist.PersistManager
    */

    public void open() throws Exception
    {
        if (open)
        {
            throw new FilePersistException("File already open") ;
        }

        fid = new RandomAccessFile( filename, "rw" ) ;

        fbk = new FileBlock( fid ) ;

        fdx = new FileIndex( fbk ) ;

        inCntl = new FileInputController( this, fbk ) ;
        outCntl = new FileOutputController( this, fbk ) ;

        open = true ;
    }

    /**
        End use of the persist manager
        <p>
        *@see JOP.persist.PersistManager
    */

    public void close() throws Exception
    {
        if (closed)
        {
            throw new FilePersistException("Already closed") ;
        }

        if (!open)
        {
            throw new FilePersistException("Not open") ;
        }

        fdx.close() ;

        closed = true ;
    }


    /**
        Save the object in the Random Access File. The object must either
        implement Persistable, have a custom marshall
        class available or it will be marshalled by native methods. The
        object and all of the objects
        referenced by the object are saved. Loops are detected and
        avoided.
        <p>
        To retrieve the object call 'streamInObject()'
        <p>
        *@see   JOP.persist.PersistManager
        *@see   JOP.persist.Persistable
        *@see   JOP.persistmarshall.PersistMarshall
        *@see   #streamInObject

        *@param     o   The object to be saved
        *@returns       The unique identifier of the object
        *@throws    java.lang.Exception
    */

    public long streamOutObject(Object o) throws Exception
    {
        if (!open)
        {
            throw new FilePersistException("Not open") ;
        }

        Vector  history = new Vector() ;

        ObjectIdentifier oid = streamOutObject( o, history, outCntl, true ) ;

        // System.out.println("Object saved as: " + oid ) ;

        return oid.getOid() ;
    }

    /**
        Retrieves an object previously saved by 'streamOutObject()'.
        The unique object id is passed and an object and all
        of the objects referenced by the object are
        returned. The object will be an instance of the same type
        that was saved and will have the same context. The
        caller of this method must explicitly
        cast the returned object before using it.
        <p>
        Within a consistency unit multiple reads of an object
        either directly (via this method) or indirectly because
        the object is referenced by another object will not
        create secondary instances.
        <p>
        *@see   JOP.persist.PersistManager

        *@param     id  Unique indentifier of the object
        *@returns       The object
    */

    public Object streamInObject(long xid) throws Exception
    {
        if (!open)
        {
            throw new FilePersistException("Not open") ;
        }

        ObjectIdentifier oid = streamInObject( xid, inCntl ) ;

        return oid.getObject() ;
    }
}

/**
    Controller to write objects to a file stream

    The format of the random access file is as follows:
*/

class FileOutputController implements JOP.persist.OutputController
{
    private FileBlock           outf ;
    private FilePersistManager  pmgr ;

    FileOutputController( FilePersistManager p, FileBlock f )
    {
        outf = f ;
        pmgr = p ;
    }

    /**
        Update an existing object. This method writes the
        object's slots to the random access file and
        updates the index reference.
        <p>
        *@param oid         the object identifier
        *@param defObject   an interface to access the objects
                            meta-data and slot values
        *@param explicit    'true' is this was the object
                            the user explicitly called streamOutObject,
                            otherwise 'false' indicates an internal
                            reference.
        *@see FileOutputController
    */
    public void updateObject( ObjectIdentifier oid, ObjectDefinition defObject, boolean explicit ) throws Exception
    {
        /**
            Get the current offset from the index
        */
        long offset = pmgr.fdx.findEntry( oid ) ;

        /**
            Create an in-memory buffer as a data output stream.
            This is used to write the object, then at the end
            the block is written to the file.
        */
        ByteArrayOutputStream outb = new ByteArrayOutputStream() ;
        DataOutputStream outd = new DataOutputStream(outb) ;

        writeObjectBuffer( outd, oid, defObject ) ;

        byte buf[] = outb.toByteArray() ;

        try
        {
            outf.updateBlock(offset, buf, outb.size()) ;
        } catch (FileBlockTooShortException ex)
        {
            long newOffset = outf.insertBlock( buf, outb.size()) ;

            pmgr.fdx.updateEntry( oid, newOffset ) ;

            outf.freeBlock( offset ) ;
        }
    }

    /**
        Adds a new object. This method writes the
        object's slots to the random access file and
        updates the index reference.
        <p>
        *@param oid         the object identifier
        *@param defObject   an interface to access the objects
                            meta-data and slot values
        *@param explicit    'true' is this was the object
                            the user explicitly called streamOutObject,
                            otherwise 'false' indicates an internal
                            reference.
        *@see FileOutputController
    */
    public void insertObject( ObjectIdentifier oid, ObjectDefinition defObject, boolean explicit ) throws Exception
    {
        /*
            Create an in-memory buffer as a data output stream.
            This is used to write the object, then at the end
            the block is written to the file.
        */
        ByteArrayOutputStream outb = new ByteArrayOutputStream() ;
        DataOutputStream outd = new DataOutputStream(outb) ;

        writeObjectBuffer( outd, oid, defObject ) ;

        byte buf[] = outb.toByteArray() ;

        long offset = outf.insertBlock(buf, outb.size()) ;

        // System.out.println("Buffer created, length = " + outb.size() + " at " + offset) ;

        pmgr.fdx.insertEntry( oid, offset ) ;

        // if (pmgr.fdx.findEntry( oid ) != offset)
        //    throw new Exception("Cannot re-find key") ;
    }


    /**
        Writes an object to a data output.
        <p>
        *@param outd        The output stream
        *@param oid         The object identifier
        *@param defObject   An interface to define meta-data and
                            slot values
        *@throws Exception  On any error
    */
    private void writeObjectBuffer(DataOutput outd, ObjectIdentifier oid, ObjectDefinition defObject) throws Exception
    {
        // System.out.println() ;
        // System.out.println("OUTPUT: OBJECT = " + defObject.getClassName() + " " + oid) ;

        outd.writeLong(oid.getOid()) ;
        outd.writeUTF(defObject.getClassName()) ;

        for (Enumeration eClass = defObject.classes() ; eClass.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)eClass.nextElement() ;

            // System.out.println("OUTPUT: CLASS = " + defClass.getClassName()) ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                int signature = defSlot.getSignatureType() ;

                if (defSlot.isArray())
                {
                    // System.out.println("An array with " + defSlot.sizeArray()) ;

                    if (defSlot.isNullArray())
                    {
                        outd.writeBoolean(false) ;
                    }
                    else
                    {
                        outd.writeBoolean(true) ;
                        outd.writeInt(defSlot.sizeArray()) ;

                        for (int n = 0 ; n < defSlot.sizeArray() ; n++)
                        {
                            // System.out.println("Write element " + n ) ;

                            if (signature == JavaTypes.TYPE_OBJECT)
                            {
                                writeSlot( outd, signature, defSlot.getRefOid(n)) ;
                            }
                            else
                            {
                                writeSlot( outd, signature, defSlot.getValue(n)) ;
                            }
                        }
                    }
                }
                else
                {
                    if (signature == JavaTypes.TYPE_OBJECT)
                    {
                        writeSlot( outd, signature, defSlot.getRefOid()) ;
                    }
                    else
                    {
                        writeSlot( outd, signature, defSlot.getValue()) ;
                    }
                }
            }
        }
    }

    /**
        Append a slot value to the output buffer
    */
    private void writeSlot( DataOutput outd, int signature, Object value ) throws Exception
    {
        if (value == null)
        {
            outd.writeBoolean(false) ;
        }
        else
        {
            outd.writeBoolean(true) ;

            // System.out.println("WRITE (" + signature + ") = " + value + " " + value.getClass().getName()) ;

            switch (signature)
            {
                case JOP.persist.JavaTypes.TYPE_CHARACTER :
                    outd.writeChar(((Character)value).charValue()) ;
                    break;
                case JOP.persist.JavaTypes.TYPE_BYTE :
                    outd.writeByte((byte)((Integer)value).intValue()) ;
                    break;
                case JOP.persist.JavaTypes.TYPE_FLOAT :
                    outd.writeFloat(((Float)value).floatValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_DOUBLE :
                    outd.writeDouble(((Double)value).doubleValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_INTEGER :
                    outd.writeInt(((Integer)value).intValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_SHORT :
                    outd.writeShort((short)((Integer)value).intValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_LONG :
                    outd.writeLong(((Long)value).longValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_BOOLEAN :
                    outd.writeBoolean(((Boolean)value).booleanValue()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_STRING :
                    outd.writeUTF((String)value) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_DATE :
                    outd.writeLong(((Date)value).getTime()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_OBJECT :
                    outd.writeLong(((ObjectIdentifier)value).getOid()) ;
                    break ;
                case JOP.persist.JavaTypes.TYPE_UNKNOWN :
                default :
                    throw new FilePersistException("Unsupported '" + signature + "'") ;
            }
        }
    }
}

/**
    INPUT CONTROLLER
*/

class FileInputController implements JOP.persist.InputController
{
    private FileBlock           inf ;
    private FilePersistManager  pmgr ;
    private DataInputStream     cache_ind ;
    private ObjectIdentifier    cache_oid ;

    FileInputController( FilePersistManager p, FileBlock f )
    {
        inf = f ;
        pmgr = p ;
    }

    public String readClassname( ObjectIdentifier oid ) throws Exception
    {
        long    offset ;

        try
        {
            offset = pmgr.fdx.findEntry( oid ) ;
        }
        catch (KeyNotFoundException ex)
        {
            throw new JOP.persist.ObjectNotFoundException(oid.toString()) ;
        }

        byte inbuf[] = inf.readBlock( offset ) ;

        ByteArrayInputStream inb = new ByteArrayInputStream(inbuf) ;
        DataInputStream ind = new DataInputStream( inb ) ;

        if (ind.readLong() != oid.getOid())
        {
            throw new FilePersistException("Corrupt OID at " + offset ) ;
        }

        String classname = ind.readUTF() ;

        cache_ind = ind ;
        cache_oid = oid ;

        return classname ;
    }

    public void readObject( ObjectIdentifier oid, ObjectDefinition defObject ) throws Exception
    {
        DataInputStream ind ;

        if (!oid.equals(cache_oid))
        {
            long    offset ;

            try
            {
                offset = pmgr.fdx.findEntry( oid ) ;
            }
            catch (KeyNotFoundException ex)
            {
                throw new JOP.persist.ObjectNotFoundException(oid.toString()) ;
            }

            byte inbuf[] = inf.readBlock( offset ) ;

            ByteArrayInputStream inb = new ByteArrayInputStream(inbuf) ;
            ind = new DataInputStream( inb ) ;
        }
        else
        {
            ind = cache_ind ;

            ind.reset() ;
        }

        readObjectBuffer( ind, oid, defObject ) ;
    }

    private void readObjectBuffer(DataInput ind, ObjectIdentifier oid, ObjectDefinition defObject) throws Exception
    {
        // System.out.println() ;
        // System.out.println("OUTPUT: OBJECT = " + defObject.getClassName() + " " + oid) ;

        if (ind.readLong() != oid.getOid())
        {
            throw new FilePersistException("Corrupt Oid") ;
        }

        ind.readUTF() ;

        for (Enumeration eClass = defObject.classes() ; eClass.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)eClass.nextElement() ;

            // System.out.println("OUTPUT: CLASS = " + defClass.getClassName()) ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                int signature = defSlot.getSignatureType() ;

                if (defSlot.isArray())
                {
                    boolean okArray = ind.readBoolean() ;

                    if (!okArray)
                    {
                        defSlot.setNullArray() ;
                    }
                    else
                    {
                        int count = ind.readInt() ;

                        for (int n = 0 ; n < count ; n++)
                        {
                            // System.out.println("Read element " + n ) ;

                            if (signature == JavaTypes.TYPE_OBJECT)
                            {
                                defSlot.setRefOid(n, (ObjectIdentifier)readSlot( ind, signature)) ;
                            }
                            else
                            {
                                defSlot.setValue(n, readSlot( ind, signature)) ;
                            }
                        }
                    }
                }
                else
                {
                    if (signature == JavaTypes.TYPE_OBJECT)
                    {
                        defSlot.setRefOid( (ObjectIdentifier)readSlot( ind, signature)) ;
                    }
                    else
                    {
                        defSlot.setValue( readSlot( ind, signature)) ;
                    }
                }
            }
        }
    }

    /**
        Append a slot value to the output buffer
    */
    private Object readSlot( DataInput ind, int signature ) throws Exception
    {
        Object  value ;

        if (ind.readBoolean() == false)
        {
            // System.out.println("readSlot = NULL") ;
            return null ;
        }

        switch (signature)
        {
            case JOP.persist.JavaTypes.TYPE_CHARACTER :
                value = new Character(ind.readChar()) ;
                break;
            case JOP.persist.JavaTypes.TYPE_BYTE :
                value = new Integer(ind.readByte()) ;
                break;
            case JOP.persist.JavaTypes.TYPE_FLOAT :
                value = new Float(ind.readFloat()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_DOUBLE :
                value = new Double(ind.readDouble()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_INTEGER :
                value = new Integer(ind.readInt()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_SHORT :
                value = new Integer((short)ind.readShort()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_LONG :
                value = new Long(ind.readLong()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_BOOLEAN :
                value = new Boolean(ind.readBoolean()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_STRING :
                value = ind.readUTF() ;
                break ;
            case JOP.persist.JavaTypes.TYPE_DATE :
                value = new Date(ind.readLong()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_OBJECT :
                value = new ObjectIdentifier(ind.readLong()) ;
                break ;
            case JOP.persist.JavaTypes.TYPE_UNKNOWN :
            default :
                throw new FilePersistException("Unsupported '" + signature + "'") ;
        }

        // System.out.println("READ (" + signature + ") = " + value ) ;

        return value ;
    }
}

