/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.jdbc ;

import java.sql.* ;

import  JOP.persist.Hashcode ;

/**
    This class provides utility functions to assign table and column
    names that are legal for the current database.
    <p>
    Identifier names are truncated to the correct length, special
    characters are removed and a hashcode is added to truncated names
    to ensure names remain unique.
*/
public class DatabaseTypes
{
    private int     maxColumnNameLength ;
    private int     maxTableNameLength ;

    DatabaseTypes( Connection dbc ) throws SQLException
    {
        DatabaseMetaData dma = dbc.getMetaData();

        maxColumnNameLength = dma.getMaxColumnNameLength() ;
        maxTableNameLength = dma.getMaxTableNameLength() ;
    }

    /**
        Return a valid identifier for this database from
        the given identifier. A valid identifier contains
        legal characters and is less than the allowable
        identifier length.
        <p>
        Identifiers are not reserved words for this database.
    */
    private String validIdentifier( String prefix, String identifier, int maxlen )
    {
        String  result = identifier ;

        result = result.trim().replace('.','_') ;

        result = result.trim().replace('/','_') ;

        if (result.length() < (maxlen - (prefix.length() + 1)))
        {
            return prefix + result ;       // a JOP name
        }
        else
        {
            //  we need to truncate the name, so we will append the hashcode
            //  to the end of the truncated name to keep it unique...

            Hashcode    hc = new Hashcode() ;   // append this to the end if we need

            hc.calc( identifier.trim().replace('/','.')) ;

            int hashVal = hc.getHashcode() ;

            if (hashVal < 0)
            {
                hashVal = -hashVal ;
            }

            String hash = Integer.toString(hashVal, 16) ;

            maxlen = maxlen - (hash.length() + prefix.length() + 2) ; // _ and one spare

            while (result.length() > maxlen && result.indexOf('_') != -1)
            {
                result = result.substring(result.indexOf('_')+1) ;
            }

            if (result.length() > maxlen)
            {
                result = result.substring(result.length()-(maxlen-1)) ;
            }

            return prefix + result + "_" + hash ;
        }
    }

    /**
        Return the name of a table based upon the name of the
        class. The table name starts with 'C_'.
        <p>
        The name is truncated to the maximum allowable length as
        returned by the JDBC driver. If the identifier is truncated
        then a hashcode value is calculated and appended to the name.
        The hashcode ensures that truncated names remain unique.
    */
    String tableName( String className )
    {
        return validIdentifier("C_", className, maxTableNameLength) ;
    }

    /**
        Return the table name for storing the elements of the array
        for the given slot of the class. The array name will start with
        'A_'.
        <p>
        The name is truncated to the maximum allowable length as
        returned by the JDBC driver. If the identifier is truncated
        then a hashcode value is calculated and appended to the name.
        The hashcode ensures that truncated names remain unique.
    */
    String arrayName( String className, String slotName )
    {
        return validIdentifier( "A_", className + "_" + slotName, maxTableNameLength ) ;
    }

    /**
        Return the column name based upon the name of a slot. The slot
        name starts with 'S_'
        <p>
        The name is truncated to the maximum allowable length as
        returned by the JDBC driver. If the identifier is truncated
        then a hashcode value is calculated and appended to the name.
        The hashcode ensures that truncated names remain unique.
    */
    String columnName( String slotName )
    {
        return validIdentifier("S_", slotName, maxColumnNameLength) ;
    }
}
