/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.jdbc ;

import  java.util.Vector;
import  java.util.Hashtable;
import  java.util.Enumeration;
import  java.sql.* ;

import  JOP.persist.JavaTypes ;
import  JOP.persist.ObjectIdentifier ;

import  java.io.PrintStream ;

/**
    The JDBC configuration utility works by testing a number
    of 'strategies' for storing and retrieving each Java type. A
    strategy defines the SQL column data type and the JDBC methods
    to be used to access that column. Testing continues until a
    successful strategy is found or until all
    strategies for the Java data type have been attempted. A successful strategy
    is one that will correctly store and retrieve the pre-defined test
    data for each Java data type. Typically the test data consists of
    the minimum, maximum, zero and null values for each Java type.
    </p><p>
    Once a successful strategy has been selected for all Java types the
    chosen strategies are saved in the JDBC database. Each time a JDBC
    PersistManager is opened the correct strategies are immediately loaded
    into memory.
    </p><p>
    When another JDBC driver is used against the same data base it is first
    necessary to run the configuration program to determine if the new
    driver is compatible with the saved strategies.
    </p><p>
    Each strategy is a dynamically loaded class that includes methods for
    reading and writing to JDBC. The methods are necessary to allow character
    escaping and other manipulations of the saved or retrieved data. It is
    possible for a JOP user to define their own strategies.
    </p>
    <p>
    To run the JDBC configuration program execute the following command:
    <pre>
    <tt>&gt;</tt><strong><tt>java JOP.persist.jdbc.JdbcConfig <i>driver_name</i> <i>url</i> <i>username</i> <i>password</i></tt></strong>
    </pre>
    This command will configure your database and store the configuration
    information in the database in the tables <b>JOPconfig</b> and <b>JOPtypes</b>.
    See the section on <a href="http://www.magna.com.au/~davidr/test.htm">testing</a> for information on
    how to setup JOP and JDBC.
*/
public class JdbcConfig
{
    JdbcConfig()
    {
    }

    /**
        Command line program entry point....
    */
    public static void main(String args[]) throws Exception
    {
        if (args.length != 4)
        {
            System.err.println("Usage: java JOP.persist.jdbc.JdbcConfig driver_name url username password") ;
        }
        else
        {
            String driver = args[0] ; // "jdbc.odbc.JdbcOdbcDriver" ;
            String url = args[1] ;
            String user = args[2] ;
            String password = args[3] ;

            JdbcConfig config = new JdbcConfig() ;

            Connection dbc = config.openDatabase(driver, url, user, password ) ;

            try
            {
                config.configDatabase( dbc, System.out ) ;

            } catch (JOP.persist.PersistException ex)
            {
                System.out.println("\nRESULT: " + ex.getMessage() + "\n") ;
            }
        }
    }

    /**
        Configure the type support for a JDBC database. Type configuration
        informatino is stored in the tables JOPconfig and JOPtypes. Each
        combination of driver and dbms type requires configuration.
        <p>
        The method is called either from JOP or from the standalone utility
        JdbcConfig.
    */
    void configDatabase( Connection dbc, PrintStream out ) throws Exception
    {
       if (isConfigured(dbc))
       {
            if (knownDriver(dbc))
            {
                if (out != null)
                {
                    out.println("\n*** Existing Configuration ***\n") ;

                    out.println("This database has already been configured for use by JOP with" ) ;
                    out.println("the current driver.") ;
                }
            }
            else if (testCompatibility( dbc ))
            {
                if (out != null)
                {
                    out.println("\n*** Comaptible Configuration exists ***\n") ;

                    insertJOPconfig( dbc ) ;

                    out.println("This database has already been configured for use by JOP. The" ) ;
                    out.println("current driver is compatible with the stored configuration ") ;
                    out.println("for this database. It is possible that your current driver is more") ;
                    out.println("capable than the driver that was used to when this database was") ;
                    out.println("configured. If your database contains data then it is not possible") ;
                    out.println("to use these extra capabilities.") ;
                }
            }
            else
            {
                if (out != null)
                {
                    out.println("\n*** Configuration NOT compatible ***\n") ;
                    out.println("This database has already been configured for use by JOP. The" ) ;
                    out.println("current driver is not compatible with the stored configuration ") ;
                    out.println("for this database and cannot be used with JOP. You can remove the") ;
                    out.println("stored configuration by dropping the tables JOPconfig and JOPtypes." ) ;
                    out.println("WARNING: Removing the stored configuration is likely to make some or") ;
                    out.println("all of the data stored in the database inaccessible or incorrect") ;

                    throw new JOP.persist.PersistException("Database re-configuration failed") ;
                }
            }
        }
        else
        {
            String typeHandlers[] = testDatabase( dbc ) ;

            boolean passed = true ;

            for (int i = 1 ; i < typeHandlers.length ; i++)
            {
                if (typeHandlers[i] == null)
                {
                    passed = false ;
                }
            }

            if (passed)
            {
                if (out != null)
                {
                    out.println("\n*** Configuration successful ***\n") ;

                    saveConfigResults( dbc, typeHandlers ) ;

                    out.println("The successful configuration has been stored in your database and will") ;
                    out.println("be used by JOP each time this database is opened. If you intend to") ;
                    out.println("use query tools or other programs to access this database you should") ;
                    out.println("check the SQL column types being used for each Java type. Other JDBC") ;
                    out.println("drivers will probably not work with this database.") ;
                }
            }
            else
            {
                if (out != null)
                {
                    out.println("\n*** Configuration failed ***\n") ;
                    out.println("One of the Java data types could not be stored and retrieved") ;
                    out.println("using the combination of JDBC driver and database manager that" ) ;
                    out.println("you are using. JOP is unable to use this configuration to persist") ;
                    out.println("objects in a reliable manner. It is recommended that you check") ;
                    out.println("you JDBC configuration and download the JDBC driver validation") ;
                    out.println("suite from Sun at www.sun.com") ;

                    throw new JOP.persist.PersistException("Database configuration failed") ;
                }
            }
        }
    }

    /**
        Open database (used when run standalone only)
    */
    private Connection openDatabase(String jdbcDriver, String jdbcUrl, String usrName, String usrPassword ) throws Exception
    {
        try {

            // Load the jdbc-odbc bridge driver

            Connection dbdbc ;
            Class d = Class.forName(jdbcDriver);

            DriverManager.registerDriver( (Driver)d.newInstance()) ;

            //  I get an error with the SUN jdbc-odbc driver
            //  when trying to set the transaction mode. So
            //  for now I will turn it off.

            dbdbc = DriverManager.getConnection( jdbcUrl, usrName, usrPassword);

            checkForWarning(dbdbc.getWarnings());

            dumpMetaData( dbdbc ) ;

            return dbdbc ;
        }
        catch (SQLException ex) {

            // A SQLException was generated.  Catch it and
            // display the error information.  Note that there
            // could be multiple error objects chained
            // together

            System.out.println ("\n*** SQLException caught ***\n");

            SQLException e = ex ;

            while (e != null)
            {
                System.out.println("SQLState: " + e.getSQLState ());
                System.out.println("Message:  " + e.getMessage ());
                System.out.println("Vendor:   " + e.getErrorCode ());

                e = e.getNextException();

                System.out.println("");
            }

            throw ex ;
        }
    }

    /**
        Used when run stand-alone only
    */
    private void dumpMetaData(Connection dbdbc) throws SQLException
    {
        // Get the DatabaseMetaData object and display
        // some information about the connection

        DatabaseMetaData dma = dbdbc.getMetaData();

        System.out.println("Connected to :" + dma.getURL());
        System.out.println("User name    :" + dma.getUserName()) ;
        System.out.println("Driver       :" + dma.getDriverName());
        System.out.println(" version     :" + dma.getDriverVersion());
        System.out.println("Database     :" + dma.getDatabaseProductName());
        System.out.println(" version     :" + dma.getDatabaseProductVersion());
        System.out.println("");

        System.out.println("Max column len :" + dma.getMaxColumnNameLength()) ;
        System.out.println("Max table len :" + dma.getMaxTableNameLength()) ;

    }

    /**
        Check and display database warnings. Source code based
        on the SimpleSelect.java distributed with JDBC. Used when run
        standalong only.
    */
    private static boolean checkForWarning(SQLWarning warn) throws SQLException
    {
        boolean rc = false;

        // If a SQLWarning object was given, display the
        // warning messages.  Note that there could be
        // multiple warnings chained together

        if (warn != null) {
            System.out.println ("\n *** Warning ***\n");
            rc = true;
            while (warn != null) {
                System.out.println ("SQLState: " +
                    warn.getSQLState ());
                System.out.println ("Message:  " +
                    warn.getMessage ());
                System.out.println ("Vendor:   " +
                    warn.getErrorCode ());
                System.out.println ("");
                warn = warn.getNextWarning ();
            }
        }
        return rc;
    }

    /**
        Returns 'true' if the DBMS is already configured, ie it has a JOPconfig
        table.
    */
    private boolean isConfigured( Connection dbc ) throws SQLException
    {
        Statement   stmt = null ;
        boolean     jopConfigExists = false ;
        int         count = 0 ;

        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select count(*) from JOPconfig" ) ;

            if (rs.next())
            {
                count = rs.getInt(1) ;
            }

            if (count > 0)
            {
                jopConfigExists = true ;
            }

            stmt.close() ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }
        }

        return jopConfigExists ;
    }

    /**
        Saves the result of the test into the database
        in JOPconfig and JOPtypes. These tables are read
        when JOP starts to pre-configure the type handlers
        for this database.
    */
    private void saveConfigResults( Connection dbc, String tcls[] ) throws Exception
    {
        Statement   stmt = null ;
        boolean     jopConfigExists = false ;
        boolean     jopTypesExists = false ;

        //  detect if the config table exists (we don't trust JDBC meta data yet)

        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select count(*) from JOPconfig" ) ;

            rs.next() ;

            jopConfigExists = true ;

            stmt.close() ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }
        }

        //  detect if the types table exists (we don't trust JDBC meta data yet)
        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select count(*) from JOPtypes" ) ;

            rs.next() ;

            jopTypesExists = true ;

            stmt.close() ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }
        }

        if (!jopConfigExists)
        {
            //  create the config table

            try
            {
                stmt = dbc.createStatement() ;

                String s1 = "create table JOPconfig (driver char(64), dbms char(64) ) " ;

                stmt.executeUpdate(s1) ;

                stmt.close() ;
            }
            catch (SQLException ex)
            {
                stmt.close() ;

                throw ex ;
            }
        }


        if (!jopTypesExists)
        {
            //  create the types table

            try
            {
                stmt = dbc.createStatement() ;

                String s1 = "create table JOPtypes (joptype integer, javatype char(64), typehandler char(64) ) " ;

                stmt.executeUpdate(s1) ;

                stmt.close() ;
            }
            catch (SQLException ex)
            {
                stmt.close() ;

                throw ex ;
            }
        }

        //  save the configuration information for this driver

        insertJOPconfig( dbc ) ;

        //  and the successful type handlers

        for (int i = 1 ; i < JavaTypes.TypeName.length ; i++)
        {
            insertJOPtype( dbc, i, JavaTypes.TypeName[i], tcls[i] );
        }
    }

    /**
        Save the name of the current driver and dbms and the version in
        the JOPconfig table
    */
    private void insertJOPconfig( Connection dbc ) throws SQLException
    {
        DatabaseMetaData dma = dbc.getMetaData();

        String  driver = dma.getDriverName().trim() + " " + dma.getDriverVersion().trim() ;
        String  dbms = dma.getDatabaseProductName().trim() + " " + dma.getDatabaseProductVersion().trim() ;

        PreparedStatement stmt = null ;

        try
        {
            stmt = dbc.prepareStatement("insert into JOPconfig (driver, dbms) values (?,?)") ;

            stmt.setString(1, driver) ;
            stmt.setString(2, dbms) ;

            stmt.executeUpdate() ;

            stmt.close() ;
        } catch (SQLException ex)
        {
            stmt.close() ;

            throw ex ;
        }
    }

    /**
        Save the type handler for a specific Java type into the JOPtypes table
    */
    private void insertJOPtype( Connection dbc, int jopType, String javaType, String typeHandler ) throws SQLException
    {
        PreparedStatement stmt = null ;

        try
        {
            stmt = dbc.prepareStatement("insert into JOPtypes (joptype, javatype, typehandler) values (?,?,?)") ;

            stmt.setInt( 1, jopType ) ;
            stmt.setString(2, javaType) ;
            stmt.setString(3, typeHandler) ;

            stmt.executeUpdate() ;

            stmt.close() ;
        } catch (SQLException ex)
        {
            stmt.close() ;

            throw ex ;
        }
    }

    /**
        Return 'true' if the current driver one of the drivers that has
        already been configured for this database.
    */
    private boolean knownDriver( Connection dbc ) throws Exception
    {
        DatabaseMetaData dma = dbc.getMetaData();
        String  driver = dma.getDriverName().trim() + " " + dma.getDriverVersion().trim() ;
        String  dbms = dma.getDatabaseProductName().trim() + " " + dma.getDatabaseProductVersion().trim() ;
        Statement   stmt = null ;
        int         count = 0 ;

        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select count(*) from JOPconfig where (driver='" + driver + "' and dbms='" + dbms + "');") ;

            if (rs.next())
            {
                count = rs.getInt(1) ;
            }

            stmt.close() ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }

            throw j ;
        }

        if (count == 1)
        {
            return true ;
        }
        else
        {
            return false ;
        }
    }

    /**
        Return 'true' if the current driver is compatible with the existing
        configuration or the current driver exists in the existing configuration.
    */
    private boolean testCompatibility( Connection dbc ) throws Exception
    {
        Statement   stmt = null ;

        TypeHandlerManager  typhdl = new TypeHandlerManager( dbc, false ) ;

        Object[][] testData = getTestData() ;

        for (int i = 1 ; i < JavaTypes.TypeName.length ; i++)
        {
            System.out.println() ;
            System.out.println(">>> " + JavaTypes.TypeName[i]) ;

            String  tablename = "JOP" + i ;

            boolean tableCreated = false ;

            TypeHandler th = typhdl.getTypeHandler(i) ;

            try
            {
                createTable(dbc, tablename, th.getDefinition(), "var") ;

                tableCreated = true ;

                for (int z = 0 ; z < testData[i].length ; z++)
                {
                    insertRow( dbc, tablename, th, z, testData[i][z] ) ;

                    Object obj = readRow( dbc, tablename, th, z ) ;

                    if (testData[i][z] == null)
                    {
                        if (obj != null)
                        {
                            throw new Exception("Null value failed") ;
                        }
                    }
                    else if (!testData[i][z].equals(obj))
                    {
                        if (i == JavaTypes.TYPE_STRING)
                        {
                            String s1 = testData[i][z].toString() ;
                            String s2 = obj.toString() ;
                            char c1[] = s1.toCharArray() ;
                            char c2[] = s2.toCharArray() ;

                            System.out.println("len c1 = " + c1.length + " c2 = " + c2.length ) ;

                            for (int ew = 0 ; ew < c1.length ; ew++)
                            {
                                if (c1[ew] != c2[ew])
                                {
                                    System.out.println("DIFF [" + ew + "] c1 = " + (int)c1[ew] + " c2 = " + (int)c2[ew]) ;
                                }
                            }
                        }

                        throw new Exception("Unequal return value '" + obj + "' expected '" + testData[i][z] + "'") ;
                    }
                }

                dropTable(dbc, tablename) ;
            } catch (Exception ex)
            {
                System.out.println("*** FAILED ***    " + ex.getMessage()) ;

                return false ;
            }

            System.out.println("*** PASSED ***") ;
        }

        return true ;
    }

    /**
        A method to provide test data.
    */
    private Object[][] getTestData()
    {
        //  setup string test data (make sure it is less than 255 after escaping
        //  otherwise we won't be able to store it in some driver configs. JOP allows
        //  255 character strings on all JDBC drivers, longer where the driver supports
        //  it.

        char lowChars[] = new char[246] ;

        for (int i = 0 ; i < 246 ; i++)
        {
            if (i < 64)
            {
                lowChars[i] = (char)i ;
            }
            else
            {
                lowChars[i] = (char)(i + 10) ;
            }
        }

        Object testData[][] =
            {
            // TYPE_UNKNOWN
                { null },
            // TYPE_OBJECT
                { new ObjectIdentifier(0),
                  new ObjectIdentifier(ObjectIdentifier.oid("ffff-ffff-ffff-ffff")),
                  new ObjectIdentifier(ObjectIdentifier.oid("0000-0000-0000-0000")),
                  null },
            // TYPE_SHORT
                { new Integer(-32767),
                  new Integer(32768),
                  new Integer(0),
                  null },
            // TYPE_LONG
                { new Long(Long.MIN_VALUE),
                  new Long(Long.MAX_VALUE),
                  new Long(0),
                  null },
            // TYPE_INTEGER
                { new Integer(Integer.MIN_VALUE),
                  new Integer(Integer.MAX_VALUE),
                  new Integer(0),
                  null },
            // TYPE_BYTE
                { new Integer(0),
                  new Integer(255),
                  null },
            // TYPE_CHARACTER
                { new Character(Character.MIN_VALUE),
                  new Character(Character.MAX_VALUE),
                  null },
            // TYPE_BOOLEAN
                { new Boolean(true),
                  new Boolean(false),
                  null },
            // TYPE_FLOAT
                { new Float(Float.MIN_VALUE),
                  new Float(Float.MAX_VALUE),
                  new Float(0),
                  null },
            // TYPE_DOUBLE
                { new Double(Double.MIN_VALUE),
                  new Double(Double.MAX_VALUE),
                  new Double(0),
                  null },
            // TYPE_STRING
                { new String(),
                  new String(" "),
                  new String(lowChars),
                  new String("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz01234567890`-=[];',./<>?:\"{}+_)(*&^%$#@!~ "),
                  null },
            // TYPE_DATE
                { new java.util.Date(70,0,1,12,0,0),
                  new java.util.Date(99,11,31,23,59,59),
                  null },
            } ;

        return testData ;
    }

    /**
        Test the capabilities of the database. Return the test
        result array, this is an array of string values
        for each Java type. The value of the string is
        null if the java type failed, otherwise it is the
        classname of the successful strategy.
    */
    private String[] testDatabase( Connection dbc ) throws Exception
    {
        // create an array for recording test results

        String      tcls[] = new String[JavaTypes.TypeName.length] ;
        int         pass[] = new int[JavaTypes.TypeName.length] ;
        TypeHandler thdl[] = new TypeHandler[JavaTypes.TypeName.length] ;

        Object[][] testData = getTestData() ;

        for (int i = 1 ; i < JavaTypes.TypeName.length ; i++)
        {
            System.out.println() ;
            System.out.println(">>> " + JavaTypes.TypeName[i]) ;

            pass[i] = -1 ;

            for (int strategy = 0 ; pass[i] == -1 && strategy < 9 ; strategy++)
            {
                String classname = "JOP.persist.jdbc.TypeHandler_" + JavaTypes.TypeName[i] + "_" + strategy ;

                try
                {
                    TypeHandler th = (TypeHandler)Class.forName(classname).newInstance() ;

                    System.out.println("  Loaded: " + JavaTypes.TypeName[i] + " strategy: " + strategy + " " + th ) ;

                    String  tablename = "JOP" + i ;

                    boolean tableCreated = false ;

                    try
                    {
                        createTable(dbc, tablename, th.getDefinition(), "var") ;

                        tableCreated = true ;

                        for (int z = 0 ; z < testData[i].length ; z++)
                        {
                            insertRow( dbc, tablename, th, z, testData[i][z] ) ;

                            Object obj = readRow( dbc, tablename, th, z ) ;

                            if (testData[i][z] == null)
                            {
                                if (obj != null)
                                {
                                    throw new Exception("Null value failed") ;
                                }
                            }
                            else if (!testData[i][z].equals(obj))
                            {
                                /*
                                if (i == JavaTypes.TYPE_STRING)
                                {
                                    String s1 = testData[i][z].toString() ;
                                    String s2 = obj.toString() ;
                                    char c1[] = s1.toCharArray() ;
                                    char c2[] = s2.toCharArray() ;

                                    System.out.println("len c1 = " + c1.length + " c2 = " + c2.length ) ;

                                    for (int ew = 0 ; ew < c1.length ; ew++)
                                    {
                                        if (c1[ew] != c2[ew])
                                        {
                                            System.out.println("DIFF [" + ew + "] c1 = " + (int)c1[ew] + " c2 = " + (int)c2[ew]) ;
                                        }
                                    }
                                }

                                if (i == JavaTypes.TYPE_OBJECT)
                                {
                                    System.out.println("hashcode = " + obj.hashCode() + "' expected '" + testData[i][z].hashCode() + "'") ;
                                }
                                */
                                throw new Exception("Unequal return value '" + obj + "' expected '" + testData[i][z] + "'") ;
                            }
                        }

                        dropTable(dbc, tablename) ;

                        pass[i] = strategy ;

                        thdl[i] = th ;

                    } catch (Exception ex)
                    {
                        System.out.println("  failed: " + ex.getMessage()) ;

                        if (tableCreated)
                        {
                            dropTable( dbc, tablename) ;
                        }
                    }

                } catch (ClassNotFoundException ex)
                {
                    // no strategy class found...
                }
            }

            if (pass[i] == -1)
            {
                System.out.println("<<< " + JavaTypes.TypeName[i] + " FAILED") ;
            }
            else
            {
                System.out.println("<<< " + JavaTypes.TypeName[i] + " PASSED") ;
            }
        }

        System.out.println("\nRESULT\n") ;

        for (int i = 1 ; i < JavaTypes.TypeName.length ; i++)
        {
            if (pass[i] == -1)
            {
                System.out.println( JavaTypes.TypeName[i] + " ** FAILED **" ) ;
            }
            else
            {
                tcls[i] = thdl[i].getClass().getName() ;

                System.out.println( JavaTypes.TypeName[i] + " " + thdl[i] ) ;
            }
        }

        return tcls ;
    }


    /**
        Read a 'test' row and return the object
    */
    private Object readRow(Connection dbc, String tablename, TypeHandler th, int index ) throws Exception
    {
        Object      retval = null ;
        Statement   stmt = null ;

        try
        {
            int count = 0 ;

            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select var from " + tablename + " where idx='" + new Integer(index).toString() + "'") ;

            while (rs.next())
            {
                if (count > 0)
                {
                    throw new SQLException("Too many rows returned") ;
                }
                else
                {
                    count = 1 ;
                }

                retval = th.readType( rs, 1 ) ;
            }

            stmt.close() ;

            if (count < 1)
            {
                throw new SQLException("No rows returned") ;
            }
        }
        catch (SQLException ex)
        {
            stmt.close() ;

            throw ex ;
        }

        return retval ;

    }

    /**
        Insert a test row into the database using the type handler.
    */
    private void insertRow(Connection dbc, String tablename, TypeHandler th, int index, Object obj) throws Exception
    {
        PreparedStatement stmt = null ;

        try
        {
            stmt = dbc.prepareStatement("insert into " + tablename + "(idx, var) values (?,?)") ;

            stmt.setString( 1, new Integer(index).toString()) ;

            th.saveType( stmt, 2, obj ) ;

            stmt.executeUpdate() ;

            stmt.close() ;
        } catch (Exception ex)
        {
            stmt.close() ;

            throw ex ;
        }
    }

    /**
        Drop a test table
    */
    private void dropTable(Connection dbc, String tablename) throws Exception
    {
        Statement   stmt = null ;

        try
        {
            stmt = dbc.createStatement() ;

            String s1 = "drop table " + tablename ;

            stmt.executeUpdate(s1) ;
        }
        catch (SQLException j)
        {
            System.out.println(j.getMessage()) ;
        }

        if (stmt != null)
        {
            stmt.close() ;
        }

        // System.out.println("*** " + tablename + " table dropped (" + coltype + ") ***") ;
    }

    /**
        Create a 'test' table and return 'true' if the table was
        successfully created
    */
    private void createTable(Connection dbc, String tablename, String coltype, String colname) throws Exception
    {
        Statement   stmt = null ;

        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select count(" + colname + ") from " + tablename ) ;

            rs.next() ;

            // System.out.println("  *** " + tablename + " already exists (" + coltype + ") ***") ;

            stmt.close() ;

            dropTable( dbc, tablename) ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }
        }

        try
        {
            stmt = dbc.createStatement() ;

            String s1 = "create table " + tablename + " (idx char(10), " + colname + " " + coltype + ") " ;

            stmt.executeUpdate(s1) ;

            stmt.close() ;

            // System.out.println("  *** " + tablename + " table created (" + coltype + ") ***") ;
        }
        catch (SQLException ex)
        {
            stmt.close() ;

            // System.out.println("  *** " + tablename + " table NOT created (" + coltype + ") ***") ;

            throw ex ;
        }
    }

    private int testTableNameLength(Connection dbc) throws Exception
    {
        String  tablename = "JJJ" ;
        boolean created = true ;
        int     length = 0 ;

        for (length = 4 ; created && length < 64 ; length++)
        {
            tablename = tablename + "J" ;

            try
            {
                createTable( dbc, tablename, "char(10)", "test") ;

                System.out.println("Table Name: " + tablename + " " + length ) ;

                dropTable( dbc, tablename ) ;

            } catch (SQLException ex)
            {
                created = false ;
            }
        }

        return length ;
    }

    private int testColumnNameLength(Connection dbc) throws Exception
    {
        String  colname = "JJJ" ;
        boolean created = true ;
        int     length = 0 ;

        for (length = 4 ; created && length < 64 ; length++)
        {
            colname = colname + "J" ;

            try
            {
                createTable( dbc, "test", "char(10)", colname) ;

                System.out.println("Column Name: " + colname + " " + length ) ;

                dropTable( dbc, "test") ;

            } catch (SQLException ex)
            {
                created = false ;
            }
        }

        return length ;
    }
}


