/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.jdbc ;

import JOP.persist.ObjectDefinition ;
import JOP.persist.ClassDefinition ;
import JOP.persist.SlotDefinition ;
import JOP.persist.ObjectIdentifier ;
import JOP.persist.OutputController ;

import  java.util.Hashtable ;
import  java.util.Enumeration ;
import  java.util.Vector ;

import  JOP.persist.JavaTypes ;

import java.net.URL;
import java.sql.*;

/**
    This class provides methods to update and insert objects into the JDBC
    data source. The class is created by the JdbcPersistManager and
    called from JOP.persist.PersistManager
    <p>
    *@see   JOP.persist.PersistManager
*/
class JdbcOutputController implements OutputController
{
    private JdbcPersistManager  parent ;
    private Hashtable           definedTables ;

    JdbcOutputController( JdbcPersistManager p )
    {
        parent = p ;

        definedTables = new Hashtable() ;
    }

    /**
        This method is called to update an object (ie an object
        with this id and of the same type already exists in
        the database.
        <p>
        This method will loop through the slots of each
        class of the object, writing each slot as we go.
    */
    public void updateObject( ObjectIdentifier oid, ObjectDefinition defObject, boolean explicit ) throws Exception
    {
        for (Enumeration e = defObject.classes() ; e.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)e.nextElement() ;

            updateTable( oid, defClass ) ;
        }
    }

    /**
        Called by updateObject to update each table. There is
        a table for each class in the hierarchy. There is also
        an index table (OidTable) that contains the name of
        the class of each object.
    */
    private void updateTable( ObjectIdentifier oid, ClassDefinition defClass ) throws Exception
    {
        //  build the SQL statement

        String  tableName = parent.config.tableName(defClass.getClassName()) ;

        StringBuffer s1 = new StringBuffer("UPDATE " + tableName + " SET " ) ;

        boolean colAdded = false ;

        for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
        {
            SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;
            String columnName = parent.config.columnName(defSlot.getName()) ;

            if (colAdded)
            {
                s1.append(",") ;
            }
            else
            {
                colAdded = true ;
            }

            s1.append(columnName + "=?") ;
        }

        // finish building the SQL command

        s1.append(" WHERE " + tableName + ".oid='" + oid + "';" ) ;

        Connection dbcon = null ;
        PreparedStatement dbstmt = null ;

        //  create the PreparedStatement and bind slot
        //  values.

        try
        {
            dbcon = parent.allocConnection(parent.CON_NORMAL) ;

            dbstmt = dbcon.prepareStatement(s1.toString()) ;

            int column = 1 ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                if (defSlot.isArray())
                {
                    deleteArray(oid, tableName, defClass, defSlot) ;

                    if (defSlot.isNullArray())
                    {
                        parent.typhdl.saveType( JavaTypes.TYPE_LONG, dbstmt, column++, null ) ;
                    }
                    else
                    {
                        long arrayid = insertArray( defClass, defSlot ) ;

                        parent.typhdl.saveType( JavaTypes.TYPE_LONG, dbstmt, column++, new Long(arrayid)) ;
                    }
                }
                else
                {
                    int type = defSlot.getSignatureType() ;

                    if (type == JavaTypes.TYPE_OBJECT)
                    {
                        parent.typhdl.saveType( JavaTypes.TYPE_OBJECT, dbstmt, column++, defSlot.getRefOid()) ;
                    }
                    else
                    {
                        parent.typhdl.saveType( defSlot.getSignatureType(), dbstmt, column++, defSlot.getValue()) ;
                    }
                }
            }

            // execute the command

            // System.out.println("SQL: " + s1.toString()) ;

            if (dbstmt.executeUpdate() != 1)
            {
                throw new JdbcPersistException("Row count not 1") ;
            }
        }
        finally
        {
            if (dbstmt != null)
            {
                dbstmt.close() ;
            }
        }
    }

    /**
        Insert a new object into the database. An insert of
        an object also updates the OidTable
    */
    public void insertObject( ObjectIdentifier oid, ObjectDefinition defObject, boolean explicit ) throws Exception
    {
        for (Enumeration e = defObject.classes() ; e.hasMoreElements() ; )
        {
            ClassDefinition defClass = (ClassDefinition)e.nextElement() ;

            insertTable( oid, defClass ) ;
        }

        parent.getIdTable().updateId( oid.getOid(), defObject.getClassName()) ;
    }

    /**
        Insert a new row in the table for each class in the
        hierarchy.
    */
    private void insertTable( ObjectIdentifier oid, ClassDefinition defClass ) throws Exception
    {
        defineTable( defClass ) ;

        String  tableName = parent.config.tableName(defClass.getClassName()) ;

        StringBuffer s1 = new StringBuffer("insert into " + tableName + " ( oid") ;
        StringBuffer s2 = new StringBuffer("VALUES ( ?" ) ;

        for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
        {
            SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;
            String columnName = parent.config.columnName(defSlot.getName()) ;

            s1.append("," + columnName) ;
            s2.append(",?") ;
        }

        /*
            Now we finish building the command
        */

        s1.append(") " ) ;
        s2.append(") " ) ;
        s1.append(s2) ;

        Connection dbcon = null ;
        PreparedStatement dbstmt = null ;

        try
        {
            dbcon = parent.allocConnection(parent.CON_NORMAL) ;

            dbstmt = dbcon.prepareStatement(s1.toString()) ;

            int column = 1 ;

            parent.typhdl.saveType( JavaTypes.TYPE_OBJECT, dbstmt, column++, oid ) ;

            for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
            {
                SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                if (defSlot.isArray())
                {
                    if (defSlot.isNullArray())
                    {
                        parent.typhdl.saveType( JavaTypes.TYPE_LONG, dbstmt, column++, null) ;
                    }
                    else
                    {
                        long arrayid = insertArray(defClass, defSlot) ;

                        parent.typhdl.saveType( JavaTypes.TYPE_LONG, dbstmt, column++, new Long(arrayid)) ;
                    }
                }
                else
                {
                    int type = defSlot.getSignatureType() ;

                    if (type == JavaTypes.TYPE_OBJECT)
                    {
                        parent.typhdl.saveType( JavaTypes.TYPE_OBJECT, dbstmt, column++, defSlot.getRefOid()) ;
                    }
                    else
                    {
                        parent.typhdl.saveType( defSlot.getSignatureType(), dbstmt, column++, defSlot.getValue()) ;
                    }
                }
            }

            /*
                and execute the command
            */

            // System.out.println("SQL: " + s1.toString()) ;

            if (dbstmt.executeUpdate() != 1)
            {
                throw new JdbcPersistException("Row count not 1") ;
            }
        }
        finally
        {
            if (dbstmt != null)
            {
                dbstmt.close() ;
            }
        }
    }

    /**
        A utility method to verify whether a table exists
        or not in the database. Used to determine whether
        or not the table should be created. This could be
        replaced by trapping specific types of exception.
    */
    private boolean tableExists(String tableName) throws Exception
    {
        /*
            First check to see if the table exists already
        */

        boolean tableExists = true ;

        Connection dbcon = null ;
        Statement dbstmt = null ;

        try
        {
            dbcon = parent.allocConnection(parent.CON_NORMAL) ;

            dbstmt = dbcon.createStatement() ;

            // System.out.println("SQL: select count(*) from " + tableName + ";") ;

            ResultSet r = dbstmt.executeQuery("select count(*) from " + tableName + ";") ;

            r.next() ;

            tableExists = true ;
        } catch (SQLException ex)
        {
            // ex.printStackTrace() ;

            tableExists = false ;
        }

        if (dbstmt != null)
        {
            dbstmt.close() ;
        }

        return tableExists ;
    }

    /**
        Define a table for the class. Fields which have a type
        that can be treated inline (normally immutable
        objects, although Date is also included) are stored
        as columns in the table, otherwise the object
        identifier of referenced objects is stored as a
        foreign key.
    */
    private void defineTable( ClassDefinition defClass ) throws Exception
    {
        String  tableName = parent.config.tableName(defClass.getClassName()) ;

        if (!definedTables.containsKey( tableName ))
        {
            /*
                First check to see if the table exists already
            */

            if (!tableExists(tableName))
            {
                /*
                    Create the table
                */

                StringBuffer s = new StringBuffer("create table " + tableName + " (oid " + parent.typhdl.getDefinition(JavaTypes.TYPE_OBJECT)) ;

                for (Enumeration eSlot = defClass.slots() ; eSlot.hasMoreElements() ; )
                {
                    SlotDefinition defSlot = (SlotDefinition)eSlot.nextElement() ;

                    /*
                        Get the name and signature for this slot. If
                        the slot is an array then a foreign key will
                        be created to another table holding the
                        elements of the array.
                    */
                    String columnName = parent.config.columnName(defSlot.getName()) ;

                    s.append(",") ;

                    if (defSlot.isArray())
                    {
                        s.append(columnName + " " + parent.typhdl.getDefinition( JavaTypes.TYPE_LONG )) ;
                    }
                    else
                    {
                        /*
                            Append the definition string
                        */

                        s.append(columnName + " " + parent.typhdl.getDefinition(defSlot.getSignatureType())) ;
                    }
                }

                /*
                    We are now ready to finish off the SQL
                    command that we have been building.
                */

                s.append(")") ;

                /*
                    Execute the SQL command using the DBO
                    connection
                */

                Connection dbcon = null ;
                Statement dbstmt = null ;

                dbcon = parent.dboConnection() ;

                dbstmt = dbcon.createStatement() ;

                // System.out.println("SQL: " + s.toString()) ;

                dbstmt.executeUpdate(s.toString()) ;

                // System.out.println("SQL: create unique index " + tableName + "OID on " + tableName + " (oid)") ;

                dbstmt.executeUpdate("create unique index " + tableName + "OID on " + tableName + " (oid)") ;

                dbcon.commit() ;

                if (dbstmt != null)
                {
                    dbstmt.close() ;
                }
            }

            definedTables.put(tableName,"EXISTS") ;
        }
    }

    /**
        Slots that are arrays are stored in a table
        that is a combination of the classname and the
        slot name. Each row has the array id (which is
        a foreign key to the referencing object) and an
        index.
        <p>
        The table has three columns the id of the array,
        the index and the element.
    */
    private void defineArray(ClassDefinition defClass, SlotDefinition defSlot) throws Exception
    {
        String arrayName = parent.config.arrayName(defClass.getClassName(), defSlot.getName()) ;

        if (!definedTables.containsKey(arrayName))
        {
            if (!tableExists(arrayName))
            {
                StringBuffer s1 = new StringBuffer("create table " + arrayName + " (arrayid " + parent.typhdl.getDefinition(JavaTypes.TYPE_LONG)
                                + ", arrayindex " + parent.typhdl.getDefinition(JavaTypes.TYPE_INTEGER)
                                + ", " ) ;

                String columnName = parent.config.columnName( defSlot.getName()) ;

                s1.append(columnName + " " + parent.typhdl.getDefinition(defSlot.getSignatureType()) + ")") ;

                Connection dbcon = null ;
                Statement dbstmt = null ;

                dbcon = parent.dboConnection() ;

                dbstmt = dbcon.createStatement() ;

                // System.out.println("SQL: " + s1.toString()) ;

                dbstmt.executeUpdate(s1.toString()) ;

                // System.out.println("SQL: create index " + arrayName + "ARRAYID on " + arrayName + " (arrayid)") ;

                dbstmt.executeUpdate("create index " + arrayName + "ARRAYID on " + arrayName + " (arrayid)") ;

                dbcon.commit() ;

                if (dbstmt != null)
                {
                    dbstmt.close() ;
                }
            }

            definedTables.put(arrayName, "EXISTS") ;
        }
    }


    private long insertArray(ClassDefinition defClass, SlotDefinition defSlot ) throws Exception
    {
        defineArray(defClass, defSlot) ;

        /*
            Some suitably random, non repeating number..ideally
            this can be an integer row-id or similar.
        */

        Long arrayid = new Long(JOP.persist.OidGenerator.newOid()) ;

        int     signature = defSlot.getSignatureType() ;
        String  arrayName = parent.config.arrayName( defClass.getClassName(), defSlot.getName()) ;

        for (int n = 0 ; n < defSlot.sizeArray() ; n++)
        {

            StringBuffer s1 = new StringBuffer("insert into " + arrayName + " (arrayid, arrayindex") ;

            /*
                Append the column information
            */

            Object obj = defSlot.getValue(n) ;

            if (obj != null)
            {
                String columnName = parent.config.columnName( defSlot.getName()) ;

                if (signature == JavaTypes.TYPE_OBJECT)
                {
                    obj = defSlot.getRefOid(n) ;

                    if (obj == null)
                    {
                        throw new JdbcPersistException("Slot '" + defSlot.getName() + "' has a null reference") ;
                    }
                }

                /*
                    Now we finish building the command
                */

                s1.append("," + columnName + ") values (?,?,?)" ) ;
            }
            else
            {
                s1.append(") values (?,?)" ) ;
            }

            /*
                Allocate a JDBC database statement
            */

            Connection dbcon = parent.allocConnection( parent.CON_ARRAY ) ;

            PreparedStatement dbstmt = dbcon.prepareStatement(s1.toString()) ;

            /*
                and now fill the statement values
            */

            parent.typhdl.saveType( JavaTypes.TYPE_LONG, dbstmt, 1, arrayid ) ;
            parent.typhdl.saveType( JavaTypes.TYPE_INTEGER, dbstmt, 2, new Integer(n)) ;

            if (obj != null)
            {
                parent.typhdl.saveType( signature, dbstmt, 3, obj ) ;
            }

            /*
                and execute the command
            */

            // System.out.println("SQL: " + s1.toString()) ;

            if (dbstmt.executeUpdate() != 1)
            {
                dbstmt.close() ;

                throw new JdbcPersistException("Row count not 1") ;
            }

            dbstmt.close() ;
        }

        return arrayid.longValue() ;
    }

    /**
        Update an array slot. This is not very efficient in
        this version, we simply delete the current rows
        from JOP_Array and re-insert a new set of rows.
        <p>
        JOP_Array is going to become a hot-spot anyway if an
        application is using lots of arrays so I need to
        optimise this in a later release.
    */
    private void deleteArray(ObjectIdentifier oid, String tableName, ClassDefinition defClass, SlotDefinition defSlot ) throws Exception
    {
        /*
            DELETE FROM JOP_ARRAY
                where JOP_ARRAY.arrayid = (SELECT col FROM Utable where oid='c79b-4830-ff3d-50d4')
        */

        Statement dbstmt = null ;

        //  create the PreparedStatement and bind slot
        //  values.

        try
        {
            Connection dbcon1 = parent.allocConnection(parent.CON_NORMAL) ;

            dbstmt = dbcon1.createStatement() ;

            StringBuffer s1 = new StringBuffer( "SELECT " ) ;
            s1.append( tableName + "." + parent.config.columnName(defSlot.getName())) ;
            s1.append( " FROM " + tableName + " WHERE (" + tableName + ".oid='" + oid.toString() + "')" ) ;

            // System.out.println("SQL: " + s1.toString()) ;

            ResultSet r = dbstmt.executeQuery(s1.toString()) ;

            if (!r.next())
            {
                throw new JdbcPersistException("Cannot read array id") ;
            }

            Long arrayid = (Long)parent.typhdl.readType( JavaTypes.TYPE_LONG, r, 1 ) ;

            if (arrayid != null)
            {
                Connection dbcon2 = parent.allocConnection(parent.CON_ARRAY) ;

                dbstmt = dbcon2.createStatement() ;

                String arrayName = parent.config.arrayName(defClass.getClassName(), defSlot.getName()) ;

                StringBuffer s2 = new StringBuffer( "DELETE FROM " + arrayName + " where " + arrayName + ".arrayid = '" + arrayid.toString() + "'") ;

                // System.out.println("SQL: " + s2.toString()) ;

                dbstmt.executeUpdate(s2.toString()) ;
            }
        }
        finally
        {
            if (dbstmt != null)
            {
                dbstmt.close() ;
            }
        }
    }
}

