/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.jdbc ;

import JOP.persist.PersistManager ;
import JOP.persist.PersistException ;
import JOP.persist.ObjectIdentifier ;

import java.net.URL;
import java.sql.*;

import  java.util.Vector ;
import  java.util.Hashtable ;
import  java.util.Enumeration ;

/**
    The JdbcPersistManager supports saving
    and retrieving objects, and complex graphs of
    objects to a database via JDBC. The JdbcPersistManager works with
    any JDBC data source and will run on all Java platforms where
    JDBC is available.
    <p>
    Before using JdbcPersistManager you should configure the
    database using <b>JOP.persist.jdbc.JdbcConfig</b>
    <p>
    The JdbcPersistManager implements several useful features:
    <p>
    <ul>
    <li type=square>
    Automatic Schema Creation. The generated schema is quite
    straightforward and can be used from end-user
    query tools or other applications.
    <li type=square>
    cached reads within a unit of work
    <li type=square>
    only updates where objects have changed (to be implemented)
    <li type=square>
    transaction support via the unit of work
    <li type=square>
    efficient SQL retrieval for slots (attributes) of an object
    <li type=square>
    in-line support for simple object types (such as String)
    thereby using foreign keys only where necessary
    </ul>
    There are several features on the drawing board, yet to be included in
    this class including mutli-user support and optimistic locking.
    <p>
    *@see    JOP.persist.jdbc.JdbcConfig
*/

public class JdbcPersistManager extends PersistManager
{
    private String  jdbcUrl = null ;
    private String  jdbcDriver = null ;
    private String  usrName ;
    private String  dboName ;
    private String  usrPassword ;
    private String  dboPassword ;
    private JdbcIdTable  idTable = null ;
    private JdbcOutputController outCntl ;
    private JdbcInputController inCntl ;

    /**
        true if multiple connections for concurrent operations
    */
    private boolean     multiConnection = false ;

    /**
        true if use explicit transactions, otherwise
        use implicit transaction mode
    */
    private boolean     transactMode = true ;

    private Connection  connects[] ;

    final static int CON_ID = 0 ;
    final static int CON_NORMAL = 1 ;
    final static int CON_ARRAY = 2 ;
    final static int CON_DBO = 3 ;

    DatabaseTypes       config = null ;
    TypeHandlerManager  typhdl = null ;

    private boolean open = false ;

    /**
        Prepares to open a connection to a specific database server. Both a
        normal user name and a database owner (dbo) user name can be specified.
        The dbo user name is used for all table and index creation
        operations.
        <p>
        This method does not open the connection, but rather
        prepares the information needed to open the connection.
        The open() method should be invoked to complete the connection.
        <p>
        *@see   #open
        *@throws Exception  Some error occured - most likely a
                            JDBC configuration error.
    */

    public JdbcPersistManager(String driver, String url, String unam, String upwd, String onam, String opwd ) throws Exception
    {
        super() ;

        jdbcDriver = driver ;
        jdbcUrl = url ;
        usrName = unam ;
        usrPassword = upwd ;
        dboName = onam ;
        dboPassword = opwd ;
    }

    /**
        Alternate constructor where <i>user_name</i> = <i>dbo_user_name</i>
    */
    public JdbcPersistManager(String driver, String url, String unam, String upwd ) throws Exception
    {
        this( driver, url, unam, upwd, unam, upwd ) ;
    }

    /**
        This is an internal method to return the objecct responsible
        for managing the table of object identifiers.
    */
    JdbcIdTable getIdTable()
    {
        return idTable ;
    }

    /**
        Establish the connection to the datasource via
        JDBC using the information passed to the constructor.
        <p>
    */
    public void open( ) throws Exception
    {
        if (open)
        {
            throw new JdbcPersistException("File already open") ;
        }

        // Load the jdbc-odbc bridge driver

        Connection dbdbc ;
        Class d = Class.forName(jdbcDriver);

        DriverManager.registerDriver( (Driver)d.newInstance()) ;

        //  I get an error with the SUN jdbc-odbc driver
        //  when trying to set the transaction mode. So
        //  for now I will turn it off.

        dbdbc = DriverManager.getConnection( jdbcUrl, usrName, usrPassword);

        // setup whether we use multiple connections etc

        setupConnectionMode(dbdbc.getMetaData()) ;

        if (transactMode)
        {
            dbdbc.setAutoCommit( false ) ;
        }

        //  get the database types setup

        typhdl = new TypeHandlerManager( dbdbc ) ;

        config = new DatabaseTypes(dbdbc) ;

        // If we were unable to connect, an exception
        // would have been thrown.  So, if we get here,
        // we are successfully connected to the URL

        // Check for, and display and warnings generated
        // by the connect.

        checkForWarning(dbdbc.getWarnings());

        dumpMetaData( dbdbc ) ;

        // open the connections passing the first connection

        setupConnections( dbdbc ) ;

        //  Methods to create the 'id table'. The id table
        //  contains object identifiers and the classnames
        //  of the objects stored with those identifiers.

        idTable = new JdbcIdTable(this) ;

        //  Setup a controller to write to the data source

        outCntl = new JdbcOutputController(this) ;

        //  and a controller to read from the data source

        inCntl = new JdbcInputController(this) ;

        open = true ;
    }

    /**
        Setup the way connections are handled for this data source. This
        may need to be changed as more data source types are
        tested
    */
    private void setupConnectionMode(DatabaseMetaData dma) throws SQLException
    {
        if (dma.getDatabaseProductName().equals("Microsoft SQL Server"))
        {
            transactMode = true ;
            multiConnection = true ;
        }
        else if (dma.getDatabaseProductName().equals("ACCESS"))
        {
            transactMode = true ;
            multiConnection = false ;
        }
        else
        {
            transactMode = true ;
            multiConnection = false ;
        }
    }

    /**
        Close a previously open connection to the persistent
        data source.
    */
    public void close() throws Exception
    {
        if (!open)
        {
            throw new JdbcPersistException("Database not open") ;
        }

        open = false ;
    }

    /**
        Start a database transaction. Does nothing in JDBC as
        transactions are implicitly opened.
    */
    protected synchronized void startTrans() throws Exception
    {
        if (!open)
        {
            throw new JdbcPersistException("Database not open") ;
        }
    }

    /**
        Commits a database transaction. This involves issuing
        commits on all connections for this database.
    */
    protected synchronized void commitTrans() throws Exception
    {
        if (!open)
        {
            throw new JdbcPersistException("Database not open") ;
        }

        for (int i = 0 ; i < connects.length ; i++)
        {
            if (connects[i] != null)
            {
                connects[i].commit() ;
            }
        }
    }

    /**
        Aborts the current database transaction. This involves issuing
        aborting on all connections for this database.
    */
    protected synchronized void abortTrans() throws Exception
    {
        if (!open)
        {
            throw new JdbcPersistException("Database not open") ;
        }

        for (int i = 0 ; i < connects.length ; i++)
        {
            if (connects[i] != null)
            {
                connects[i].rollback() ;
            }
        }
    }

    /**
        If we get garbage collected then close the database.
    */
    protected void finalize()
    {
        try
        {
            if (open)
            {
                close() ;
            }
        }
        catch (Exception ex)
        {
        }
    }

    /**
        Save the object on the database. The object can
        implement Persistable, have a custom marshall
        class available or can be automatically marshalled. The
        object and all of the objects referenced by the
        object are saved. Loops are detected and
        avoided.
        <p>
        To retrieve the object call 'streamInObject()'
        <p>
        *@see    #streamInObject

        *@param     o   The object to be saved
        *@returns       The unique identifier of the object
        *@throws    java.lang.Exception
    */
    public long streamOutObject(Object o) throws Exception
    {
        if (!open)
        {
            throw new JdbcPersistException("Not open") ;
        }

        Vector  history = new Vector() ;
        ObjectIdentifier oid = null ;

        oid = streamOutObject( o, history, outCntl, true ) ;

        return oid.getOid() ;
    }

    /**
        Retrieves an object previously saved by 'streamOutObject()'.
        The unique object id is passed and an object and all
        of the objects referenced by the object are
        returned. The object will be an instance of the same type
        that was saved and will have the same context. The
        caller of this method must explicitly
        cast the returned object before using it.
        <p>
        Within a consistency unit multiple reads of an object
        either directly (via this method) or indirectly because
        the object is referenced by another object will not
        create secondary instances.
        <p>
        *@see    JOP.persist.PersistManager#streamInObject
        *@see    #streamOutObject
        <p>
        *@param     id  Unique indentifier of the object
        *@returns       The object
    */
    public Object streamInObject(long xid) throws Exception
    {
        ObjectIdentifier oid ;

        if (!open)
        {
            throw new JdbcPersistException("Not open") ;
        }

        oid = streamInObject( xid, inCntl ) ;

        return oid.getObject() ;
    }


    /**
        Check and display database warnings. Source code based
        on the SimpleSelect.java distributed with JDBC.
    */
    private static boolean checkForWarning(SQLWarning warn) throws SQLException
    {
        boolean rc = false;

        // If a SQLWarning object was given, display the
        // warning messages.  Note that there could be
        // multiple warnings chained together

        if (warn != null) {
            System.out.println ("\n *** Warning ***\n");
            rc = true;
            while (warn != null) {
                System.out.println ("SQLState: " +
                    warn.getSQLState ());
                System.out.println ("Message:  " +
                    warn.getMessage ());
                System.out.println ("Vendor:   " +
                    warn.getErrorCode ());
                System.out.println ("");
                warn = warn.getNextWarning ();
            }
        }
        return rc;
    }


    /**
        Setup the connection array given the first
        connection (a user mode connection)
    */
    private void setupConnections( Connection dbc ) throws SQLException
    {
        connects = new Connection[4] ;

        // this first block assigns a DBO connection and
        // the generic user connection

        if (dboName.equals(usrName))
        {
            // user name and dbo name are identical

            connects[CON_DBO] = dbc ;

            if (multiConnection == false)
            {
                connects[CON_ID] = dbc ;
            }
        }
        else
        {
            // this is the first dbo user

            connects[CON_ID] = dbc ;

            connects[CON_DBO] = DriverManager.getConnection( jdbcUrl, dboName, dboPassword);

            if (transactMode)
            {
                try
                {
                    connects[CON_DBO].setAutoCommit( false ) ;
                } catch (SQLException ex)
                {
                    if (!jdbcDriver.equalsIgnoreCase("jdbc.odbc.JdbcOdbcDriver"))
                    {
                        throw ex ;
                    }
                }
            }
        }

        if (multiConnection == false)
        {
            connects[CON_ID] = connects[CON_ID] ;
            connects[CON_NORMAL] = connects[CON_ID] ;
            connects[CON_ARRAY] = connects[CON_ID] ;
        }
    }

    /**
        Allocates a connection handle. There are three
        connections handle types.
        <p>
        0 = JOP_Id table
        1 = normal tables
        2 = JOP_ARRAY table
        3 = DBO connection
    */
    Connection allocConnection(int typ) throws SQLException
    {
        Connection c = connects[typ] ;

        if (c == null)
        {
            connects[typ] = DriverManager.getConnection( jdbcUrl, usrName, usrPassword);

            if (transactMode)
            {
                try
                {
                    connects[typ].setAutoCommit( false ) ;
                } catch (SQLException ex)
                {
                    if (!jdbcDriver.equalsIgnoreCase("jdbc.odbc.JdbcOdbcDriver"))
                    {
                        throw ex ;
                    }
                }
            }

            c = connects[typ] ;
        }

        return c ;
    }

    /**
        Return a a DBO connection
    */
    Connection dboConnection() throws Exception
    {
        return allocConnection(CON_DBO) ;
    }

    /**
        Just print the driver version etc
    */
    private void dumpMetaData(Connection dbdbc) throws SQLException
    {
        // Get the DatabaseMetaData object and display
        // some information about the connection

        DatabaseMetaData dma = dbdbc.getMetaData();

        System.out.println("Connected to :" + dma.getURL());
        System.out.println("User name    :" + dma.getUserName()) ;
        System.out.println("Driver       :" + dma.getDriverName());
        System.out.println(" version     :" + dma.getDriverVersion());
        System.out.println("Database     :" + dma.getDatabaseProductName());
        System.out.println(" version     :" + dma.getDatabaseProductVersion());
        System.out.println("");
    }
}


