/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.persist.jdbc ;

import  java.sql.* ;

import  JOP.persist.JavaTypes ;

/**
	The TypeHandler maps the basic Java types to SQL
	data types and JDBC driver methods. Unfortunately these
	classes are necessary because:
    <p>
    <ol>
    <li>
	JDBC drivers do not have to implement all data types
	<li>
	DBMS vary in their support for data types
	<li>
	Many things are left unspecified and therefore
	behaviour varies from driver to driver and dbms
	to dbms..
    </ol>
    As an example, I am using two ODBC drivers against the
    exact same database. One driver allows null 'timestamp'
    values the other does not, one driver automatically
    trims spaces from strings, the other does not, one
    driver convers Character('\0') to null, the other
    converts it to "", the underlying DBMS does not
    support BIGINT (ie Long) values.
    <p>
    As you can see an approach is necessary to manage this
    situation.
    <p>
    These classes are designed as follows:
    <p>
    <ul>
    <li>
    each built in Java type has a 'type handler'
	<li>
    each 'type handler' supports a number of strategies
    for saving that type (eg direct mapping to the
    JDBC data type, conversion to a VARCHAR etc).
	<li>
    a configuration utility determines the strategies
    for each type handler that will work with a specific
    combination of driver and database. These combinations
    are stored in the database.
	<li>
    when JOP needs to save or read one of the basic Java
    types it invokes the appropriate type handler, the
    type handler uses the pre-determined strategy to
    save/read the data
    </ul>
    <p>
    Once a database has been initialised with a specific
    set of strategies it will not be possible to use a
    different Jdbc driver unless the exact same strategies
    can be used with identical results.
    </p><p>
    The utility program JOP.persist.jdbc.JdbcConfig must
    be used to pre-configure the strategy and also to
    print a report of the driver and database capabilities.
    </p><p>
    User defined type handlers can be added by implementing
    the JOP.persist.jdbc.TypeHander interface and by
    defining a class with the correct naming convention.
*/
class TypeHandlerManager
{
    private TypeHandler th[] = new TypeHandler[JavaTypes.TypeName.length] ;

    /**
        Initialise a type handler for the given database
    */
    TypeHandlerManager( Connection dbc ) throws DriverNotConfiguredException, SQLException, IllegalAccessException, ClassNotFoundException, InstantiationException
    {
        this( dbc, true ) ;
    }

    /**
        Initialise the type handler and optionally check to see if the driver
        has been configured.
    */

    TypeHandlerManager( Connection dbc, boolean check ) throws DriverNotConfiguredException, SQLException, IllegalAccessException, ClassNotFoundException, InstantiationException
    {
        Statement   stmt = null ;

        if (check)
        {
            //  check to see if driver has been configured

            DatabaseMetaData dma = dbc.getMetaData();

            String  driver = dma.getDriverName().trim() + " " + dma.getDriverVersion().trim() ;
            String  dbms = dma.getDatabaseProductName().trim() + " " + dma.getDatabaseProductVersion().trim() ;

            int     count = 0 ;

            try
            {
                stmt = dbc.createStatement() ;

                ResultSet rs = stmt.executeQuery("select count(*) from JOPconfig where (driver='" + driver + "' and dbms='" + dbms + "');") ;

                if (rs.next())
                {
                    count = rs.getInt(1) ;
                }

                stmt.close() ;
            }
            catch (SQLException j)
            {
                if (stmt != null)
                {
                    stmt.close() ;
                }

                count = 0 ;
            }

            if (count == 0)
            {
                throw new DriverNotConfiguredException("run JOP.persist.jdbc.JdbcConfig") ;
            }
        }

        try
        {
            stmt = dbc.createStatement() ;

            ResultSet rs = stmt.executeQuery("select joptype, typehandler from JOPtypes") ;

            while (rs.next())
            {
                int     jopType = rs.getInt(1) ;
                String  typeHandler = rs.getString(2).trim() ;

                System.out.println("Loading type handler: " + jopType + " = " + typeHandler ) ;

                th[jopType] = (TypeHandler)Class.forName(typeHandler).newInstance() ;
            }

            stmt.close() ;
        }
        catch (SQLException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }

            throw j ;
        }
        catch (ClassNotFoundException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }

            throw j ;
        }
        catch (IllegalAccessException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }

            throw j ;
        }
        catch (InstantiationException j)
        {
            if (stmt != null)
            {
                stmt.close() ;
            }

            throw j ;
        }
    }


    /**
        Return the TypeHandler for a given Java type.
    */
    TypeHandler getTypeHandler( int type )
    {
        return th[type] ;
    }

    /**
        Save the 'object' by binding it to column of the Jdbc
        statement 'stmt'. The type is one of the constants
        defined in JOP.persist.JavaTypes.
    */
    void saveType( int type, PreparedStatement stmt, int column, Object object ) throws SQLException
    {
        th[type].saveType( stmt, column, object ) ;
    }

    /**
        Read the type and return an object
    */
    Object readType( int type, ResultSet rs, int column ) throws SQLException
    {
        return th[type].readType( rs, column ) ;
    }

    String getDefinition(int type)
    {
        return th[type].getDefinition() ;
    }
}

