/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.util ;

import JOP.persist.PersistManager ;
import JOP.persist.PersistController ;
import JOP.persist.Persistable ;
import JOP.persist.PersistException ;
import JOP.persist.ObjectIdentifier ;

import java.util.Date ;
import java.util.Hashtable ;
import java.util.Enumeration ;

/**
    A HashtableDB implements a subset of the functionality
    of a Hasthable. The HashtableDB supports the late
    fetching of objects, ie objects are only fetched
    from the database when they are referenced - hence
    performance is improved. The semantics
    of Hashtable are maintained so that HashtableDB and
    Hasthable can be substituted for one another in applications
    programs.
    <p>
    One of the main uses of HashtableDB is support indexed
    collections of persistent objects.
    <p>
    Like a Hashtable when a HasthableDB is saved to a database by
    calling streamOutObject(), the
    hashtable, its keys and all referenced objects will
    be automatically saved.
    <p>
    A typical application will have a 'root' HashtableDB with
    a well known object id (read from an environment variable
    etc). When this HashableDB is read it contains references
    to other HashtableDB's and other application objects.
    <p>
    If you streamOutObject() a HashtableDB to a different
    PersistManager than the one that was used to read the
    HashtableDB (or first save the HashtableDB) then the
    HashtableDB will be fully resolved (ie all referenced objects
    will be read) before the HashtableDB is written to the second
    PersistManager.
*/

public class HashtableDB implements Persistable
{
    Hashtable                   ht ;            // keys plus oids
    PersistManager              pmgr ;
    Hashtable                   newObjects ;    // keys plus objects
    Hashtable                   inuseObjects ;
    Hashtable                   pmgrTables ;
    private static int          htWork = 0 ;    // trans sequence

    public HashtableDB()
    {
        clear() ;
        pmgrTables = new Hashtable() ;
    }

    /**
        Remove all entries from the hashtable
    */
    public synchronized void clear()
    {
        if (ht == null)
        {
            ht = new Hashtable() ;
        }
        else
        {
            ht.clear() ;
        }

        newObjects = new Hashtable() ;
        inuseObjects = new Hashtable() ;
    }

    /**
        Test if the hashtable contains the given key and return 'true'
        if the key is found.
    */
    public boolean containsKey( Object key )
    {
        if (ht.containsKey(key) || newObjects.containsKey(key))
        {
            return true ;
        }
        else
        {
            return false ;
        }
    }

    /**
        Remove an object with the given key from the hashtable.
    */
    public void remove(Object key)
    {
        ht.remove(key) ;
        inuseObjects.remove(key) ;
        newObjects.remove(key);
    }

    /**
        Add a new object to the hashtable accessible by
        the key value.
    */
    public synchronized void put( Object key, Object obj )
    {
        if (ht.get(key) != null || newObjects.get(key) != null)
        {
            throw new RuntimeException("HashtableDB does not support duplicate keys") ;
        }
        else
        {
            newObjects.put( key, obj ) ;
        }
    }

    /**
        Retrieve an object from the hashtable with the given key. If
        the object is not currently in memory or in the JOP cache
        the object will be retrieved from the persistent database. Returns
        'null' if the object is not in the hashtable or cannot be
        retrieved.
        <p>
        Any exceptions from the PersistManager whilst retrieving the
        object are caught and a strace trace is printed.
        <p>
    */
    public synchronized Object get( Object key )
    {
        /* select the 'ht' based upon the stream manager */

        Object  o = ht.get(key) ;

        if (o != null)
        {
            if (o instanceof ObjectIdentifier)
            {
                return get(((ObjectIdentifier)o).getOid()) ;
            }
            else
            {
                return o ;
            }
        }
        else
        {
            return newObjects.get( key ) ;
        }
    }

    synchronized Object get( long oid )
    {
        Long key = new Long(oid) ;

        Object o = inuseObjects.get( key ) ;

        try
        {
            if (o == null && pmgr != null)
            {
                /*  the object has not been read yet */

                o = pmgr.streamInObject( oid ) ;

                inuseObjects.put( key, o ) ;
            }
        } catch (Exception ex)
        {
            ex.printStackTrace() ;
            return null ;
        }

        return o ;
    }

    /**
        Return an enumerator of all objects over their key values. As the
        'getNext()' method is called for the enumerator objects will
        be retrieved from the persistent database (or the JOP cache).
    */
    public Enumeration keys()
    {
        return new HashtableDBEnumerator(ht.keys(), this, true) ;
    }

    /**
        Return an enumerator of all objects in the hashtable. As the
        'getNext()' method is called for the enumerator objects will
        be retrieved from the persistent database (or the JOP cache).
    */
    public Enumeration elements()
    {
        return new HashtableDBEnumerator(ht.elements(), this, false) ;
    }

    /**
        Called by the PersistManager when this object is retrieved
        from the persistent database.
        <p>
        To retrieve a HashtableDB object call the 'streamInObject()' method
        or the 'streamInRoot()' method of the PersistManager.
    */
    public synchronized void streamIn( PersistController c ) throws Exception
    {
        pmgr = c.getPersistManager() ;

        ht = (Hashtable)c.getObject("ht") ;

        // System.out.println("StreamIn = " + ht) ;

        /*
            reset the memory cache tables to be consistent
            with the fact that we have just read the object
            from the database
        */

        newObjects = new Hashtable() ;
        inuseObjects = new Hashtable() ;
    }

    /**
        Called by the PersistManager when this object is saved
        to the persistent database. To save a HashtableDB call the
        'streamOutObject()' or 'streamOutRoot()' methods of the
        PersistManager.
    */
    public synchronized void streamOut( PersistController c ) throws Exception
    {
        /*
            Save the persistance manager for later use. The
            purpose of this routine is to save the Hashtable
            to the database (keys and oid's) and to set
            the memory 'inuse' and 'newObject' tables to
            reflect the state of the DBMS
        */

        PersistManager pmgrout = c.getPersistManager() ;

        if (pmgr != null && pmgr != pmgrout)
        {
            /*
                This is a different PersistManager than was
                originally used to save or read the HashtableDB.
            */

            HashtableDB ot = (HashtableDB)pmgrTables.get(pmgrout) ;

            if (ot == null)
            {
                ot = new HashtableDB() ;
                pmgrTables.put( pmgrout, ot ) ;
            }

            /*
                Clear the output persist manager specific
                hashtable
            */

            ot.clear() ;

            /*
                Get the resolved objects for every object in
                the hashtable
            */

            for (Enumeration e = this.keys() ; e.hasMoreElements() ; )
            {
                Object key = e.nextElement() ;
                Object obj = this.get( key ) ;

                ot.put( key, obj ) ;
            }

            ot.streamOut( c ) ;
        }
        else
        {
            pmgr = pmgrout ;

            /*
                For any objects that have been added to the
                hashtable, stream them out and get their object
                id's then save the id in the hashtable.
            */

            for (Enumeration e = newObjects.keys() ; e.hasMoreElements() ; )
            {
                Object key = e.nextElement() ;
                Object obj = newObjects.get( key ) ;

                if ((obj instanceof String) ||
                    (obj instanceof Integer) ||
                    (obj instanceof Date) ||
                    (obj instanceof Boolean) ||
                    (obj instanceof Float) ||
                    (obj instanceof Long) ||
                    (obj instanceof Double) ||
                    (obj instanceof Character))
                {
                    ht.put( key, obj ) ;
                }
                else
                {
                    long xid = pmgr.streamOutObject( obj ) ;

                    ObjectIdentifier id = new ObjectIdentifier( xid ) ;

                    // System.out.println("Moving key=" + key + ", obj=" + obj + " to hashtable as " + id ) ;

                    ht.put( key, id ) ;
                }
            }

            /*
                Remove the record of the new objects
            */

            newObjects = new Hashtable() ;

            /*
                Any objects that were referenced also need to
                be streamed out
            */

            for (Enumeration e = inuseObjects.keys() ; e.hasMoreElements() ; )
            {
                Long key = (Long)e.nextElement() ;
                Object obj = inuseObjects.get( key ) ;

                long id = pmgr.streamOutObject( obj ) ;

                if (id != key.longValue())
                {
                    throw new Exception("Inconsistent (hashtable) object id " + id + " " + key.longValue()) ;
                }
            }

            /*
                reset the list of 'read' objects
            */

            inuseObjects = new Hashtable() ;

            c.put( "ht", ht ) ;
        }
    }

    /**
        Called by the PeristManager to define the hashtable.
    */
    public void defineObject( PersistController c ) throws Exception
    {
        c.defineSlot("JOP.util.HashtableDB", "ht", PersistController.TYPE_OBJECT ) ;
    }
}


class HashtableDBEnumerator implements Enumeration
{
    private Enumeration         enum ;
    private HashtableDB         parent ;
    private boolean             newObjects = false ;
    private boolean             keys = false ;

    HashtableDBEnumerator( Enumeration e, HashtableDB p, boolean dokeys )
    {
        enum = e ;
        parent = p ;
        newObjects = false ;
        keys = dokeys ;
    }

    public synchronized boolean hasMoreElements()
    {
        if (enum.hasMoreElements())
        {
            return true ;
        }

        if (newObjects)
        {
            return false ;
        }
        else
        {
            newObjects = true ;

            if (keys)
            {
                enum = parent.newObjects.keys() ;
            }
            else
            {
                enum = parent.newObjects.elements() ;
            }

            return enum.hasMoreElements() ;
        }
    }

    public synchronized Object nextElement()
    {
        if (keys)
        {
            /*
                we are iterating over the hashtables keys
            */

            return enum.nextElement() ;
        }
        else
        {
            /*
                we are iterating over the hashtables elements
            */

            if (newObjects)
            {
                return enum.nextElement() ;
            }
            else
            {
                Object  o = enum.nextElement() ;

                if (o instanceof ObjectIdentifier)
                {
                    // enum.nextElement() ;

                    if (o != null)
                    {
                        return parent.get(((ObjectIdentifier)o).getOid()) ;
                    }
                    else
                    {
                        return null ;
                    }
                }
                else
                {
                    return o ;
                }
            }
        }
    }
}

