/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.util ;

import JOP.persist.PersistController ;
import JOP.persist.Persistable ;
import JOP.persist.PersistException ;
import JOP.persist.PersistManager ;
import JOP.persist.ObjectIdentifier ;

/**
    RefDB offers the ability to have a reference to an
    object that is read from the persistent store
    only when it is referenced. This can improve performance, especially
    when an object contains references to complex objects that are
    only occasionally traversed.
    <pre>
    public class MyClass
    {
        RefDB   relatedObject ;

        void setRelatedObject( AnotherObject o )
        {
            relatedObject.setObject( o ) ;
        }

        AnotherObject getRelatedObject() throws Exception
        {
            return (AnotherObject)relatedObject.getObject() ;
        }
    }
    </pre>
    In the above example the instances of MyClass can be saved and restored
    as needed. When an instance of MyClass is restored the instance of
    AnotherObject (stored in the 'relatedObject' RefDB) is not read
    from the persistent database until the 'getRelatedObject()'
    method is called.
*/

public class RefDB implements Persistable
{
    private PersistManager      pmgr ;
    private Object              obj = null ;
    private long                oid = 0 ;
    private boolean             readAttempted = false ;

    /**
        All objects that implement JOP.persist.Persistable require a public
        constructor that takes no arguments.
    */
    public RefDB()
    {
    }

    /**
        Initialise the reference with an existing object.
    */
    public RefDB(Object o)
    {
        setObject(o) ;
    }

    /**
        Change the object referenced. The referenced object will be saved when the
        object containing the reference is saved.
    */
    public void setObject(Object o)
    {
        obj = o ;
    }

    public String toString()
    {
        if (obj == null)
        {
            ObjectIdentifier id = new ObjectIdentifier(oid) ;
            return getClass().getName() + ":" + id.toString() ;
        }
        else
        {
            return obj.toString() ;
        }
    }

    /**
        Get the referenced object. If the referenced object is not in memory
        then it will be read from the persistent store. Returns 'null' if
        no object was stored in the RefDB.
        <p>
        *@throws    JOP.persist.ObjectNotFoundException an error occured and
                    the RefDB is no longer correct.
    */
    public Object getObject() throws Exception
    {
        if (obj == null && !readAttempted)
        {
            if (oid == 0)
            {
                obj = null ;
            }
            else
            {
                obj = pmgr.streamInObject( oid ) ;

                readAttempted = true ;
            }
        }

        return obj ;
    }

    /**
        Called by the PersistManager. Required to implement JOP.persist.Persistable
    */
    public synchronized void streamIn( PersistController c ) throws Exception
    {
        pmgr = c.getPersistManager() ;

        oid = c.longValue("ref") ;

        obj = null ;

        readAttempted = false ;
    }

    /**
        Called by the PersistManager. Required to implement JOP.persist.Persistable
    */
    public synchronized void streamOut( PersistController c ) throws Exception
    {
        pmgr = c.getPersistManager() ;

        if (obj != null)
        {
            // System.out.println("Obj = " + obj ) ;

            oid = pmgr.streamOutObject(obj) ;
        }

        c.put("ref", oid ) ;
    }

    /**
        Defines a RefDB as an object identifier (ie a long)
    */
    public void defineObject( PersistController c ) throws Exception
    {
        c.defineSlot("JOP.util.RefDB", "ref", PersistController.TYPE_LONG ) ;
    }
}


