/*
    Copyright (c) 1996, David Rothwell. All Rights Reserved.

    http://www.magna.com.au/~davidr     mailto:davidr@magna.com.au

    Permission to use, copy, modify, and distribute this software
    and its documentation for NON-COMMERCIAL purposes and without
    fee is hereby granted provided that this copyright notice
    appears in all copies. This software is distributed on an 'as is'
    basis without warranty.

    Release: 0.4a   11-Sep-1996

 **/

package JOP.util ;

import JOP.persist.PersistController ;
import JOP.persist.Persistable ;
import JOP.persist.PersistException ;
import JOP.persist.PersistManager ;
import JOP.persist.ObjectIdentifier ;

import java.util.Vector ;
import java.util.Enumeration ;
import java.util.Hashtable ;

/**
    VectorDB offers a subset of the functionality of a Vector with
    the benefit of supporting a late fetch on object
    references.
    <p>
    The VectorDB maintains a vector of objects that
    have been accessed (either by an enumeration or directly) and ensures
    that all those objects are saved when the
    VectorDB instance is saved.
*/
public class VectorDB implements Persistable
{
    Vector              vt ;
    PersistManager      pmgr ;
    Vector              newObjects ;
    Hashtable           inuseObjects ;
    Hashtable           pmgrTables ;

    public VectorDB()
    {
        removeAllElements() ;

        pmgrTables = new Hashtable() ;
    }

    /**
        Remove all the elements of this vector.
    */
    public void removeAllElements()
    {
        if (vt == null)
        {
            vt = new Vector() ;
        }
        else
        {
            vt.removeAllElements() ;
        }

        newObjects = new Vector() ;
        inuseObjects = new Hashtable() ;
    }

    /**
        Add a new element to the vector at the next available position. The
        VectorDB does not support insertAt() - use an array for this capability.
    */
    public synchronized void addElement( Object obj )
    {
        newObjects.addElement( obj ) ;
    }

    /**
        Return the number of elements in the vector
    */
    public int size()
    {
        return vt.size() + newObjects.size() ;
    }

    /**
        Return an enumeration of all elements in the vector. As each enumeration
        is accessed via nextElement() the object is read from the persist
        manager if requried.
    */
    public Enumeration elements()
    {
        return new VectorDBEnumerator(vt.elements(), this ) ;
    }

    /**
        An internal method
    */
    synchronized Object get( long oid )
    {
        Long key = new Long(oid) ;

        Object o = inuseObjects.get( key ) ;

        try
        {
            if (o == null && pmgr != null)
            {
                /*  the object has not been read yet */

                o = pmgr.streamInObject( oid ) ;

                inuseObjects.put( key, o ) ;
            }
        } catch (Exception ex)
        {
            ex.printStackTrace() ;
            return null ;
        }

        return o ;
    }

    /**
        Method to support the JOP.persist.Persistable interface
    */
    public synchronized void streamIn( PersistController c ) throws Exception
    {
        pmgr = c.getPersistManager() ;

        vt = (Vector)c.getObject("vt") ;

        /*
            reset the memory cache tables to be consistent
            with the fact that we have just read the object
            from the database
        */

        newObjects = new Vector() ;
        inuseObjects = new Hashtable() ;
    }

    /**
        Method to support the JOP.persist.Persistable interface
    */
    public synchronized void streamOut( PersistController c ) throws Exception
    {
        /*
            Save the persistance manager for later use. The
            purpose of this routine is to save the Vector
            to the database and to set
            the memory 'inuse' and 'newObject' tables to
            reflect the state of the DBMS
        */

        PersistManager pmgrout = c.getPersistManager() ;

        if (pmgr != null && pmgr != pmgrout)
        {
            /*
                This is a different PersistManager than was
                originally used to save or read the HashtableDB.
            */

            VectorDB ot = (VectorDB)pmgrTables.get(pmgrout) ;

            if (ot == null)
            {
                ot = new VectorDB() ;
                pmgrTables.put( pmgrout, ot ) ;
            }

            /*
                Clear the output persist manager specific
                hashtable
            */

            ot.removeAllElements() ;

            /*
                Get the resolved objects for every object in
                the hashtable
            */

            for (Enumeration e = this.elements() ; e.hasMoreElements() ; )
            {
                ot.addElement(e.nextElement()) ;
            }

            ot.streamOut( c ) ;
        }
        else
        {
            pmgr = pmgrout ;

            /*
                For any objects that have been added to the
                vector, stream them out and get their object
                id's then save the id in the hashtable.
            */

            for (Enumeration e = newObjects.elements() ; e.hasMoreElements() ; )
            {
                Object obj = e.nextElement() ;

                long id = pmgr.streamOutObject( obj ) ;

                // System.out.println("Moving key=" + key + ", obj=" + obj + " to hashtable as " + id ) ;

                vt.addElement(new Long(id)) ;
            }

            /*
                Remove the record of the new objects
            */

            newObjects = new Vector() ;

            /*
                Any objects that were referenced also need to
                be streamed out
            */

            for (Enumeration e = inuseObjects.keys() ; e.hasMoreElements() ; )
            {
                Long key = (Long)e.nextElement() ;
                Object obj = inuseObjects.get( key ) ;

                long id = pmgr.streamOutObject( obj ) ;

                if (id != key.longValue())
                {
                    throw new PersistException("Inconsistent (vector) object id" + id + " " + key.longValue()) ;
                }
            }

            /*
                reset the list of 'read' objects
            */

            inuseObjects = new Hashtable() ;

            c.put( "vt", vt ) ;
        }
    }

    /**
        Method to support the JOP.persist.Persistable interface. A VectorDB is
        simply defined as an array of objects. The PersistManager looks after
        the rest.
    */
    public void defineObject( PersistController c ) throws Exception
    {
        c.defineSlot("JOP.util.VectorDB", "vt", PersistController.TYPE_OBJECT ) ;
    }
}


class VectorDBEnumerator implements Enumeration
{
    private Enumeration         enum ;
    private VectorDB            parent ;
    private boolean             newObjects = false ;

    VectorDBEnumerator( Enumeration e, VectorDB p )
    {
        enum = e ;
        parent = p ;
        newObjects = false ;
    }

    public synchronized boolean hasMoreElements()
    {
        if (enum.hasMoreElements())
        {
            return true ;
        }

        if (newObjects)
        {
            return false ;
        }
        else
        {
            newObjects = true ;

            enum = parent.newObjects.elements() ;

            return enum.hasMoreElements() ;
        }
    }

    public synchronized Object nextElement()
    {
        if (newObjects)
        {
            return enum.nextElement() ;
        }
        else
        {
            Long     htid ;

            htid = (Long)enum.nextElement() ;

            if (htid != null)
            {
                return parent.get(htid.longValue()) ;
            }
            else
            {
                return null ;
            }
        }
    }
}

